import java.awt.*;        // Point
import java.util.*;       // Vector
import javax.swing.*;     // UIManager

/**
* This is the main class for Economic Circular Flows program.
* An ECF model has nodes representing economic actors and flows transferring
* payments, real entities etc. between the nodes. The model has different
* layers for e.g. payment flows and real flows.
* <p>
* ECF models are described in
* <a href="http://www.ecf-theory.se">
*          http:www.ecf-theory.se</a>
*/
public class ECF {

	/** The program version. */
	public static String szVersionProgram = "2.9.7.5";
	/** The program release date. */
	public static String szDate = "2007-01-14";
	/** A compatible data version. */
	public static String szVersionData = "2.9 from 2006-01-11";
	/** The main program interface. */
	public static ECFMainFrame frameMain;
	/** Time axis. */
	public static TimeAxis timeaxis;
	/** Layers */
	public static Vector vrlayer = new Vector();
	/** The units used by the model. */
	public static Vector vrunits = new Vector();
	/** The flows used by the model. */
	public static Vector vrflow = new Vector();
	/** The top model in the model hierarchy. */
	public static ECFModel modelTop;
	/** The open model, topmodel or submodel. */
	public static ECFModel modelOpen;
	/** The calculator performs the calculations on the model. */
	public static Calculator calculator;
	/** Flow path in clipboard*/
	public static FlowPath flowpathClipboard = null;

	/** Get number of model layers.
	* @return Number of model layers. */
	public static int getNoOfLayers() { return vrlayer.size(); }

	/** Get time axis of the mode.
	* @return The time axis. */
	public static TimeAxis getTimeAxis() { return timeaxis; }

	/** Get the units used by the model.
	* @return A Vector with the sets of units used by the model. */
	public static Vector getUnits() { return vrunits; }

	/** Add a flow to the model.
	* @param flow The flow to be added. */
	public static void addFlow(Flow flow) {
		vrflow.add(flow);
		Calculator.calculate();
	} // addFlow

	/** Get flow with index iflow.
	* @param iflow Index of flow.
	* @return The flow. */
	public static Flow getFlow(int iflow) {
		int cflow = vrflow.size();
		if (iflow<cflow) {
			return (Flow)vrflow.elementAt(iflow);
		} else {
			System.out.println("ECF.getFlow: Error:: iflow=" + iflow +
				" cflow=" + cflow);
			return null;
		} // if/else
	} // getFlow

	/** Get the flows used by the model.
	* @return A Vector with the flows used by the model. */
	public static Vector getFlows() { return vrflow; }

	/** Get the index of a flow. */
	public static int getIndexOfFlow(Flow flow) {
		return vrflow.indexOf(flow);
	} // getIndexOfFlow

	/** Remove flow from model.
	* @param flow The flow to be removed. */
	public static void removeFlow(Flow flow) {
		vrflow.removeElement(flow);
	} // removeFlow

	/** Start a new model.*/
	public static void startNewModel() {
		// Create general environment for models
		Calculator.clear();
		timeaxis = new TimeAxis();
		vrunits = Units.generateStandardUnits();
		vrlayer = new Vector();
		vrflow = new Vector();
		Flow.fShowDiff = false;
		// Create model
		modelTop = new ECFModel("Reset");
		modelOpen = modelTop;
		vrlayer.add(new Layer("Layer #1"));
		if (frameMain!=null) ECF.frameMain.showCalculatorWarning(false);
	} // startNewModel

	/** Creates the ECF program. */
	public ECF() {
		System.out.println("Welcome to the ECF program version " + szVersionProgram);

		// Set look and feel to the same as current operating system.
		UIManager uiman = new UIManager();
		try {
			uiman.setLookAndFeel(uiman.getSystemLookAndFeelClassName());
		} catch(Exception e) {System.out.println(e);}

		// Create model
		startNewModel();
		// Open main frame
		frameMain = new ECFMainFrame();
		// Calculator
		calculator = new Calculator(new Point(frameMain.getX()+
				frameMain.getWidth()/2,frameMain.getY()+20));
	}
	// ECF

	/** Add a layer to model. */
	public static void addLayer() {
		vrlayer.add(new Layer("Layer #"+ (getNoOfLayers()+1)));
		modelTop.addLayer();
	} // addLayer

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public static void clearUnits(int ilayer) { modelTop.clearUnits(ilayer); }

	/** Close submodel. */
	public static void closeSubModel(ECFModel modelSuper) {
//		System.out.println("ECF.closeSubModel: modelSuper=" + modelSuper);
		if (modelSuper!=null) modelOpen = modelSuper;
		else modelOpen = modelTop;
	} // closeSubModel

	/** Decode interface and model from XML code.
	* @param szXMLCode The xml description of the interface and the model. */
	public static void decodeXML(String szXMLCode) throws XMLParserException {
		startNewModel();
//		Calculator.clear();
//		timeaxis.clearTimeAxis();
		vrlayer.clear();
		vrunits.clear();
		modelTop.clearModel();
		modelOpen = modelTop;
		XMLParser xml = null;

		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();

//				System.out.println("ECF.decodeXML: szTag=" + szTag);
				if (szTag.equals("szVersionData"))  {
					if (szSub.compareTo(ECF.szVersionData)!=0    ) {
						JOptionPane.showMessageDialog(frameMain,
	    					"You are opening a model data file of version " + szSub + " \n" +
	    					"This program version works with data version " + ECF.szVersionData + "\n",
	    					"ECF Model File from old program",
	    					JOptionPane.PLAIN_MESSAGE);
					} // if
				} // szVersionData

				if (szTag.equals("frameMain"))  {
					frameMain.decodeXMLFrame(szSub);
				}
				if (szTag.equals("Zoom_iscale")) {
					ECFMainFrame.setZoom(Integer.valueOf(szSub).intValue());
				}
				if (szTag.equals("timeaxis"))  {
					timeaxis.decodeXML(szSub);
				}
				if (szTag.equals("TimeAxis_iincSelected")) {
					timeaxis.setSelectedIncrement(Integer.valueOf(szSub).intValue());
				}
				if (szTag.equals("layer"))  {
					Layer layer = new Layer();
					layer.decodeXML(szSub);
					vrlayer.add(layer);
				}
				if (szTag.equals("units"))  {
					Units units = new Units();
					units.decodeXML(szSub);
					vrunits.add(units);
				}
				// Flow totals or differences, show connectors
				if (szTag.equals("Flow_fShowDiff")) {
					Flow.fShowDiff = Boolean.valueOf(szSub).booleanValue() ;
				}
				// Show flow numbers
				if (szTag.equals("Flow_fShowFlowNumbers")) {
					Flow.fShowFlowNumbers = Boolean.valueOf(szSub).booleanValue() ;
				}
				if (szTag.equals("FlowConnector_fShowConnectors")) {
					FlowConnector.fShowConnectors = Boolean.valueOf(szSub).booleanValue() ;
				}
				// Flows
				if (szTag.equals("flow"))  {
					Flow flow = new Flow();
					flow.decodeXML(szSub);
					vrflow.add(flow);
					// Calculator.addFlow(flow);
				}
				if (szTag.equals("modelTop"))  {
//					System.out.println("ECF.decodeXML: modelTop");
					modelTop.decodeXML(szSub);
				}
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("ECF:" + e);
		} // try/catch

		if (vrunits.size()==0) vrunits  = Units.generateStandardUnits();
	} // decodeXML

	/** Get the layer with index ilayer
	* @param ilayer Index of the layer.
	* @return The layer.*/
	public static Layer getLayer(int ilayer) {
		try { return (Layer)vrlayer.elementAt(ilayer);
		} catch(Exception e) {
	        // Index may be out of range when menu actions File->New and File->Load
			// System.out.println("ECFModel.getLayer: ilayer=" + ilayer + " vrlayer.size()=" + vrlayer.size());
			// System.out.println(e);
		} // try/catch
		return null; // Dummy return for compiler
	}

	/** Get the layers of the model.
	* @return A Vector of Layer.*/
	public static Vector getLayers() {
		return vrlayer;
	} // getLayers

	/** List XML code.
	* @return XML code. */
	public static String listXML(String szBlanks1) {
		Iterator iter;
		String szBlanks = ""; // To give one compact line
		String szXML = "";

		// Version
		szXML += szBlanks + "<szVersionProgram>" + szVersionProgram + "</szVersionProgram>\n";
		szXML += szBlanks + "<szVersionData>" + szVersionData + "</szVersionData>\n";
		String szDateDataSaved = ECFUtil.getCurrentTime();
		szXML += szBlanks + "<szDateDataSaved>" + szDateDataSaved + "</szDateDataSaved>\n";

		// Main frame
		szXML += szBlanks + "<frameMain>" + frameMain.listXML(szBlanks + "  ") +
			"</frameMain>\n";
		szXML += "<Zoom_iscale>" + ECFMainFrame.getZoom()  +
			"</Zoom_iscale>";
		// TimeAxis
		szXML += szBlanks +"<timeaxis>\n";
		szXML += timeaxis.listXML("  "+szBlanks);
		szXML += szBlanks +"</timeaxis>\n";
		szXML += "<TimeAxis_iincSelected>" + timeaxis.getSelectedIncrement() +
			"</TimeAxis_iincSelected>";

		// Layers
		iter = vrlayer.iterator();
		while(iter.hasNext()) {
			Layer layer = (Layer)iter.next();
			szXML += szBlanks +"<layer>";
			szXML += layer.listXML("");
			szXML += "</layer>";
		} // while layers

		// Units
		iter = vrunits.iterator();
		while(iter.hasNext()) {
			Units units = (Units)iter.next();
			szXML += szBlanks +"<units>";
			szXML += units.listXML("  "+szBlanks);
			szXML += szBlanks +"</units>";
		} // while units
		// Flow totals or differences, show connectors
		szXML += "<Flow_fShowDiff>" + Flow.fShowDiff + "</Flow_fShowDiff>";
		szXML += "<Flow_fShowFlowNumbers>" + Flow.fShowFlowNumbers + "</Flow_fShowFlowNumbers>";
		szXML += "<FlowConnector_fShowConnectors>" + FlowConnector.fShowConnectors +
			"</FlowConnector_fShowConnectors>";
		// Flows
//		Vector vrflow = Calculator.getFlows();
		iter = vrflow.iterator();
		while(iter.hasNext()) {
			Flow flow = (Flow)iter.next();
			szXML += szBlanks +"<flow>" + flow.listXML("") + "</flow>";
		} // while flows
		szXML += szBlanks + "\n<modelTop>\n" + modelTop.listXML(szBlanks + "  ") +
			szBlanks + "\n</modelTop>\n";
		return szXML;
	} // listXML

	/** Open submodel.*/
	public static void openSubModel(SubModel submodel) {
//		System.out.println("ECF.openSubModel:");
		modelOpen = submodel;
		frameMain.repaint();
	} // openSubModel

	/** Set the units used by the model.
	* @param vrunits A Vector with the sets of units used by the model.*/
	public void setUnits(Vector vrunits) { this.vrunits = vrunits; }

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public static void setUnitsIfNotSet(int ilayer,Units units) {
		modelTop.setUnitsIfNotSet(ilayer,units);
	} // setUnitsIfNotSet

	/** Start the program by ECF.main();
	* @param args No arguments are used. */
	public static void main(String[] args) {
		ECF ecf = new ECF();
	}// main

//			Throwable trow = new Throwable();
//			trow.printStackTrace();

} // class ECF