import java.awt.*;        // Graphics2D
import java.awt.event.*;  // MouseEvent
import java.awt.geom.AffineTransform;
import javax.swing.*;     // JOptionPane
import java.util.*;       // Vector

/**
* The <code>ECFModel</code> class describes an Economic Circular Flows model.
* An ECF model has nodes representing economic actors and flows transferring
* payments, real entities etc. between the nodes. The model has different
* layers for e.g. payment flows and real flows.
* <p>
* ECF models are described in
* <a href="http://www.ecf-teori.se">
*          http://www.ecf-theory.se</a>
*/

public class ECFModel implements FlowPathHolder, KeyListener {

	/** Default color of nodes and flowpaths. */
	public final static Color COLOR_DEFAULT = Color.yellow;
	/** Hit color of nodes and flowpaths, indicates the node was hit while
	* dragging flowconnector.*/
	public final static Color COLOR_HIT = Color.green;
	/** Selected color of nodes and flowpaths, indicates the item was selected by mouse. */
	public final static Color COLOR_SELECTED = new Color(255,100,100); // Pink
	/** Color when an object has a associated conditon, as flow balance or prescribed flow. */
	public final static Color COLOR_CONDITION = new Color(255,170,0); // Orange
	public final static Color COLOR_PRESCRIBED = COLOR_CONDITION;
	public final static Color COLOR_CALCULATED = new Color(0,170,255);
	public final static Color COLOR_TRANSPERANT = new Color(255,255,255,0);
	public final static Color COLOR_WARNING = new Color(255,0,0);
	/** Grid size */
	public static int GRID_SIZE = 10;
	/** Flag indicating if model was changed. */
	public static boolean fModelWasChanged = false;
	/** Display scale for model diagram.*/
	public static double scaleDisplay = 0.7171;

	/** The name of this model. */
	protected String szName = "Top model";
	/** Labels for explanatory text,
	* a vector of <a href="ExplanatoryLabel.html">ExplanatoryLabels</a>
	* for putting comments in the diagram area.*/
	protected Vector vrlabel = new Vector();
	/** Selected label*/
	protected ExplanatoryLabel labelSelected;
	/** Vector of all names of models, nodes and flows */
	protected Vector vrNameAll = new Vector();
	/** Vector of nodes */
	protected Vector vrnode = new Vector();
	/** Vector of flowpaths */
	protected Vector vrflowpath = new Vector();
	/** Vector of all flowpaths */
//	private static Vector vrAllFlowPaths = new Vector();
	/** Starting x-coordinate of a LayerPanel for dragging nodes. */
	protected int xStart;
	/** Starting y-coordinate of a LayerPanel for dragging nodes. */
	protected int yStart;
	/** Selected flow path*/
	protected FlowPath flowpathSelected = null;
	/** Node that was hit while dragging.*/
	protected ECFNode nodeHit;
	/** Node that was resized.*/
	protected ECFNode nodeResize = null;
	/** Submodel.*/
	protected ECFModel modelSuper = null;


	/** Creates a new ECF model with no nodes or flows. */
	public ECFModel() {
	} // ECFModel

	/** Creates a new ECF model with no nodes or flows,
	* indices of nodes and flows are reset.
	* @param szReset Dummy argument. */
	public ECFModel(String szReset) {
		Sector.resetIndex();
		Junction.resetIndex();
		Splitter.resetIndex();
		Source.resetIndex();
		ECFProcess.resetIndex();
		InputOutput.resetIndex();
		Terminal.resetIndex();
		Flow.resetIndex();
		fModelWasChanged = false;
		ECFMainFrame.setZoom(0);
	} // ECFModel

	/**
	* Adds a <code>FlowPath</code> to the model. The flow path will be
	* drawn with the reciever endpoint at the panel coordinates (x,y).
	* @param ilayer the index of the layer containing the flow path.
	* @param (x,y) the panel coordinates of the reciever endpoint
	*        of the flow path symbol.
	*/
	public void addFlowPath(int ilayer, int x, int y) {
		FlowPath flowpath = new FlowPath(this,ilayer,x,y);
		vrflowpath.add(flowpath);
		fModelWasChanged = true;
	} // addFlowPath

	/**
	* Adds a <code>FlowPath</code> to the model. The flow path will be
	* drawn with the reciever endpoint at the panel coordinates (x,y).
	* @param ilayer the index of the layer containing the flow path.
	* @param (x,y) the panel coordinates of the reciever endpoint
	*        of the flow path symbol.
	* @param flow the flow to be put into the flow path.
	*/
	public void addFlowPath( int ilayer, int x, int y, Flow flow) {
		FlowPath flowpath = new FlowPath(this,ilayer,x,y,flow);
		vrflowpath.add(flowpath);
		fModelWasChanged = true;
	} // addFlowPath

	/**
	* Adds a copy of the<code>FlowPath</code> to the model.
	* The flow path will be
	* drawn with the reciever endpoint at the panel coordinates (x,y).
	* @param ilayer the index of the layer containing the flow path.
	* @param (x,y) the panel coordinates of the reciever endpoint
	*        of the flow path symbol.
	* @param flow the flow to be put into the flow path.
	*/
	public void addFlowPathCopy( int ilayer, int x, int y, Flow flow) {
		FlowPath flowpath = new FlowPath(this,ilayer,x,y,flow);
		vrflowpath.add(flowpath);
		flow.add(flowpath);
		fModelWasChanged = true;
	} // addFlowPath

	/**
	* Adds a <code>IputOutput</code> node to this model.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the input-output symbol.
	*/
	public void addInputOutput(int x, int y) {
		vrnode.add(new InputOutput(this,x,y));
		fModelWasChanged = true;
	} // addSector

	/**
	* Adds a <code>Junction</code> node to this model.
	* @param ilayer the index of the layer containing the node.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the junction symbol.
	*/
	public void addJunction( int ilayer, int x, int y) {
		vrnode.add(new Junction(this,ilayer,x,y));
		fModelWasChanged = true;
	} // addJunction

	/**
	* Adds a <code>GraphLabel</code> node to this model.
	* @param ilayer the index of the layer containing the label.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the label.
	*/
	public void addLabel( int ilayer, int x, int y) {
		vrlabel.add(new ExplanatoryLabel("Explanation",ilayer,x,y));
		fModelWasChanged = true;
	} // addJunction

	/** Add a layer to model. */
	public void addLayer() {
		// Add a layer to all existing nodes, (Only used by sectors).
		Iterator iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
				node.addLayer(ECF.getNoOfLayers()-1);
		} // while
		fModelWasChanged = true;
	} // addLayer

	/**
	* Adds a <code>Process</code> node to this model.
	* @param ilayer  the index of the layer containing the node.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the process symbol.
	*/
	public void addProcess( int ilayer, int x, int y) {
		vrnode.add(new ECFProcess(this,ilayer,x,y));
		fModelWasChanged = true;
	} // addProcess

	/**
	* Adds a <code>Sector</code> node to this model.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the sector symbol.
	*/
	public void addSector(int x, int y) {
		vrnode.add(new Sector(this,x,y));
		fModelWasChanged = true;
	} // addSector

	/**
	* Adds a <code>Source</code> node to this model.
	* @param ilayer  the index of the layer containing the node.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the source symbol.
	*/
	public void addSource( int ilayer, int x, int y) {
		vrnode.add(new Source(this,ilayer,x,y,"with new source flow"));
		fModelWasChanged = true;
	} // addSource

	/**
	* Adds a <code>Splitter</code> node to this model.
	* @param ilayer the index of the layer containing the node.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the junction symbol.
	*/
	public void addSplitter( int ilayer, int x, int y) {
		vrnode.add(new Splitter(this,ilayer,x,y));
		fModelWasChanged = true;
	} // addSplitter

	/**
	* Adds a <code>Terminal</code> node to this model.
	* @param ilayer the index of the layer containing the node.
	* @param (x,y) the panel coordinates of the upper left corner
	*        of the junction symbol.
	*/
	public void addTerminal( int ilayer, int x, int y) {
		vrnode.add(new Terminal(this,ilayer,x,y));
		fModelWasChanged = true;
	} // addTerminal

	/** Remove all nodes and flowpaths.*/
	public void clearModel() {
		// Delete flows
		Iterator iter = vrflowpath.iterator();
		while (iter.hasNext()) {
			FlowPath flowpath = (FlowPath)iter.next();
			flowpath.disconnectThis();
			ECF.removeFlow(flowpath.getFlow());
		} // while
		// Disconnect all flowpaths from nodes to avoid
		// that the garbage collector can not break the double
		// association.
		// Clear variables from splitters, processes etc.
		// Delete submodel from sectors.
		iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			node.disconnectThis();
			node.clearVariables();
			node.deleteSubmodel();
		} // while
		vrlabel.clear();
		vrnode.clear();
		vrflowpath.clear();
//		vrAllFlowPaths.clear();
		flowpathSelected = null;
		ECF.flowpathClipboard = null;
	} // clearModel

	/** Clear time axis and all variable data (flows etc). */
	public void clearVariableData() {
		ECF.timeaxis.clearVariableData();
	} // clearVariableData

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		// Clear units of flowpaths
		Iterator iter = vrflowpath.iterator();
		while (iter.hasNext()) {
			FlowPath flowpath = (FlowPath)iter.next();
			flowpath.clearUnits(ilayer);
		} // while
		// Clear units of sub models of sectors
		iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			node.clearUnits(ilayer);
		} // while
	} // clearUnits

	/** Close submodel. */
	public void closeSubModel() {
		ECF.closeSubModel(modelSuper);
	} // closeSubModel

	/** Calculate flows. */
	public void collectVariables() {
		Calculator.clear();

		// Flows
		Calculator.vrflow = ECF.vrflow;

		// Flow ratios, Productivities, Prices
		Iterator iter = getNodeIterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			if (node instanceof Splitter) {
				Calculator.addFlowRatios(((Splitter)node).getFlowRatios());
			} // if splitter
			if (node instanceof ECFProcess) {
				Calculator.addProductivities(((ECFProcess)node).getProductivities());
			} // if splitter
			if (node instanceof Sector) {
				Calculator.addPrices(((Sector)node).getPriceRelations());
			} // if splitter
		} // while nodes
	} // collectVariables

	/** Copy selected flow to clipboard. */
	public void copyToClipboard() {
		if (!(flowpathSelected==null)) ECF.flowpathClipboard = flowpathSelected;
	} // copyToClipboard

	/** Copy all values at selected time to all values of all variables. */
	public void copyVariables() {

		Iterator iter;
		int itime = ECF.getTimeAxis().getSelectedIncrement();

		// Flows
//		Vector vrflow = Calculator.getFlows();
		Vector vrflow = ECF.getFlows();
		iter = vrflow.iterator();
		while (iter.hasNext()) {
			Flow flow = (Flow)iter.next();
			flow.copyToAll(itime);
		} // while
		// Flow ratios
		Vector vrflowratio = Calculator.getFlowRatios();
		iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			flowratio.copyToAll(itime);
		} // while
		// Prices
		Vector vrprice = Calculator.getPrices();
		iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			price.copyToAll(itime);
		} // while
		// Productivities
		Vector vrproductivity = Calculator.getProductivities();
		iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
			productivity.copyToAll(itime);
		} // while
	} // copyVariables

	/**
	* Decode model from XML code.
	* @param szXMLCode The xml description of the model.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		String szTag = "-";
		String szSub = "";
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				szTag = xml.tag();
				szSub = xml.sub();
//				System.out.println("ECFModel.decodeXML: szTag=" + szTag);
				// Model name
				if (szTag.equals("szName")) {
					szName = szSub;
				} // if szName
				if (szTag.equals("label"))  {
					ExplanatoryLabel label = new ExplanatoryLabel("Explanation");
					label.decodeXML(szSub);
					vrlabel.add(label);
				}
				if (szTag.equals("flowpath"))  {
					FlowPath flowpath = new FlowPath(this,null);
					flowpath.decodeXML(szSub);
					vrflowpath.add(flowpath);
				} // flowpath
				if (szTag.equals("sector")) {
					Sector sector = new Sector(this,20,20);
					sector.decodeXML(szSub);
					vrnode.add(sector);
				} // sector
				if (szTag.equals("junction")) {
					Junction junction = new Junction(this,0,30,30);
					junction.decodeXML(szSub);
					vrnode.add(junction);
				} // junction
				if (szTag.equals("process")) {
					ECFProcess process = new ECFProcess(this,0,30,30);
					process.decodeXML(szSub);
					vrnode.add(process);
				} // junction
				if (szTag.equals("inputoutput")) {
					InputOutput inputoutput = new InputOutput(this,20,20);
					inputoutput.decodeXML(szSub);
					vrnode.add(inputoutput);
				} // sector
				if (szTag.equals("splitter")) {
					Splitter splitter = new Splitter(this,0,30,30);
					splitter.decodeXML(szSub);
					vrnode.add(splitter);
				} // junction
				if (szTag.equals("source")) {
					Source source = new Source(this,0,40,40);
					source.decodeXML(szSub);
					vrnode.add(source);
				} // source
				if (szTag.equals("terminal")) {
					Terminal terminal = new Terminal(this,0,40,40);
					terminal.decodeXML(szSub);
					vrnode.add(terminal);
				} // terminal
				szXMLCode = xml.remainder();
			} // while
			Calculator.calculate();
		} catch(Exception e) {

			Throwable trow = new Throwable();
			trow.printStackTrace();

			JOptionPane.showMessageDialog(ECF.frameMain,
				"Could not decode XML model file: \n" + e +
				"\nMail your model file to olert@swipnet.se",
				"ECFModel.decodeXML",JOptionPane.WARNING_MESSAGE);
			System.out.println("ECFModel.decodeXML: " + e + "\n");
			System.out.println("  szTag= '" + szTag + "'\n");
			System.out.println("  szSub= '" + szSub + "'\n");
			System.exit(1);
		} // try/catch
	} // decodeXML

	/**
	* Deselects all nodes of this model.
	*/
	public void deselectAllNodes() {
		Iterator iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			node.setSelected(false);
		} // while
	} // deselectAllNodes

	/**
	* Deletes selected nodes and flow paths of this model.
	*/
	public void deteteSelectedElements() {
		// Delete Label
		if (labelSelected!=null) {
			vrlabel.remove(labelSelected);
			fModelWasChanged = true;
			return;
		}  // Label
		// Delete flowpath
		if (flowpathSelected!=null && flowpathSelected.hitLabel()) {
			flowpathSelected.disconnectThis();
			vrflowpath.remove(flowpathSelected);
//			vrAllFlowPaths.remove(flowpathSelected);
			Flow flow = flowpathSelected.getFlow();
			if (flow.hasMoreFlowPaths()) flow.remove(flowpathSelected);
			else ECF.removeFlow(flow);
			fModelWasChanged = true;
			Calculator.calculate();
			return;
		} // flowpath
		// Delete nodes
		Vector vrnodeKeep = new Vector();  // New vector for ramaining nodes
		Iterator iter = vrnode.iterator(); // removeElement(Object obj) may
		while (iter.hasNext()) {           // not be applied to vrnode while
			ECFNode node = (ECFNode)iter.next();  // iterating.
			if (node.isSelected()) {
				node.disconnectThis();
				node.clearVariables();
				node.deleteSubmodel();
				fModelWasChanged = true;
			} else
				vrnodeKeep.add(node);
		} // while
		vrnode = vrnodeKeep;
		Calculator.calculate();
	} //

	/**
	* Finds a flowpath of this model. An <code>FlowPath</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return FlowPath if a flowpath was found, else <code>null</code>.
	*/
	public FlowPath findFlowPath(int ilayer,int x,int y) {
		Iterator iter = vrflowpath.iterator();
		while (iter.hasNext()) {
			FlowPath flowpath = (FlowPath)iter.next();
			boolean fFound = flowpath.contains(ilayer,x,y);
			if (fFound) {
				return flowpath;
			} // if
		} // while
		return null;
	} // findFlowPath

	/**
	* Finds a label of this model. An <code>ExplanatoryLabel</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return ExplanatoryLabel if a label was found, else <code>null</code>.
	*/
	public ExplanatoryLabel findLabel(int ilayer,int x,int y) {
		Iterator iter = vrlabel.iterator();
		while (iter.hasNext()) {
			ExplanatoryLabel label = (ExplanatoryLabel)iter.next();
			boolean fFound = label.contains(ilayer,x,y);
			if (fFound) {
				return label;
			} // if
		} // while
		return null;
	} // findLabel

	/**
	* Finds a node of this model. An <code>ECFNode</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>. Returns <code>null</code> if no node is
	* found.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return ECFNode if a node was found, else <code>null</code>.
	*/
	protected ECFNode findNode(int ilayer,int x,int y) {
		// EFFECTS: If (x,y) lies within nodes symbol, returns the node,
		//   else returns null.
		Iterator iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			boolean fFound = node.contains(ilayer,x,y);
			if (fFound) {
				return node;
			} // if
		} // while
		return null;
	} // findNode

	/** Get a vector of the flows of the model.
	* Could be replaced by call to ECF.getFlows()*/
	public Vector getFlows() {
		return ECF.getFlows();
	} // getFlows

	/** Get a vector of the flows ot the model.
	* Use to check consistency with flows in Calculator. */
/*	public Vector getFlows() {
		Vector vrflow = new Vector();
		// Flow paths at this model.
		Iterator iterPath = vrflowpath.iterator();
		while (iterPath.hasNext()) {
			FlowPath flowpath = (FlowPath)iterPath.next();
			Flow flow = flowpath.getFlow();
			if (!vrflow.contains(flow)) vrflow.add(flow);
		} // while flowpath
		// Flow paths at submodels.
		Iterator iterNode = vrnode.iterator();
		while (iterNode.hasNext()) {
			ECFNode node = (ECFNode)iterNode.next();
			if (node instanceof Sector) {
				Sector sector = (Sector)node;
				SubModel submodel = sector.getSubModel();
				if (submodel!=null) {
					Vector vrflowSub= submodel.getFlows();
					Iterator iterFlowSub = vrflowSub.iterator();
					while (iterFlowSub.hasNext()) {
						Flow flow = (Flow)iterFlowSub.next();
						if (!vrflow.contains(flow)) vrflow.add(flow);
					} // while
				} // if submodel
			} // if node
			if (node instanceof Source) {
				Flow flow = ((Source)node).getSourceFlow();
				if (!vrflow.contains(flow)) vrflow.add(flow);
			} // if source
		} // while flowpath
		return vrflow;
	} // getFlows
*/
	/** Get the flow path with index iflowpath.
	* @param iflowpath The index of the flow path. */
	public FlowPath getFlowPath(int iflowpath) {
		return (FlowPath)vrflowpath.elementAt(iflowpath); }

	/** Get the index of a flowpath of the model.
	* @param flowpath The flowpath for which the index is wanted.
	* @return The index of the flowpath. */
	public int getFlowPathIndex(FlowPath flowpath) {
		int iflowpath = vrflowpath.indexOf(flowpath);
		return iflowpath;
	} // getFlowPathIndex

	/** Get model iterator.
	* @return A model iterator.*/
	public Iterator getModelIterator() {
		Vector vrmodel = new Vector();
		vrmodel.add(this);
		Iterator iterTop = vrnode.iterator();
		while (iterTop.hasNext()) {
			ECFNode node = (ECFNode)iterTop.next();
			if (node instanceof Sector) {
				Sector sector = (Sector)node;
				SubModel submodel = sector.getSubModel();
				if (submodel!=null) {
					Iterator iterSub = submodel.getModelIterator();
					while (iterSub.hasNext()) {
						ECFModel modelSub = (ECFModel)iterSub.next();
						vrmodel.add(modelSub);
					} // while
				} // if submodel
			} // if node
		} // while
		return vrmodel.iterator();
	} // getModelIterator

	/** Get node iterator.
	* @return A node iterator.*/
	public Iterator getNodeIterator() {
		Vector vrnodeAll = new Vector();
		Iterator iterTop = vrnode.iterator();
		while (iterTop.hasNext()) {
			ECFNode node = (ECFNode)iterTop.next();
			vrnodeAll.add(node);
			if (node instanceof Sector) {
				Sector sector = (Sector)node;
				SubModel submodel = sector.getSubModel();
				if (submodel!=null) {
					Iterator iterSub = submodel.getNodeIterator();
					while (iterSub.hasNext()) {
						ECFNode nodeSub = (ECFNode)iterSub.next();
						vrnodeAll.add(nodeSub);
					} // while
				} // if submodel
			} // if node
		} // while
		return vrnodeAll.iterator();
	} // getNodeIterator

	/** Get the name of the model.
	* @return The name of the model. */
	public String getName() { return szName; }

	/** List XML code for this model.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		Iterator iter;
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		// Labels
		iter = vrlabel.iterator();
		while(iter.hasNext()) {
			ExplanatoryLabel label = (ExplanatoryLabel)iter.next();
			szXML += szBlanks +"<label>" + label.listXML("") + "</label>";
		} // while labels
		// Flowpaths
		iter = vrflowpath.iterator();
		while(iter.hasNext()) {
			FlowPath flowpath = (FlowPath)iter.next();
			szXML += szBlanks +"<flowpath>";
			szXML += flowpath.listXML("");
			szXML += "</flowpath>";
		} // while flowpaths
		// Model name
		szXML += szBlanks +"<szName>" + szName + "</szName>";
		// Nodes
		iter = vrnode.iterator();
		while(iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			szXML += szBlanks +"<"+node.xmltag()+">";
			szXML += node.listXML("  "+szBlanks);
			szXML += szBlanks +"</"+node.xmltag()+">";
		} // while nodes
		return szXML;
	} // listXML

	/** Make a new empty model. */
	public void newModel() {
		clearModel();
		addLayer();
	} // newModel

	/**
	* Paints the labels, nodes and flow paths of the model.
	* @param g2 the graphics context of a panel or a canvas to which
	*           this model is drawn.
	* @param ilayer the index of the model layer beeing drawn.
	*/
	public void paint(Graphics2D g2,int ilayer) {
		// System.out.println("ECFModel.paint:");
		// if (ECF.frameMain!=null) ECF.frameMain.setTitleMainFrame();
		for (int inode=0;inode<vrnode.size();inode++) {
			((ECFNode)vrnode.get(inode)).paint(g2,ilayer);
		}// for
		for (int iflowpath=0;iflowpath<vrflowpath.size();iflowpath++) {
			((FlowPath)vrflowpath.get(iflowpath)).paint(g2,ilayer);
		}// for
		for (int ilabel=0;ilabel<vrlabel.size();ilabel++) {
			((ExplanatoryLabel)vrlabel.get(ilabel)).paint(g2,ilayer);
		}// for
	} // paint

	/**
	* Selects a node or flow path in layer <code>ilayer</code> if
	* the model element contains the coordinates <code>(x,y)</code>,
	* otherwise deselects all elements. Sets the start point
	* <code>(xStart,yStart)</code> for dragging nodes or a flow path
	* label.
	* @param fAddSelection if true adds element to previous selections,
	*        else selects only this node or flow path.
	* @param fEditProperties if true opens a window for editing the
	*        properties of this element.
	* @param ilayer the index of the layer to contain the element.
	* @param (x,y) the coordinates to be contained by the element.
	*/
	public void selectModelElement(boolean fAddSelection,
				boolean fEditProperties, int ilayer, int x,int y) {

		// Deselect labels and flow paths
		if (labelSelected!=null) labelSelected.setDeselected();
		labelSelected = null;
		if (flowpathSelected!=null) flowpathSelected.setDeselected();
		flowpathSelected = null;

		// Find label
		labelSelected = findLabel(ilayer,x,y);
		if (labelSelected!=null) {
			deselectAllNodes();
			xStart = x; yStart = y;
			if (!fEditProperties) labelSelected.setStartPos();
			else {
				labelSelected.openDialog();
				fModelWasChanged = true;
			} // if/else
			return;
		}

		// Find flow path
		flowpathSelected = findFlowPath(ilayer,x,y);
		if (flowpathSelected!=null) {
			deselectAllNodes();
			xStart = x; yStart = y;
			if (!fEditProperties) flowpathSelected.setStartPos();
			else {
				flowpathSelected.openDialog();
				fModelWasChanged = true;
			} // if/else
			return;
		}

		ECFNode nodeSelected = findNode(ilayer,x,y);
		if (nodeSelected==null) {
			// No node found, deselect all
			deselectAllNodes();
		} else {
			// A node was found
			xStart = x; yStart = y;
			if (!fEditProperties)  {
				// Do not edit,set selected only
				if (!fAddSelection) deselectAllNodes();
				nodeSelected.setSelected(true);
				Iterator iter = vrnode.iterator();
				while (iter.hasNext()) {
					ECFNode node = (ECFNode)iter.next();
					node.setStartPos();
				} // while
			} else  {
				// Edit node properties
				nodeSelected.openDialog();
				fModelWasChanged = true;
			} // if mask
		} // if nodeSelected

	} // selectModelElement

	/** Set the name of the model.
	* @param szName The name of the model. */
	public void setName(String szName) {
		this.szName =szName;
		fModelWasChanged = true;
	} // setName

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(int ilayer,Units units) {
		// Set units of flowpaths
		Iterator iter = vrflowpath.iterator();
		while (iter.hasNext()) {
			FlowPath flowpath = (FlowPath)iter.next();
			flowpath.setUnitsIfNotSet(ilayer,units);
		} // while
		// Set units of sub models of sectors
		iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			node.setUnitsIfNotSet(ilayer,units);
		} // while
		fModelWasChanged = true;
	} // setUnitsIfNotSet

	/** Snap selected node to grid. */
	public void snapNodeToGrid() {
		Iterator iter = vrnode.iterator(); // removeElement(Object obj) may
		while (iter.hasNext()) {           // not beappied to vrnode while
			ECFNode node = (ECFNode)iter.next();  // iterating.
			if (node.isSelected()) {
				node.snapToGrid();
			} // if
		} // while
	} // snapNodeToGrid

	public String toString() { return "ECFModel " + szName; }


	// Mouse events, invoked by PointerTool
	// ------------------------------------
	/**
	* Translates the selected nodes or flowpath label by the distance by which
	* the mouse was dragged since last selection.
	* @param ilayer index of model layer.
	* @param (x,y) the end point in panel coordinates to which the mouse
	*        was dragged from the point <code>(xStart,yStart)</code>.
	*/
	public void mouseDraggedTo(MouseEvent e, int ilayer) {
		int x = (int)(e.getX()/scaleDisplay);
		int y = (int)(e.getY()/scaleDisplay);
		int dx = x - xStart;
		int dy = y - yStart;
		Iterator iter = vrnode.iterator();
		while (iter.hasNext()) {
			ECFNode node = (ECFNode)iter.next();
			if (node.isSelected()) {
				node.translate(dx,dy);
				fModelWasChanged = true;
			} // if
		} // while
		// Translate connector or label of flowpath
		if (flowpathSelected!=null) {
			flowpathSelected.translate(dx,dy);
			fModelWasChanged = true;
	 		ECFNode nodeHit = findNode(ilayer,x,y);
	 		if (nodeHit!=null && !flowpathSelected.hitLabel())  {
				nodeHit.setHit(true);
			} // if
			else if (this.nodeHit!=null) this.nodeHit.setHit(false);
			this.nodeHit = nodeHit;
		} // if
		// Translate explanatory label
		if (labelSelected!=null) {
			labelSelected.translate(dx,dy);
			fModelWasChanged = true;
		} // if
	} // mouseDraggedTo

	/**
	* Sets cursor on mouse move to resize cursor if sector hit.
	* @param panel the panel on which the model layer is drawn.
	* @param (x,y) the end point in panel coordinates to which the mouse
	*        was dragged from the point <code>(xStart,yStart)</code>.
	*/
	public void mouseMovedTo(MouseEvent e, LayerPanel panel) {
		int x = (int)(e.getX()/scaleDisplay);
		int y = (int)(e.getY()/scaleDisplay);
		if (nodeResize!=null) nodeResize.quitResize();
		nodeResize=findNode(0,x,y);
		if (nodeResize!=null) panel.setCursor(nodeResize.getResizeCursor(x,y));
		else panel.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	} // mouseMovedTo

	/** Select model element when mouse pressed.
	* @param e The mouse event.
	* @param ipanel The index of the active panel.*/
	public void mousePressed(MouseEvent e, int ipanel) {
		boolean fAddSelection = e.isShiftDown();
		boolean fEditProperties = e.isMetaDown();
		int x = (int)(e.getX()/scaleDisplay);
		int y = (int)(e.getY()/scaleDisplay);
		selectModelElement(fAddSelection,fEditProperties,ipanel,x,y);
		if (e.isControlDown()) copyToClipboard();
	} // mousePressed

	/** Mouse released, connect node and flowpath if node was hit.
	* @param ipanel The index of the active panel.*/
	public void mouseReleased(int ipanel) {
		snapNodeToGrid();
		try {
			if (nodeHit!=null && flowpathSelected!=null) {
				nodeHit.connectFlowPath(ipanel, flowpathSelected);
				nodeHit = null;
				flowpathSelected = null;
			} // if nodeHit
		} catch(Exception e) {
			System.out.println(e);
		}
	} // mouseReleased



	// KeyListener methods
	// -------------------

	/**  Check id key typed is Crtl+C==3. */
	public void keyTyped(KeyEvent e) {
		if (e.getKeyChar()==3) copyToClipboard();
	} // keyTyped

	/** Not used. */
	public void keyReleased(KeyEvent e) {
	} // keyReleased

	/** Not used. */
	public void keyPressed(KeyEvent e) {
	} // keyPressed
} // ECFModel