import java.awt.*;
import java.awt.event.*;  // ActionListener
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;

/**
* A node of the economic system. A node is an actor in the economy,
* for example a <code>Sector</code>, <code>Junction</code> or
* <code>Source</code>. It is connected to other nodes by flowpaths.
*/
public abstract class ECFNode {

	/** No side of the node is referenced, from for example a
	* <A href=FlowConnector.html>FlowConnector</A>. */
	public final static int NONE  = 0;
	/** East (right) side of the node is referenced, from for example a
	* <A href=FlowConnector.html>FlowConnector</A>. */
	public final static int EAST  = 1;
	/** West (left) side of the node is referenced, from for example a
	* <A href=FlowConnector.html>FlowConnector</A>. */
	public final static int WEST  = 2;
	/** North (upper) side of the node is referenced, from for example a
	* <A href=FlowConnector.html>FlowConnector</A>. */
	public final static int NORTH = 3;
	/** South (bottom) side of the node is referenced, from for example a
	* <A href=FlowConnector.html>FlowConnector</A>. */
	public final static int SOUTH = 4;

	/** Number of layers of this node. */
	protected int clayer =1;
	/** The fill color for the node symbol. */
	protected Color colorFill = ECFModel.COLOR_DEFAULT;
	/** Indicates if the node is selected for delete or translate actions.*/
	protected boolean fSelected = false;
	/** Height of rectangle enclosing node symbol. */
	protected int height = 30;
	/** The model to which the node belongs. */
	protected ECFModel model = null;
	/** Parent model to which the node belongs. */
	protected ECFModel modelParent = null;
	/** Paints the interior, edges and name of the node symbol. */
	protected Painter painter;
	/** Shape of node symbol. */
	protected Shape shape;        // Shape of node
	/** Name of the node. */
	protected String szName = "Node";
	/** Text geometry for node name. */
	protected TextGeometry textName = new TextGeometry(szName);
	/** Nodeconnectors for each layer. */
	protected Vector vrNodeConnectors = new Vector();
	/** Width of rectangle enclosing node symbol. */
	protected int width = 80;
	/** Width of border of node symbol (pixels). */
	protected int widthStroke = 1;
	/** X-coordinate of upper left corner of node symbol. */
	protected int x;
	/** X-coordinate where mouse pressed before translating. */
	protected int xStart;
	/** Y-coordinate of upper left corner of node symbol. */
	protected int y;
	/** Y-coordinate where mouse pressed before translating. */
	protected int yStart;

	/**
	* Creates a new node of the ECF model.
	* @param x,y coordinates of the upper left corner of the node symbol
	*/
	public ECFNode(ECFModel model, int x, int y) {
		this.model = model;
		modelParent = ECF.modelOpen;
		this.x = x;
		this.y = y;
		snapToGrid();
		widthStroke = 1;
		makePainter();
		shape = getShape();
	} // ECFNode

	/**
	* Creates a new node of the ECF model.
	* @param x,y coordinates of the upper left corner of the node symbol
	* @param colorFill the fill color.
	*/
	public ECFNode(ECFModel model, int x, int y, Color colorFill) {
		this.model = model;
		modelParent = ECF.modelOpen;
		this.x = x;
		this.y = y;
		this.colorFill = colorFill;
		snapToGrid();
		widthStroke = 1;
		makePainter();
		shape = getShape();
	} // ECFNode

	/** Add a layer with index ilayer to node (sectors only).
	* @param ilayer Index of the layer to be added.*/
	public void addLayer(int ilayer) { ;
	} // addLayer

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		// Clear units of sub models of sectors
		// Implemented in Sectors and Sources
	} // clearUnits

	/** Clear variables of nodes. Implemented in splitter,
	* processes etc. */
	public void clearVariables() {
	} // clearVariables()

	/** Connect flow path to this node.
	* @param ilayer Index of the layer of the flow path
	* @param flowpath the flow path to be connected. */
	public void connectFlowPath(int ipanel, FlowPath flowpath) {
		// Find node connectors
		Iterator iter = vrNodeConnectors.iterator();
		while (iter.hasNext()) {
			NodeConnectors conn = (NodeConnectors)iter.next();
			if (ipanel==conn.getLayerIndex()) {
				conn.connectFlowPath(flowpath);
				break;
			} // if
		} // while
		// Update fill color and recalculate
		updateNodeContext();
	} // connectFlowPath

	/** Update fill color of node symbol and recalculate flows. */
	protected void updateNodeContext() {
		colorFill = ECFModel.COLOR_DEFAULT;
		makePainter();
		Calculator.calculate();
	} // updateNodeContext

	/**
	* Checks if the node belongs to panel <code>ipanel</code> and contains
	* the coordinates <code>x,y</code>.
	* @param ipanel index of the panel to which the node belongs
	* @param x,y coordinates to be contained within the node symbol
	*/
	public abstract boolean contains(int ipanel,int x,int y);

	/**
	* Decode node from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				decodeXMLItems(szTag,szSub);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("ECFNode.decodeXML: " + e +
			"\n  " + szXMLCode);
			System.exit(2);
		} // try/catch
	} // decodeXML

	/**
	* Decode node items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		if (szTag.equals("szName")) setName(szSub);
		if (szTag.equals("height")) height = Integer.parseInt(szSub);
		if (szTag.equals("width")) width = Integer.parseInt(szSub);
		if (szTag.equals("x")) x = Integer.parseInt(szSub);
		if (szTag.equals("y")) y = Integer.parseInt(szSub);
		if (szTag.equals("vrNodeConnectors")) decodeXMLNodeConnectors(szSub);
	} // decodeXML

	/**
	* Decode node connectors from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLNodeConnectors(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			Iterator iter = vrNodeConnectors.iterator();
			while (iter.hasNext()) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("NodeConnectors")) {
					NodeConnectors conn = (NodeConnectors)iter.next();
					conn.decodeXML(szSub);
				} // if
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
		} // try/catch
	} // decodeXMLNodeConnectors

	/** Delete the submodel of the node, only applicable for Sector.
	* Implemented in the sector class.*/
	public void deleteSubmodel() {
	} // deleteSubmodel

	/**
	* Release associations to flow paths before deleting node.
	*/
	public void	disconnectThis() {
		Iterator iter = vrNodeConnectors.iterator();
		while (iter.hasNext()) {
			NodeConnectors conns = (NodeConnectors)iter.next();
			conns.disconnectAllFlows();
		} // while
	} // deleteThis

	/** Get an iterator for all equations of this node.
	* @return An equations iterator. */
	public Iterator getEquationsIterator() throws Exception {
		Vector vrNodeEquations = new Vector();

		// Get node specific equations
		Vector vrequations = getNodeSpecificEquations();
		Iterator iterNodeSpecEqu = vrequations.iterator();
		while (iterNodeSpecEqu.hasNext()) {
			vrNodeEquations.add(iterNodeSpecEqu.next());
		} // while

		// Get equations specific for the layers of the node
		Iterator iterLayer = vrNodeConnectors.iterator();
		while (iterLayer.hasNext()) {
			NodeConnectors conns = (NodeConnectors)iterLayer.next();
			if (conns.hasFlowBalance()) {
				vrNodeEquations.add(conns.getFlowBalanceEquation());
			} // if
		} // while

		return vrNodeEquations.iterator();
	} // getEquationsIterator

	/** Get the flow path with index iflowpath.
	* @param iflowpath The index of the flow path. */
	public FlowPath getFlowPath(int iflowpath) {
		return (FlowPath)model.getFlowPath(iflowpath);}

	/** Get the name of the node.
	* @return The name of the node. */
	public String getName() { return szName; }

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public abstract NodeConnectors getNodeConnectors(int ilayer);

	/** Get parent model.
	* @return The parent model of the sector. */
	public ECFModel getParentModel() {
		return modelParent;
	} // getParentModel

	/** Get node specific equations, to be specified in subtypes.
	* @return A vector of <a href="LinearEquation.html"><code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() throws Exception { return new Vector();}

	/** Get number of model layers.
	* @return Number of model layers. */
	public int getNoOfLayers() { return vrNodeConnectors.size(); }

	/**
	* Get the surrounding rectangle of this node symbol.
	*/
	protected Rectangle getRect() {
		return new Rectangle(x,y,width,height);
	} // getRect

	public Cursor getResizeCursor(int x,int y) {
		return new Cursor(Cursor.DEFAULT_CURSOR);
	} // getResizeCursor

	/**
	* Get the shape of this node symbol.
	*/
	protected abstract Shape getShape();

	/** Check if node has a flow balance for this layer.
	* @return true if has flow balance, else false.*/
	public abstract boolean hasFlowBalance(int ilayer);

	/**
	* Check if a the node is selected.
	* @return fSelected = true if selected, else false
	*/
	public boolean isSelected() { return fSelected;}

	/** List XML code for this node.
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = "";
		szXML += szBlanks +"<szName>"+szName+"</szName>";
		szXML += "<height>"+height+"</height>";
		szXML += "<width>"+width+"</width>";
		szXML += "<x>"+x+"</x>";
		szXML += "<y>"+y+"</y>";
		szXML += szBlanks +"<vrNodeConnectors>";
		Iterator iter = vrNodeConnectors.iterator();
		while (iter.hasNext()) {
			NodeConnectors nodeconn = (NodeConnectors)iter.next();
			szXML += szBlanks +"  <NodeConnectors>";
			szXML += nodeconn.listXML("    "+szBlanks);
			szXML += szBlanks +"  </NodeConnectors>";
		} // while
		szXML += szBlanks +"</vrNodeConnectors>";
		return szXML;
	} // listXML

	/**
	* Make a painter to paint the symbol of the node.
	*/
	protected void makePainter() {
		Painter painterFill = new FillPainter(colorFill);
		Stroke stroke = new BasicStroke(widthStroke);
		Painter painterDraw = new DrawPainter(Color.black,stroke);
		painter = new MultiPainter(painterFill,painterDraw);
	} // makePainter

	/** Open a dialog window for node properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		JFrame mainframe = ECF.frameMain;
		int x = mainframe.getX();
		int y = mainframe.getY();
		ECFNodeDialog dialog = new ECFNodeDialog(mainframe, this, new Point(x+20,y+20));
		dialog.show();
	} // openDialog

	/**
	* Paint the symbol with painter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		// REQUIRES. g2 is not null.
		if (g2==null) return;
		painter.paint(g2, getShape());
		if (hasFlowBalance(ilayer))  {
			// Colored rectange for flow balance condition
			NodeConnectors conns = getNodeConnectors(ilayer);
			int xFB = x-4; int yFB = y-4;
			if (conns!=null && conns.mayApplyFlowBalance()) {
				g2.setColor(ECFModel.COLOR_CONDITION);
				g2.fillRect(xFB,yFB,15,10);
			} // if
			// Text for flow balance condition
			g2.setColor(Color.black);
			g2.drawString("FB",xFB,yFB+10);
		} // if
	} // paint

	/** Set resize side to NONE (applies to Sector). */
	public void quitResize() {;}

	/** Receive a message that a new flow path connector has been added
	* from NodeConnectors when connecting a new flow path by the
	* connectFlowPath method above.
	* @param iconnNew Index of added connector. */
	public void registerNewConnector(int iconnNew) {
	} // registerNewConnector

	/** Remove flow from node subtype, to be implemented in subtype. */
	public void removeFlowReference(FlowPath flowpath) {;}

	/** Reposition all node connectors after adding or removing a flowpath.*/
	protected void repositionConnectors() {
		Iterator iter = vrNodeConnectors.iterator();
		while (iter.hasNext()) {
			NodeConnectors nodeconn = (NodeConnectors)iter.next();
			nodeconn.setPositions();
		} // while
	} // repositionConnectors

	/** Set if node has a flow balance for this layer.
	* @param f true if has flow balance, else false.*/
	public void setHasFlowBalance(int ilayer, boolean f) {;}

	/**
	* Change color of the node symbol to indicate that the node was hit
	* while dragging a <A href= "FlowConnector.html">FlowConnector</A>.
	*/
	public void setHit(boolean fHit) {
			if (fHit) colorFill = ECFModel.COLOR_HIT;
			else colorFill = ECFModel.COLOR_DEFAULT;
			makePainter();
	} // setHit

	/** Set the name of the node.
	* @param szName The name to be set.
	*/
	public void setName(String szName) {
		this.szName = szName;
		textName.setText(szName);
	} // setName

	/**
	* Set the node in selected state.
	* @param fSelect true if node is selected, else false
	*/
	public void setSelected(boolean fSelect) {
		if (fSelect) {
			fSelected = true;
			colorFill = ECFModel.COLOR_SELECTED;
		} else {
			fSelected = false;
			colorFill = ECFModel.COLOR_DEFAULT;
		} // if
		makePainter();
	} // setSelected

	/**
	* Set start position (before translating) of the node to is current
	* position.
	*/
	public void setStartPos() {
		xStart = x;   yStart = y;  // Upper left corner start position
	} // setStartPos

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(int ilayer,Units units) {
		// Set units of sub models of sectors
		// Implemented in Sectors and Sources
	} // setUnitsIfNotSet

	/** Snap to grid. */
	public void snapToGrid() {
		double tol = 0.5;
		x = ECFModel.GRID_SIZE*(int)Math.round((x+tol)/ECFModel.GRID_SIZE);
		y = ECFModel.GRID_SIZE*(int)Math.round((y+tol)/ECFModel.GRID_SIZE);
		width = ECFModel.GRID_SIZE*(int)Math.round((width+tol)/ECFModel.GRID_SIZE);
		height = ECFModel.GRID_SIZE*(int)Math.round((height+tol)/ECFModel.GRID_SIZE);
		repositionConnectors();
	} // snapToGrid

	/** Show properties of this ECFNode
	* @return A string telling the properties.*/
	public String toString() {
		return "ECFNode: szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

    /**
    * Translate the node symbol.
    * @param dx translation in x direction (pixels)
    * @param dy translation in y direction (pixels)
    */
	public void translate(int dx, int dy) {
		x = Math.min(Math.max(0,xStart + dx),LayerPanel.PANEL_WIDTH-width);
		y = Math.min(Math.max(0,yStart + dy),LayerPanel.PANEL_HEIGHT-height);
		repositionConnectors();
	} // translate

	/** Return the XML code tag fot this node.
	* @return XML tag. */
	public String xmltag() {return "node";}

} // ECFNode