import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.util.*;

/**
* A process is a node that is connected
* to flows of different kinds. It belongs to one model layer only. A process
* transforms input flows to output flows. The ratio  between an output flow
* and an input flow is called a productivity factor (productivity).
*/
public class ECFProcess extends AbstractJunction {

	/** A vector of <a href="Productivity.html">Productivitys</a>*/
	protected Vector vrproductivity = new Vector();
	/** Index of process. */
	static int iprocess = 1;
	/** Reset process index. */
	public static void resetIndex() { iprocess = 1; }

	/**
	* Creates a new process.
	* @param time axis the time axis for flow ratios.
	* @param ilayer index of the layer to which the process belongs
	* @param x,y coordinates of the upper left corner of the process symbol
	*/
	public ECFProcess(ECFModel model, int ilayer, int x, int y) {
		super(model,ilayer,x,y);
		setName("Process"+iprocess);
		radius = 20; // Radius of the symbol.
		width = 2*radius;
		height = 2*radius;
		conn.setHasFlowBalance(false);
		iprocess++;
	} // ECFProcess

	/** Clear variables of nodes, for process remove productivities. */
	public void clearVariables() {
		System.out.println("ECFProcess.clearVariables:");
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
		System.out.println("ECFProcess.clearVariables:" + productivity);
			Calculator.removeProductivity(productivity);
		} // while
		vrproductivity.clear();
	} // clearVariables

	/**
	* Decode process items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		if (szTag.equals("Productivity")) {
			Productivity productivity = new Productivity(this);
			productivity.decodeXML(szSub);
			vrproductivity.add(productivity);
		} // if
		super.decodeXMLItems(szTag,szSub);
	} // decodeXMLItems

	/** Get att flows into the process.
	* @return A vector of Flow objects representing flows into the process.*/
	public Vector getFlowsIn() {
		Vector vrflowIn = new Vector();

		// Check all connectors and find inflows
		NodeConnectors conns = getNodeConnectors();
		Iterator iter = conns.getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			if (conn.isHeadEnd()) {
				vrflowIn.add(conn.getFlowPath().getFlow());
			} // if
		} // while

		return vrflowIn;
	} // getFlowsIn

	/** Get att flows out from the process.
	* @return A vector of Flow objects representing flows out from the process.*/
	public Vector getFlowsOut() {
		Vector vrflowOut = new Vector();

		// Check all connectors and find outflows
		NodeConnectors conns = getNodeConnectors();
		Iterator iter = conns.getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			if (conn.isTailEnd()) {
				vrflowOut.add(conn.getFlowPath().getFlow());
			} // if
		} // while

		return vrflowOut;
	} // getFlowsOut

	/** Get a Vector of Productivitys for the process.
	* @return A Vector of <a href="Productivity.html">Productivities</a>.
	*/
	public Vector getProductivities() { return vrproductivity; }

	/** Get node specific equations for productivities.
	* @return A vector of <a href="LinearEquation.html">
	*         <code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() throws Exception {
		Vector vrequations = new Vector();
		// Flow process ratios
		Iterator iterProductivity = vrproductivity.iterator();
		while (iterProductivity.hasNext()) {
			Productivity productivity = (Productivity)iterProductivity.next();
			if (productivity.isPrescribed())
				vrequations.add(getProductivityEquation(productivity));
		} // while
		return vrequations;
	} // getNodeSpecificEquations


	/** Get a productivity equation for for the process.
	* @param productivity A <code>Productivity</code> object.
	* @return An equation for flow ratio.
	*/
	protected LinearEquation getProductivityEquation(Productivity productivity)
		throws Exception {

		int cflows = Calculator.getNumberOfFlows();
		LinearEquation equation= new LinearEquation(cflows);  // Flows + right member

		// Set coefficient for inflow
		double prodty = productivity.getCalcValuePrescribed();
		if (prodty==0.0) throw new Exception("Productivity " +
					productivity.getName() +" of process " + model.getName() +
					"/" + szName +  " is zero");
		equation.setCoefficient(productivity.getFlowDenominator().getIndexOfFlow(),
			prodty);
		// Set coefficient for outflow
		equation.setCoefficient(productivity.getFlowNominator().getIndexOfFlow(),-1);

		equation.setType("Productivity");
//		equation.setNodeName(getName());
		equation.setNode(this);
		equation.setLayerIndex(ilayer);

		return equation;
	} // getProductivityEquation

	/** Creates a <code>Shape</code> of a line from the center <code>p1</code>
	* of the process symbol to the connection point <code>p2</code> of an outflow.
	* @param p1 The center of the process symbol.
	* @param p2 The connection point of an outflow.
	* @return The <code>Shape</code> of a line from point <code>p1</code>
	*         to point <code>p2</code>.*/
	private Shape getShapeRatioLine(Point2D p1, Point2D p2) {
		Line2D line = new Line2D.Double(p1,p2);
		return line;
	} // getShapeRatioLine

	/** List XML code for the process.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = super.listXML(szBlanks);
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
			szXML += szBlanks +"<Productivity>";
			szXML += productivity.listXML("  "+szBlanks);
			szXML += szBlanks +"</Productivity>";
		} // while
		return szXML;
	} // listXML

	/**
	* Make a painter to paint the line to indicate that
	* a flow ratio has been set.
	* @param colorFill the color of the line.
	* @return A Painter.
	*/
	protected Painter makePainterRatioLine(Color colorFill) {
		Painter painterFill = new FillPainter(colorFill);
		int widthStroke = 5;
		Stroke stroke = new BasicStroke(widthStroke);
		Painter painterDraw = new DrawPainter(colorFill,stroke);
		return new MultiPainter(painterFill,painterDraw);
	} // makePainter

	/** Open a dialog window for process properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		Point pnt = new Point(ECF.frameMain.getX()+20,ECF.frameMain.getY()+20);
		ProcessDialog dialog = new ProcessDialog(this, pnt);
		dialog.show();
	} // openDialog

	/**
	* Paint the process.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {

		super.paint(g2,ilayer);
		if (g2==null) return;

		if (this.ilayer==ilayer)  {

			// Lines indicating pairs of flows that are related by
			// a productivity relation,
			// from flow connector to flow connector.
			paintRatioLines(g2);

			// A capitol "P" for splitter in the center of the circle
			g2.setColor(Color.black);
			TextGeometry text = new TextGeometry(new Font("Times",Font.BOLD,18),"P");
			Rectangle rectBounds = text.getBounds();
			int h = (int)rectBounds.getHeight();
			int w = (int)rectBounds.getWidth();
			text.paint(g2, x+radius-w/2, y+radius+h/2);
		}
	} // paint

	/** Paint lines inside the process symbol that indicate the flows
	* for which ratios have been specified.
	* Lines run from center to flow connectors.
	* @param g2 graphics context. */
	private void paintRatioLines(Graphics2D g2) {

		// Check all productivitys
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			// Productivity
			Productivity productivity = (Productivity)iter.next();
			Flow flowIn = productivity.getFlowDenominator();
			Point2D posConnIn = null;
			Flow flowOut = productivity.getFlowNominator();
			Point2D posConnOut = null;
			// Find corresponding flow connector
			NodeConnectors conns = getNodeConnectors();
			Iterator iterConns = conns.getNodeConnectorsIterator();
			while (iterConns.hasNext()) {
				FlowConnector conn = (FlowConnector)iterConns.next();
				FlowPath flowpath = conn.getFlowPath();
				Flow flowConn = flowpath.getFlow();
				if (flowConn==flowIn) posConnIn = conn.getRefPoint();
				if (flowConn==flowOut) posConnOut = conn.getRefPoint();
			} // while connectors
			if (posConnIn!=null && posConnOut!=null) {
				// Draw a line
				Shape shapeRatioLine = getShapeRatioLine(posConnIn,posConnOut);
				if (productivity.isPrescribed()) {
					// Draw a fat colored line
					Color colorFill = ECFModel.COLOR_PRESCRIBED;
					Painter painterRatioLine = makePainterRatioLine(colorFill);
					painterRatioLine.paint(g2,shapeRatioLine);
				} else
					// Draw a thin black line
					g2.setPaint(ECFModel.COLOR_CALCULATED);
					g2.draw(new Line2D.Double(posConnIn,posConnOut));
			} // if ratio line

		} // while productivitys
	} // paintRatioLines

	/**
	* Paints the productivity curves with axes in a diagram.
	* @param g The graphical context.
	* @param xOrigin The x-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param yOrigin The y-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param widthXAxis The width (pixels) of the time axis.
	* @param heightYAxis The height (pixels) of the value axis.
	*/
	public void paintProductivityCurves(Graphics g, int xOrigin, int yOrigin,
		int widthXAxis, int heightYAxis) {
		System.out.println("ECFProcess.paintProductivityCurves: 1");
		// Find maximum of productivity curves
		Productivity productivity;
		double max = 0.0;
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			productivity = (Productivity)iter.next();
			max = Math.max(max,productivity.maxValue());
		} // while
		// Value axis
		System.out.println("ECFProcess.paintProductivityCurves: 2");
		ValueAxis valueaxis = new ValueAxis(max," ");
		valueaxis.paint(g, xOrigin, heightYAxis, yOrigin);
		// Time axis
		if (ECF.timeaxis.getTimeIncrements().size()==0) return;
		ECF.timeaxis.paint(g, xOrigin, xOrigin+widthXAxis, yOrigin);
		// Paint productivity curves
		System.out.println("ECFProcess.paintProductivityCurves: 3");
		iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			productivity = (Productivity)iter.next();
			ValuesPlot valuesplot = new ValuesPlot(ECF.timeaxis.getTimeIncrements(),
				productivity.getDisplayValues(),productivity.getDisplayColors());
			valuesplot.paint(g,xOrigin,yOrigin,widthXAxis,heightYAxis,valueaxis.getScale(),
		           ECF.timeaxis.getSelectedIncrement());
		} // while
		System.out.println("ECFProcess.paintProductivityCurves: end");
	} // paintProductivityCurves

	/** Remove flow from the process and its corresponding productivity.
	* @param flowpath The flow path which productivity shall be removed. */
	public void removeFlowReference(FlowPath flowpath) {

		// Find and remove flow reference
		Flow flow = flowpath.getFlow();
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
			if (productivity.getFlowNominator()==flow || productivity.getFlowDenominator()==flow) {
				ECF.timeaxis.remove(productivity);
				vrproductivity.remove(productivity);
				Calculator.removeProductivity(productivity);
				break;
			} // if
		} // while

		Calculator.calculate();
	} // removeFlowReference

	/** Show properties of this ECFProcess
	* @return A string telling the properties.*/
	public String toString() {
		return "ECFProcess: szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag for this process.
	* @return XML tag. */
	public String xmltag() {return "process";}

} // ECFProcess