import javax.swing.*;     // JFileChooser
import java.io.*;         // File
import java.util.*;		  // Date
import java.text.*;       // DateFormat

/** Exports model data to file.*/
public class Exporter {

	protected String szCommandStart = "";
	protected String szCommentSign = "% ";

	/** The name of the XML model file. */
	protected String szNameXMLFile;
	/** File extension of exported file. */
	protected String szExtension;
	/** File chooser. */
	protected JFileChooser chooser;
	/** Time axis.*/
	protected TimeAxis timeaxis;
	/** Vector of flow objects. */
	protected Vector vrflow;
	/** Vector of flow ratio objects. */
	protected Vector vrflowratio;
	/** Vector of price objects. */
	protected Vector vrprice;
	/** Vector of vrproductivity objects. */
	protected Vector vrproductivity;
	/** Linear equation system. */
	protected LinEquSystem eqsys = null;
	/** A vector of equations that were not applied. */
	protected Vector vreqNotApplied = null;

	/** Creates a Exporter for writing model data to file.
	* @param szNameXMLFile The name of the XML model file.
	* @param szExtension File extension of exported file. */
	public Exporter(String szNameXMLFile, String szExtension) {
		this.szNameXMLFile = szNameXMLFile;
		this.szExtension = szExtension;
		chooser = new JFileChooser();
		chooser.addChoosableFileFilter(new ECFFileFilter(szExtension));
		chooser.setCurrentDirectory(new File("."));
		String szNameXML = (new File(szNameXMLFile)).getName();
		String szNameExport = ECFUtil.getFileNameWithoutExtension(szNameXML);
		chooser.setSelectedFile(new File(szNameExport.replace(' ','_')+"."+szExtension));
	} // Exporter

	/** Get the file name without extension.
	* @return The file name up to the last period character ".".
	*/
/*	private String getFileNameWithoutExtension(String szFullName) {
		System.out.println("Exporter.getFileNameWithoutExtension: szFullName=" + szFullName);
		int icharPeriod = szFullName.indexOf(".");
		String szShortName = szFullName.substring(0,icharPeriod);
		return szShortName;
	} // getFileNameNoExtension
*/
	/** Export data. */
	public void export() {
		openFileChooser();
	} // export

	/** Open the file chooser. */
	protected void openFileChooser() {
		String szFileName = "ECFModel." + szExtension;
		String szPath = System.getProperty("user.dir");

		int returnVal = chooser.showSaveDialog(ECF.frameMain);
		if(returnVal == JFileChooser.APPROVE_OPTION) {
			try {
				szFileName = chooser.getSelectedFile().getName();
				szPath = chooser.getSelectedFile().getAbsolutePath();
				FileWriter filewr = new FileWriter(szPath);
				filewr.write(listData());
				filewr.close();
			} catch(IOException excIO) {
				String szMsg = "Could not export to file  '" + szFileName +
					" '\n" + excIO.getMessage();
				JOptionPane.showMessageDialog(ECF.frameMain,szMsg);
			} catch(Exception exc) {
				String szMsg = "Error when exporting to file  '" + szFileName +
					" '\n" + exc;
				JOptionPane.showMessageDialog(ECF.frameMain,szMsg);
			} // try/catch
		} // if
	} // openFileChooser

	/** Returns the index offset for vectors (0 or 1),
	* should be replaced in subclass. */
	protected int indexOffset() { return 0;}

	/** Lists the model data for the specific file format. */
	protected String listData() {
		String szExport = "";
		// Heading
		szExport += listHeading();
		// Time histories
		szExport += listTimeHistories();
		// Equations that were not used
		szExport += listEquationsNotApplied();
		// Show equation system at selected time
		szExport += listAtSelectedTime();
		// Calculations using equation system
		szExport += listCalculationFormulas();
		return szExport;
	} // listData

	/** Calculation formulas, implemented for each output type. */
	public String listCalculationFormulas() {
		String szExport = "";
		return szExport;
	} // listCalculationFormulas

	/** List the heading.
	* @return A string with the heading.*/
	public String listHeading() {
		return listHeading(szCommentSign);
	} // listHeading

	/** List the heading.
	* @param szCommentSign The comment sign.
	* @return A string with the heading.*/
	public String listHeading(String szCommentSign) {
		Date date = new Date();
		DateFormat dateformat = DateFormat.getDateInstance(DateFormat.MEDIUM);
		// Heading
		String szExport = szCommentSign + "Results from the ECF program:       "
					+ dateformat.format(date) + ";\n"
					+ szCommentSign + "Model: " + szNameXMLFile;
		szExport += "\n" + szCommentSign + "Time histories;\n" +
			szCommentSign + "=======================================;\n\n";
		return szExport;
	} // listHeading

	/** List the time histories of variables.
	* @return A string listing time histories.*/
	public String listTimeHistories() {
		String szExport = "";
		// Time axis
		if (timeaxis!=null)	szExport += timeaxis.export();
		// Data vectors
		szExport += listFlowTimeHistories();
		szExport += listFlowRatioTimeHistories();
		szExport += listPriceTimeHistories();
		szExport += listProductivityTimeHistories();
		return szExport;
	} // listTimeHistories

	public String listAtSelectedTime() {
		return listAtSelectedTime(szCommentSign);
	} // listAtSelectedTime

	/** List equation systen at selected time.
	* @param szCommentSign The comment sign.
	* @return The equation system for the selected time.
	*/
	public String listAtSelectedTime(String szCommentSign) {

		// Selected time
		String szExport = "\n\n" + szCommentSign + "Model time: " +
			timeaxis.getSelectedTime() + ";\n" +
			szCommentSign + "=======================================;";

		// Flows
		szExport += "\n\n" + listFlows(szCommentSign,indexOffset());

		// Compact view of the equation system
		if (eqsys!=null)
			szExport += "\n\n" + eqsys.toStringCompact(szCommentSign,indexOffset());

		// Equation system
		if (eqsys!=null) szExport += "\n\n" + listEquationSystem();

		return szExport;
	} // listAtSelectedTime

	/** List flow time histories.
	* @return A string with semicolon(;) separates values and line feeds
	* between each flow line. */
	protected String listFlowTimeHistories() {
		// Flows as a function of time
		String szFlows = "";
		if (vrflow==null)	return szFlows;
		Iterator iter = vrflow.iterator();
		while (iter.hasNext()) {
			szFlows += ((Flow)iter.next()).export();
		} // while
		return szFlows;
	} // listFlowTimeHistories

	/** List flow ratio time histories.
	* @return A string with semicolon(;) separates values and line feeds
	* between each flow ratio line. */
	protected String listFlowRatioTimeHistories() {
		// Flows as a function of time
		String szFlowRatios = "";
		if (vrflowratio==null)	return szFlowRatios;
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			szFlowRatios += ((FlowRatio)iter.next()).export();
		} // while
		return szFlowRatios;
	} // listFlowRatioTimeHistories

	/** List price histories.
	* @return A string with semicolon(;) separates values and line feeds
	* between each price line. */
	protected String listPriceTimeHistories() {
		// Flows as a function of time
		String szPrices = "";
		if (vrprice==null)	return szPrices;
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			szPrices += ((Price)iter.next()).export();
		} // while
		return szPrices;
	} // listPriceTimeHistories

	/** List productivity histories.
	* @return A string with semicolon(;) separates values and line feeds
	* between each productivity line. */
	protected String listProductivityTimeHistories() {
		// Flows as a function of time
		String szProductivities = "";
		if (vrproductivity==null)	return szProductivities;
		Iterator iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			szProductivities += ((Productivity)iter.next()).export();
		} // while
		return szProductivities;
	} // listProductivityTimeHistories

	/** List flows at selected time.
	* @param szCommentSign The comment sign.
	* @param indexOffset Index offset for vectors (0 or 1).
	* @return A string with a list of the flows of the model.*/
	protected String listFlows(String szCommentSign, int indexOffset) {
		String szFlows = szCommentSign + "Flows:\n";
		Iterator iter = vrflow.iterator();
		while (iter.hasNext()) {
			Flow flow = (Flow)iter.next();
			int iflow = vrflow.indexOf(flow);
			szFlows += szCommentSign + ECFUtil.padCharacters(iflow+indexOffset,3,' ') + ": " + flow + "\n";
		} // while
		return szFlows;
	} // listFlows

	/** List flows with a given state at selected time.
	* @param szCommentSign The comment sign.
	* @param indexOffset Index offset for vectors (0 or 1).
	* @param ilayer Index of the layer of the flow.
	* @param state The state of the flow: Variable.DEFAULT,PRESCRIBED or CALCULATED.
	* @return A string with a list of the flows of the model.*/
	protected String listFlows(String szCommentSign, int indexOffset, int ilayer, int state) {
		String szFlows = szCommentSign + "Flows:\n";
		int cflowFound = 0;
		Iterator iter = vrflow.iterator();
		while (iter.hasNext()) {
			Flow flow = (Flow)iter.next();
			int iflow = vrflow.indexOf(flow);
			if (ilayer==flow.getLayerIndex() && state==flow.getState()) {
				szFlows += szCommentSign + ECFUtil.padCharacters(iflow+indexOffset,3,' ') +
					": " + flow + "\n";
				cflowFound++;
			} // if
		} // while
		if (cflowFound==0) szFlows += "No flows were " + Variable.STATES[state] + "\n";
		return szFlows;
	} // listFlows

	/** List flow ratios with a given state at selected time.
	* @param szCommentSign The comment sign.
	* @param indexOffset Index offset for vectors (0 or 1).
	* @param ilayer Index of the layer of the flow ratio.
	* @param state The state of the flow ratio: Variable.DEFAULT,PRESCRIBED or CALCULATED.
	* @return A string with a list of the flow ratios of the model.*/
	protected String listFlowRatios(String szCommentSign, int indexOffset, int ilayer, int state) {
		String szFlowRatios = szCommentSign + "FlowRatios:\n";
		int cflowratioFound = 0;
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			int iflowratio = vrflowratio.indexOf(flowratio);
			if (ilayer==flowratio.getLayerIndex() && state==flowratio.getState()) {
				szFlowRatios += szCommentSign + ECFUtil.padCharacters(iflowratio+indexOffset,3,' ') +
					": " + flowratio + "\n";
				cflowratioFound++;
			} // if
		} // while
		if (cflowratioFound==0) szFlowRatios += "No flow ratios were " + Variable.STATES[state] + "\n";
		return szFlowRatios;
	} // szFlowRatios

	/** List variables with a given state at selected time.
	* @param vrvariable A vector of variables to be listed.
	* @param szType1 Type of variable in words, e.g "FlowRatio".
	* @param szType2 Type of variable in words, e.g "flow ratio".
	* @param szCommentSign The comment sign.
	* @param indexOffset Index offset for vectors (0 or 1).
	* @param ilayer Index of the layer of the flow ratio.
	* @param state The state of the flow ratio: Variable.DEFAULT,PRESCRIBED or CALCULATED.
	* @return A string with a list of the flow ratios of the model.*/
	protected String listVariableAtTimeWithState(Vector vrvariable,
		String szType1, String szType2,
		String szCommentSign, int indexOffset, int ilayer, int state) {
		String szList = szCommentSign + szType1 + "s:\n";
		int cvariablesFound = 0;
		Iterator iter = vrvariable.iterator();
		while (iter.hasNext()) {
			Variable variable = (Variable)iter.next();
//			System.out.println("Exporter.listVariableAtTimeWithState: name= " +
//				 variable.getName() + " state=" + variable.getState());
			int ivariable = 1 + vrvariable.indexOf(variable);
			if (ilayer==variable.getLayerIndex() && state==variable.getState()) {
				szList += szCommentSign + ECFUtil.padCharacters(ivariable+indexOffset,3,' ') +
					": " + variable.toStringState(szType1) + "\n";
				cvariablesFound++;
			} // if
		} // while
		if (cvariablesFound==0) szList += "No "+ szType2 + "s were " + Variable.STATES[state] + "\n";
		return szList;
	} // szFlowRatios

	/** List equations that were not applied.
	* @return A string with flow equations.*/
	protected String listEquationsNotApplied() {
		return listEquationsNotApplied(szCommentSign);
	} // listEquationsNotApplied

	/** List equations that were not applied.
	* @return A string with flow equations.*/
	protected String listEquationsNotApplied(String szCommentSign) {

		if (vreqNotApplied==null || vreqNotApplied.size()==0)
			return "\n\n" + szCommentSign + "All equations were used.\n";
		String szEqu = "\n\n" + szCommentSign + "Equations that were not used:\n";
		Iterator iterEq = vreqNotApplied.iterator();
		while (iterEq.hasNext()) {
			LinearEquation eq = (LinearEquation)iterEq.next();
			String szNameLayer = ECF.getLayer(eq.getLayerIndex()).getName();
			szEqu += szCommentSign + eq.getType() + " of node " + eq.getNodeName() +
				" in layer " + szNameLayer + "\n";
			szEqu += szCommentSign + eq + "\n";
		} // while
		return szEqu;
	} // listEquationsNotApplied

	/** List constants and equation system.
	* @return A string with assignments for constants
	* and line feeds between each line.
	* To be implemented in subclasses. */
	protected String listEquationSystem() { return ""; }

	/** Set time history data.
	* @param timeaxis Time axis.
	* @param vrflow Vector of flow objects.
	* @param vrflowratio Vector of flow ratio objects.
	* @param vrprice Vector of price objects.
	*/
	public void setTimeHistories(TimeAxis timeaxis,
		Vector vrflow, Vector vrflowratio, Vector vrprice, Vector vrproductivity) {
		this.timeaxis = timeaxis;
		this.vrflow = vrflow;
		this.vrflowratio = vrflowratio;
		this.vrprice = vrprice;
		this.vrproductivity = vrproductivity;
	} // setTimeHistories

	/** Set equation system and equations that ver not applied.
	* @param eqsys Linear equation system.
	* @param vreqNotApplied A vector of equations that were not applied. */
	public void setEqSys(LinEquSystem eqsys, Vector vreqNotApplied) {
		this.eqsys = eqsys;
		this.vreqNotApplied = vreqNotApplied;
	} // setEqSys

} // Exporter