import java.util.*;
import javax.swing.*;     // JOptionPane

/**
* A <code>Flow</code> of resources may be monetary or real.
* The flow describes the kind of a resource and the quantity.<p>
* Prescribed and calculated flow rates are stored as vectors in
* <A href="Variable.html">Variables</A>.
* The <code>get</code> and <code>set</code> methods operate on the
* the Variable value that corresponds to the current time increment.
*/
public class Flow extends Variable {

	/** Index of flow. */
	protected int iflow = 0;
	/** Index of the layer to which the flow belongs. */
	protected int ilayer = 0;
	/** Index of flow for naming purpose. */
	static int iflowName = 1;
	/** Reset flow index iflowName. */
	public static void resetIndex() { iflowName = 1; }
	/** Vector of flowpaths that reference this flow. */
	protected Vector vrflowpath = new Vector();
	/** Flag = false if calculated flows shall be shown, = true if
	* flows differences shall be shown.. */
	public static boolean fShowDiff = false;
	/** Flag = true if flow numbers shall be shown, = false otherwise */
	public static boolean fShowFlowNumbers = false;

	/** Set if flow diagram shall show calculated flows or
	* the difference between calculated and measured flows.
	* Toggles between the two states. */
	public static void showCalcMeasFlowDifference() {
		Flow.fShowDiff = !fShowDiff;
	} // showCalcMeasFlowDifference

	/** Set if flow diagram shall show calculated flows or
	* the difference between calculated and measured flows.
	* @param fShowDiff = false is calculated flows shall be shown, = true if
	* flows differences shall be shown. */
	public static void showCalcMeasFlowDifference(boolean fShowDiff) {
		Flow.fShowDiff = fShowDiff;
	} // showCalcMeasFlowDifference

	/** Set if flow diagram shall show flow numbers.
	* Toggles between the two states. */
	public static void showFlowNumbers() {
		Flow.fShowFlowNumbers = !fShowFlowNumbers;
	} // showCalcMeasFlowDifference

	/** Create a flow of resources with ilayer = 0.
	* @param timeaxis The time axis against which the Flow is plotted.
	*/
	public Flow() {
		super();
		ilayer = 0;
		setDefaultUnits();
		szName = "Flow"+iflowName;
		iflowName++;
	} // Flow

	/** Create a flow of resources.
	* @param ilayer Index of the layer to which the flow belongs.
	*/
	public Flow(int ilayer) {
		super();
		this.ilayer = ilayer;
		setDefaultUnits();
		szName = "Flow"+iflowName;
		iflowName++;
	} // Flow

	/** Create a flow of resources.
	* @param ilayer Index of the layer to which the flow belongs.
	* @param szName Name of flow
	*/
	public Flow(int ilayer, String szName) {
		super();
		this.ilayer = ilayer;
		this.szName = szName;
		setDefaultUnits();
	} // Flow

	/** Add reference to flow path.
	* @param flowpath Flow path to be added. */
	public void add(FlowPath flowpath) {
		vrflowpath.add(flowpath);
	} // add

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		if (this.ilayer==ilayer) {
			super.clearUnits();
		} // if
	} // clearUnits

	/**
	* Decode flow from XML code.
	* @param szXMLCode The xml description of this model.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("ilayer")) ilayer = Integer.parseInt(szSub);
				if (szTag.equals("Variable")) super.decodeXML(szSub);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
		} // try/catch
		findAndSetUnits();
	} // decodeXML

	/**
	* Find a <A href="Units.html">Units</A> set from a vector vrunits that matches
	* szNameUnits and set the units of the flow to that instance.
	* @param vrunits A Vector of Units.
	*/
	public void findAndSetUnits() {
		Vector vrunits = ECF.vrunits;
		if (vrunits==null) return;
		Iterator iter = vrunits.iterator();
		while (iter.hasNext()) {
			Units units = (Units)iter.next();
			if (units==null) {
				System.out.println("Flow.findAndSetUnits: Can not find units.");
				JOptionPane.showMessageDialog(ECF.frameMain,
	    		"Flow.findAndSetUnits: Can not find units.",
	    		"ECF find units",
	    		JOptionPane.WARNING_MESSAGE);
	    	} else if (szNameUnits.compareTo(units.getName())==0)	setUnits(units);
		} // while
	} // findAndSetUnits

	/**
	* Magnitude of prescribed flow expressed in display units.
	* @return Flow rate in current units.
	*/
	public double getFlowPrescribed() {
		Units units = getUnits();
		double ratePrescribed = getBaseValuePrescribed();
		if (units!=null)  {
			return ratePrescribed*units.getFactor();
		} else  {
			return ratePrescribed;
		} // if/else
	} // getFlowPrescribed

	/**
	* Magnitude of prescribed flow as base value.
	* @return Flow rate as base value.
	*/
	public double getFlowPrescribedBase() {
		return super.getBaseValuePrescribed(); }

	/** Get index of flow.
	* @return Index of flow. */
	public int getIndexOfFlow() { return ECF.getIndexOfFlow(this); }
//	public int getIndexOfFlow() { return iflow; }

	/** Get index of layer for this flow.
	* @return Index of layer. */
	public int getLayerIndex() { return ilayer; }

	/** Get the name of the flow. */
	public String getNameForLabel() {
		String szFlowNameLabel = "";
		if (fShowFlowNumbers) szFlowNameLabel += "No." + (1+ECF.getIndexOfFlow(this)) + " ";
		if (vrflowpath.size()<=1) szFlowNameLabel += szName;
		else szFlowNameLabel += szName + "(" + vrflowpath.size() + ")";
		return szFlowNameLabel;
	} // getName

	/** Get the pixel value of the Flow.
	* @param cpixel The maximum width of an arrow symbol when the flow
	* has max display value. */
	public int getPixelValue(int cpixel) {
		if (fShowDiff) return getPixelValueCalcMeasDiff(cpixel);
		else if (isPrescribed()) return getPixelValuePrescribed(cpixel);
		else return getPixelValueCalculated(cpixel);
	} // getPixelValue

	/** Get the scaling factor for setting flows in equation systems.
	* Flow in calculations = Base flow / scale.
	* Base flow = Flow in calculations * scale.
	* @return Scaling factor for flows.*/
	public double getScaleForCalc() {
		Units units = getUnits();
		if (units==null) return 1.0;
		return units.getScaleForCalc();
	} // getScaleForCalc

	/** Get the string with flow value and display unit.
	* @return A string with decimal number and unit, e.g. "1.3 MCU"
	*/
	public String getStringQuant() {
		String szFlowQuantLabel = "";
		if (fShowDiff) szFlowQuantLabel += getStringQuantCalcMeasDiff();
		else if (isPrescribed()) szFlowQuantLabel += getStringQuantPrescribed();
		else szFlowQuantLabel += getStringQuantCalculated();
		return szFlowQuantLabel;
	} // getStringQuant

	public boolean hasMoreFlowPaths() { return vrflowpath.size()>1;}

	/** List XML code for the flow.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += "<ilayer>"+ilayer+"</ilayer>";
		szXML += "<Variable>"+ super.listXML("") +"</Variable>";
		return szXML;
	} // listXML

	/** Remove refernce to flow path.
	* @param flowpath Flow path to be removed. */
	public void remove(FlowPath flowpath) { vrflowpath.remove(flowpath); }

	/** Set defalt <A href="Units.html">Units</A> of a new flow.
	* The defalult is the first occurence
	* of a units set that is associated with the layer.
	* @param vrunits A vector of defined units sets.*/
	public void setDefaultUnits() {
		Units units = null;
		Iterator iter = ECF.vrunits.iterator();
		while (iter.hasNext()) {
			units = (Units)iter.next();
			if (units.getLayerIndex()==ilayer) {
				setUnits(units);
				break;
			} // if
		} // while
	} // setDefaultUnits

	/** Set default <A href="Units.html">Units</A> of a new flow. The default is the first occurence
	of a units set that is associated with the layer.
	* @param vrunits A vector of defined Units sets.*/
	public void setDefaultUnits(Vector vrunits) {
		if (vrunits!=null) {
			Iterator iter = vrunits.iterator();
			while (iter.hasNext()) {
				Units units = (Units)iter.next();
				if (units.getLayerIndex()==ilayer) {
					setUnits(units);
					break;
				} // if
			} // while
		} // if
	} // setDefaultUnits

	/**
	* Set magnitude of calculated flow, base value.
	* @param rate Flow magnitude as base value.
	*/
	public void setFlowCalculated(double rate) {
		setBaseValueCalculated(rate);
	} // setFlowCalculated

	/** Set prescribed flow expressed in display units and flow state to PRESCRIBED.
	* @param rate Prescribed flow magnitude expressed in current units.*/
	public void setFlowPrescribed(double rate) {
		Units units = getUnits();
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		setBaseValuePrescribed(rate/factor);
		setPrescribed();
		Calculator.calculate();
	} // setFlowPrescribed

	/** Set index of the Flow.
	* @param iflow Index of flow. */
	public void setIndexOfFlow(int iflow) { this.iflow = iflow; }

	/** Set layer index.
	* @param ilayer Index of the layer to which the flow belongs. */
	public void setLayerIndex(int ilayer) { this.ilayer = ilayer; }

	/** Show properties of the Flow
	* @return A string telling the properties.*/
	public String toString() {
		int cflowpath = vrflowpath.size();
		String szNameNoLineFeeds = ECFUtil.removeLineFeeds(szName);
		if (cflowpath>1) szNameNoLineFeeds += "("+cflowpath+ ")";
		String sz = "\tFlow: ";
		sz += Quantity.padToLength(szNameNoLineFeeds,25) + " ";
		if (super.state==DEFAULT)    sz += "DEFAULT=0  ";
		if (super.state==CALCULATED) sz += "CALCULATED= " + getStringQuantCalculated();
		if (super.state==PRESCRIBED) sz += "PRESCRIBED= " + getStringQuantPrescribed();
		return sz;
	} // toString

} // Flow