import java.awt.*;
import java.awt.geom.*;

/**
* A <code>FlowConnector</code> is attached to one end of a
* <A href="FlowPath.html">FlowPath</A> line. It can be dragged to
* connect to nodes. The flowpath line is dragged together with the
* connector.The connection is done from the node side
* via <A href="NodeConnectors.html">NodeConnectors</A>.
* <p>
* The <code>FlowConnector</code> class has only metods for
* disconnection flows: <br>
* - <code>disconnectNode()</code> disconnects a node if connected via the
* <code>undoAssociation()</code> method of the
* <A href="NodeConnectors.html">NodeConnectors</A> class. <br>
* - <code>translate(dx,dy)</code> disconnects this FlowConnector from the node,
* via the <a href="#disconnectNode()">disconnectNode()</a> method.
* <p>
* Flow connectors are drawn as white squares at the head end of flow paths
* and as grey squares at the tail end: <br>
* <img src="FlowConnectorsAndNodes.gif">
* <p>
* The flow connector has a reference point (x,y) where arrow head of
* the flow path is attached to the sector. The offset is a vector from
* the upper left corner of the connector square to the reference point.<br>
* <img src="FlowConnectorOffset.gif">
*/
public class FlowConnector extends Rectangle {

	/**  Flag = true if FlowConnectors are shown.*/
	protected static boolean fShowConnectors = true;
	/** Toggle if conectors are shown or hidden. */
	public static void hideshow() {
		if (fShowConnectors) fShowConnectors = false; else fShowConnectors = true;
	} // hideshow

	/** The flowpath to which the connector is attached. */
	protected FlowPath flowpath;
	/** Index of end of flowpath line (1=FlowPath.END_HEAD,2=FlowPath.END_TAIL) */
	protected int iend;
	/** Side of the node to which the the connector is connected. */
	protected int sideNode = ECFNode.NONE;
	/** Flag =true if connector is hidden, else =false. */
	protected boolean fHidden = false;

	Painter painter;  // The painter of the connector symbol.
	int xStart = 0;   // Start x-position before dragging.
	int yStart = 0;   // Start y-position before dragging.
	/** Side length of connector symbol. */
	public static int SIDE = 10;
	ECFNode node = null; // Node to which this is connected
	NodeConnectors nodeconn= null; // NodeConnectors to which this is connected

	/**
	* Creates a connector that is attached to the end <code>iend</code>
	* of a flowpath.
	* @param flowpath The flowpath.
	* @param iend index of the end of the flowpath.
	* @param pnt the reference point (x,y) of the connector =
	* the coordinates of the corresponding end of the flowpath.
	*/
	public FlowConnector(FlowPath flowpath, int iend, Point pnt) {
		super(pnt);
		this.flowpath = flowpath;
		this.iend =iend;
		makeGeom();
		setLocation((int)(getX()-offset().getX()),
				(int)(getY()-offset().getY()));
	} // FlowConnector

	/** Check if connector symbol contains the point (x,y).
	* Can not be found if not shown.
	* @param x x-coordinate of point.
	* @param y y-coordinate of point.
	* @return true if point shown and contained within symbol, else false.
	*/
	public boolean contains(int x, int y) {
		if (!fShowConnectors || fHidden) return false;
		return contains((double)x,(double)y);
	} // contains

	/** Disconnect the connector from node. */
	public void disconnectNode() {
		// Disconnect this connector before deleting flow line
		if (nodeconn!=null) nodeconn.undoAssociation(this);
		nodeconn= null;
	} // disconnect

	/** Get index of end of flow.
	* @return 1 if arrow end, 2 if base end */
	public int getEndOfFlow() { return iend; }

	/** Get flowpath.
	* @return FlowPath. */
	public FlowPath getFlowPath() { return flowpath; }

	/** Get the reference point of the connector.
	* @return reference point (x,y).
	*/
	public Point2D getRefPoint() {
		double x = getX() + offset().getX();
		double y = getY() + offset().getY();
		return new Point2D.Double(x,y);
	} // getPos

	/** Get side of node to which connector is connected.
	* @return side of node.*/
	public int getSide() {return sideNode;}

	/** Hide the connector permanently. */
	public void hidePermanently() { fHidden = true; }

	/** Check if connector is at head (destination) end of flow path.
	* @return true if head end, else false*/
	public boolean isHeadEnd() {return iend==1;}

	/** Check if connector is at tail (source) end of flow path.
	* @return true if tail end, else false*/
	public boolean isTailEnd() {return iend==2;}

	/** List XML code for this flowconnector.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += "<iend>"+iend+"</iend>";
		if (flowpath instanceof FlowPathOfFrame) {
			FlowPathOfFrame flowpathofframe = (FlowPathOfFrame)flowpath;
			szXML += "<FlowPathOfFrame>";
			szXML += flowpathofframe.listXML("");
		 	szXML += "</FlowPathOfFrame>";
		} else {
			szXML += "<flowpath>"+flowpath.getFlowPathIndex()+"</flowpath>";
		} // if/else
		return szXML;
	} // listXML

	/**
	* Make the geometry of the connector symbol and create painter.
	*/
	private void makeGeom() {
		Dimension dim = new Dimension(SIDE,SIDE);
		Painter painterFill;
		setSize(dim);
		if (iend==1)
			painterFill = new FillPainter(Color.white);
		else
			painterFill = new FillPainter(Color.gray);
		Stroke stroke = new BasicStroke(1);
		Painter painterDraw = new DrawPainter(Color.black,stroke);
		painter = new MultiPainter(painterFill,painterDraw);
	} // makeGeom

	/** Get offset of the reference point from the upper left corner
	* of connector symbol.
	* @return A 2D vector (x,y) from the upper left corner of the connector
	* to the reference point. */
	public Point2D offset() {
		if (sideNode==ECFNode.WEST)
			return new Point2D.Double(SIDE,SIDE/2);
		else if (sideNode==ECFNode.EAST)
			return new Point2D.Double(0,SIDE/2);
		else if (sideNode==ECFNode.NORTH)
			return new Point2D.Double(SIDE/2,SIDE);
		else if (sideNode==ECFNode.SOUTH)
			return new Point2D.Double(SIDE/2,0);
		else
			return new Point2D.Double(SIDE/2,SIDE/2);
	} // offset

	/** Paint the connector symbol. Do not paint if not shown.
	* @param g the graphics context.
	*/
	public void paint(Graphics g) {
		if (!fShowConnectors || fHidden) return;
		Graphics2D g2 = (Graphics2D)g;
		painter.paint(g2, this);
	} // paint

	/** Paint the connector symbol. Do not paint if not shown.
	* @param g2 the graphics context.
	* @param fIsPrescribed Flag = true if price is prescribed, else = false.
	*/
	public void paintPriceMarker(Graphics2D g2, boolean fIsPrescribed) {
		// Coordinates of reference point
		Point2D pntRef = getRefPoint();
		int x = (int)pntRef.getX();
		int y = (int)pntRef.getY();
		int dx,dy; // offset for price marker upper left corner
		int dxP = 1; int dyP= 9; // extra offset for string "P"
		if (sideNode==ECFNode.WEST) {
			dx = 0; dy = -SIDE/2; }
		else if (sideNode==ECFNode.EAST) {
			dx = -SIDE; dy = -SIDE/2; }
		else if (sideNode==ECFNode.NORTH) {
			dx = -SIDE/2; dy = 0; }
		else if (sideNode==ECFNode.SOUTH) {
			dx = -SIDE/2; dy = -SIDE; }
		else {
			dx = 0; dy = 0; }
		// Draw a square and the letter "P" inside.
		if (fIsPrescribed)
			g2.setColor(ECFModel.COLOR_PRESCRIBED);
		else
			g2.setColor(ECFModel.COLOR_CALCULATED);
		g2.fillRect(x+dx,y+dy,SIDE,SIDE);
		g2.setColor(Color.black);
		g2.drawString("P",x+dx+dxP,y+dy+dyP);
	} // paint

	/** Set association to the <code>NodeConnectors</code> of the node connecting to this.
	* Shall only be called by the <code>connectFlowPath</code> method of the
	* <code>NodeConnectors</code> class.
	* @param nodeconn the nodeconnectors of the node to which this is being connected. */
	public void setAssociation(NodeConnectors nodeconn) {this.nodeconn = nodeconn;}

	/** Set the position of this connector and update flowline.
	* @param x the x-coordinate of the reference point.
	* @param y the y-coordinate of the reference point.
	*/
	public void setPos(int x, int y) {
		setLocation(x-(int)offset().getX(), y-(int)offset().getY());
		flowpath.updateLine();
	} // setPos

	/** Set the position of this connector and do not update flowline.
	* @param x the x-coordinate of the reference point.
	* @param y the y-coordinate of the reference point.
	*/
	public void setPosOnly(int x, int y) {
		setLocation(x-(int)offset().getX(), y-(int)offset().getY());
	} // setPos

	/** Set side of node to which the flow connector is connected and update flowline.
	* @param sideNode The side of the node to which this node is connected,
	* see <A href="ECFNode.html#field_summary">ECFNode Field Summary</A>.
	*/
	public void setSide(int sideNode) {
		if (sideNode>=ECFNode.NONE && sideNode<=ECFNode.SOUTH)
			this.sideNode = sideNode;
		else
			this.sideNode = ECFNode.NONE;
		flowpath.updateLine();
	} // setSide

	/** Set side of node to which this is connected and update flowline.
	* @param sideNode The side of the node to which this node is connected,
	* see <A href="ECFNode.html">ECFNode</A>.
	*/
	public void setSideOnly(int sideNode) {
		// See offset();
		if (sideNode>=ECFNode.NONE && sideNode<=ECFNode.SOUTH)
			this.sideNode = sideNode;
		else
			this.sideNode = ECFNode.NONE;
	} // setSide

	/**
	* Set start position of the connector before dragging.
	*/
	public void setStartPos() {
		xStart = (int)getX();
		yStart = (int)getY();
	} // setStartPos

	/**
	* Return a description of the connector.
	* @return A description of the connector.
	*/
	public String toString() {
		String szSide;
		if (sideNode==ECFNode.WEST) szSide="WEST";
		else if (sideNode==ECFNode.EAST) szSide="EAST";
		else if (sideNode==ECFNode.NORTH) szSide="NORTH";
		else if (sideNode==ECFNode.SOUTH) szSide="SOUTH";
		else szSide="NONE";
		return "FlowConnector end=" + iend + " side=" + szSide;
	} // toString

	/**
	* Make XML description of the connector.
	* @param szBlanks A blank string inserted at the beginning of each text row.
	* @return A string with line feeds separating text rows of the description.
	*/
	public String toXML(String szBlanks) {
		String sz = "\n" + szBlanks + "<Connector>";
			  sz += "\n" + szBlanks + "   <End>" + iend + "</End>";
			  // sz += "\n" + szBlanks + "   <FlowPath>" + flowpath.findIndexOfFlowPath() +"</FlowPath>";
			  sz += "\n" + szBlanks + "</Connector>";
		return sz;
	} // toXML

	/**
	* Translate the connector symbol by (dx,dy), disconnect node
	* if connected.
	* @param dx translation in x direction.
	* @param dy translation in y direction.
	*/
	public void translate(int dx, int dy) {
		disconnectNode();
		sideNode = ECFNode.NONE;
		setLocation(xStart+dx, yStart+dy);
	} // moveBy

	/** Second action when disconnecting a flowpath from a node. <br>
	* Shall only be called from the <code>disconnectOnOneSide()</code> method of
	* the <code>NodeConnectors</code> class.*/
	public void undoAssociation() {
		sideNode = ECFNode.NONE;
		nodeconn = null;
	} // undoAssociation

} // FlowConnector