import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.util.*;

/**
* A <code>FlowPath</code> connects two nodes. Resources flow in the direction
* indicated by the arrow. Resources may be monetary or real. Many flows may
* be associated to one flowpath. The flows describe the kind of resource and
* the quantity.
* <p>
* Each flowpath belongs to one model layer. It has two end points with connectors
* that can be dragged by the mouse. They can be connected to a node if dragged
* over that node and then released. The connection is done from the node side
* via <A href="NodeConnectors.html">NodeConnectors</A>.
* <p>
* The <code>FlowPath</code> class has only metods for disconnection flows: <br>
* - <code>disconnectThis()</code> disconnects both ends (before deleting the flow path) <br>
* - <code>translate(dx,dy)</code> disconnects the translated end via the
* <A href="FlowConnector.html">FlowConnector</A> class.
*/
public class FlowPath {

	/** The head end (with arrow) of a flow path. */
	public static int END_HEAD = 1;
	/** The tail end (with no arrow) of a flow path. */
	public static int END_TAIL = 2;

	/** Connector at the head end of the flowpath. */
	protected FlowConnector conn1 = null;
	/** Connector at the tail end of the flowpath. */
	protected FlowConnector conn2 = null;
	/** Flow of resources. */
	protected Flow flow = null;

	/** Selected connector, if found, else <code>null</code>.*/
	protected FlowConnector connSelected = null;
	/** Indicate if label is selected. */
	protected boolean fLabelSelected = false;
	/** Indicate if label is shown. */
	protected boolean fLabelShown = true;
	/** Index of the layer to which the flowpath belongs. */
	protected int ilayer;
	/** The model/modelframe to which the flow path belongs. */
	protected FlowPathHolder model = null;
	/** Type of side to which the flowpath is connected on node at the head end,
	* see <code>ECFNode</code>. */
	protected int sideNode1 = ECFNode.NONE;
	/** Type of side to which the flowpath is connected on node at the tail end,
	* see <code>ECFNode</code>. */
	protected int sideNode2 = ECFNode.NONE;
	/** The name of the flowpath. */
//	protected String szName = "Flow";
	/** X-coordinate of the head end.*/
	protected int x1;
	/** X-coordinate of the head end.*/
	protected int x2;
	/** X-coordinate of center of label. */
	protected int xLabel = 0;
	/** Offset in x direction of label, relative to middle point */
	protected int xOffsetLabel = 0;
	/** X-coordinate of label offset before translating. */
	protected int xStart;
	/** Y-coordinate of the tail end.*/
	protected int y1;
	/** Y-coordinate of the tail end.*/
	protected int y2;
	/** Y-coordinate of center of label. */
	protected int yLabel = 0;
	/** Offset in y direction of label, relative to middle point */
	protected int yOffsetLabel = 0;
	/** Y-coordinate of label offset before translating. */
	protected int yStart;
	/** Label of the this flowpath. */
	private GraphLabel label = null;
	/** Number of pixels for max error vidth. */
	protected int cpixelMaxArrowWidth = 15;

	/**
	* Creates a flowpath in layer <code>ilayer</code> from point
	* <code>(x2,y2)</code> to point <code>(x1,y1)</code> containing
	* flow.
	* @param model The parent model of the flow path.
	* @param ilayer Index of the layer to which the flow path belongs.
	* @param x1,y1 Coordinates of the head end.
	* @param x2,y2 Coordinates of the tail end.
	* @param flow  The flow of the flow path.
	*/
	public FlowPath(FlowPathHolder model, int ilayer,int x1,int y1,int x2,int y2, Flow flow) {
		this.model = model;
		this.ilayer = ilayer;
		this.x1 = x1;
		this.y1 = y1;
		this.x2 = x2;
		this.y2 = y2;
		this.flow = flow;
		label = new GraphLabel("Flow");
		setSide1(ECFNode.NONE);
		setSide2(ECFNode.NONE);
		conn1 = new FlowConnector(this,1,new Point(x1,y1));
		conn2 = new FlowConnector(this,2,new Point(x2,y2));
	} // FlowPath

	/**
	* Creates a flowpath in layer <code>ilayer</code> from point
	* <code>(x2,y2)</code> to point <code>(x1,y1)</code>.
	* @param model The parent model of the flow path.
	* @param ilayer Index of the layer to which the flow path belongs.
	* @param x1,y1 Coordinates of the head end.
	* @param x2,y2 Coordinates of the tail end.
	*/
	public FlowPath(FlowPathHolder model, int ilayer,int x1,int y1,int x2,int y2) {
		this.model = model;
		this.ilayer = ilayer;
		this.x1 = x1;
		this.y1 = y1;
		this.x2 = x2;
		this.y2 = y2;
		flow = new Flow(ilayer);
		flow.add(this);
		ECF.addFlow(flow);
		// Calculator.addFlow(flow);
		label = new GraphLabel("Flow");
		setSide1(ECFNode.NONE);
		setSide2(ECFNode.NONE);
		conn1 = new FlowConnector(this,1,new Point(x1,y1));
		conn2 = new FlowConnector(this,2,new Point(x2,y2));
	} // FlowPath

	/**
	* Creates a vertical flowpath of length 100 in layer <code>ilayer</code>
	* from point <code>(x1,y1+100)</code> to point <code>(x1,y1)</code>.
	* @param model The parent model of the flow path.
	* @param ilayer Index of the layer to which the flow path belongs.
	* @param x1,y1 Coordinates of the head end.
	*/
	public FlowPath(FlowPathHolder model, int ilayer,int x1,int y1) {
		this(model,ilayer,x1,y1,x1,y1+100);
	} // FlowPath

	/**
	* Creates a vertical flowpath of length 100 in layer <code>ilayer</code>
	* from point <code>(x1,y1+100)</code> to point <code>(x1,y1)</code>.
	* @param model The parent model of the flow path.
	* @param ilayer Index of the layer to which the flow path belongs.
	* @param x1,y1 Coordinates of the head end.
	* @param flow the flow to be put into the flow path.
	*
	*/
	public FlowPath(FlowPathHolder model, int ilayer,int x1,int y1, Flow flow) {
		this(model,ilayer,x1,y1,x1,y1+100,flow);
} // FlowPath

	/**
	* Creates a vertical flowpath of length 100 in layer 0
	* from point <code>(10,110)</code> to point <code>(10,10)</code>.
	* @param model The parent model of the flow path.
	* @param flow the flow to be put into the flow path.
	*/
	public FlowPath(FlowPathHolder model, Flow flow) {
		this(model,0,10,10,10,110,flow);
	} // FlowPath

	/**
	* Creates a vertical flowpath of length 100 in layer 0
	* from point <code>(10,110)</code> to point <code>(10,10)</code>.
	* @param model The parent model of the flow path.
	*/
	public FlowPath(FlowPathHolder model) {
		this(model,0,10,10,10,110);
	} // FlowPath

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		if (this.ilayer==ilayer) flow.clearUnits(ilayer);
	} // clearUnits

	/**
	* Checks if the <code>FlowPath</code> belongs to layer <code>ilayer</code>
	* and contains the coordinates (x,y) in either label or connectors.
	* Sets label as selected or one of the end connectors.
	* @return true if contained, false otherwise.
	*/
	public boolean contains(int ilayer,int x,int y) {
		if (this.ilayer==ilayer)  {
			if (fLabelShown) {
				fLabelSelected = label.contains(x,y);
				if (fLabelSelected)  {
			     	connSelected = null;
					return true;
				} // if fLabelSelected
			} // if fLabelShown
			if (conn1.contains(x,y)) {
		     	connSelected = conn1;
				return true;
			}
			if (conn2.contains(x,y)) {
		     	connSelected = conn2;
				return true;
			}
		}
		return false;
	} // contains

	/**
	* Decode flowpath from XML code.
	* @param szXMLCode The xml description of this model.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("ilayer")) ilayer = Integer.parseInt(szSub);
				if (szTag.equals("x1")) x1 = Integer.parseInt(szSub);
				if (szTag.equals("y1")) y1 = Integer.parseInt(szSub);
				if (szTag.equals("x2")) x2 = Integer.parseInt(szSub);
				if (szTag.equals("y2")) y2 = Integer.parseInt(szSub);
				if (szTag.equals("xOffsetLabel")) xOffsetLabel = Integer.parseInt(szSub);
				if (szTag.equals("yOffsetLabel")) yOffsetLabel = Integer.parseInt(szSub);
				if (szTag.equals("sideNode1")) sideNode1 = Integer.parseInt(szSub);
				if (szTag.equals("sideNode2")) sideNode2 = Integer.parseInt(szSub);
				if (szTag.equals("flowindex")) {
					int iflow = Integer.parseInt(szSub);
					Flow flowTemp = ECF.getFlow(iflow);
					if (flowTemp!=null) { flow = flowTemp; flow.add(this); }
					else {
						String szMsg = "Flowpath.decodeXML could not find flow " +
						"with index=" + iflow;
						JOptionPane.showMessageDialog(ECF.frameMain,szMsg);
					} // if/else
				}// if flowindex
				szXMLCode = xml.remainder();
			} // while
			conn1.setSideOnly(sideNode1);
			conn1.setPosOnly(x1,y1);
			conn2.setSideOnly(sideNode2);
			conn2.setPosOnly(x2,y2);
		} catch(Exception e) {
			System.out.println("FlowPath.decodeXML: " + e);
		} // try/catch
	} // decodeXML

	/**
	* Decode only label position from flowpath from XML code.
	* @param szXMLCode The xml description of this model.
	*/
	public void decodeXMLLabelPos(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("xOffsetLabel")) xOffsetLabel = Integer.parseInt(szSub);
				if (szTag.equals("yOffsetLabel")) yOffsetLabel = Integer.parseInt(szSub);
				szXMLCode = xml.remainder();
			} // while
			conn1.setSideOnly(sideNode1);
			conn1.setPosOnly(x1,y1);
			conn2.setSideOnly(sideNode2);
			conn2.setPosOnly(x2,y2);
		} catch(Exception e) {
			System.out.println("FlowPath.decodeXML: " + e);
		} // try/catch
	} // decodeXMLLabelPos

	/** Disconect both ends from nodes if connected.*/
	public void disconnectThis() {
		conn1.disconnectNode();
		conn2.disconnectNode();
	} // disconnectThis

	/** Export FlowPath values to file *.dat
	* @return A string with rows of label and values on each line.
	*/
	public String export() {
		return flow.export();
	} // export

	/** Get the flow of the flow path.
	* @return The Flow of the flow path.*/
	public Flow getFlow() { return flow; }

	/** Get the index of the flow path in the parent model. */
	public int getFlowPathIndex() {
		if (model!=null)
			return model.getFlowPathIndex(this);
		else
			// ModelFrame flowpaths (TwinFlowConnector).
			return 88;
		} // getFlowPathIndex

	/** Hide flow connector permanently at iend. */
	public void hideConnectorPermanently(int iend) {
		if (iend==END_HEAD) conn1.hidePermanently();
		else if (iend==END_TAIL) conn2.hidePermanently();
	} // hideConnectorPermanently

	/** Hide flow label permanently. */
	public void hideLabelPermanently() {
		fLabelShown = false;
	} // hideConnectorPermanently


	/** Check if label was hit.
	* @return true if label was hit, else false. */
	public boolean hitLabel() {return fLabelSelected;}

	/** List XML code for this flowpath.
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = "";
		szXML += "<ilayer>"+ilayer+"</ilayer>";
		szXML += "<x1>"+x1+"</x1>";
		szXML += "<y1>"+y1+"</y1>";
		szXML += "<x2>"+x2+"</x2>";
		szXML += "<y2>"+y2+"</y2>";
		szXML += "<xOffsetLabel>"+xOffsetLabel+"</xOffsetLabel>";
		szXML += "<yOffsetLabel>"+yOffsetLabel+"</yOffsetLabel>";
		szXML += "<sideNode1>"+sideNode1+"</sideNode1>";
		szXML += "<sideNode2>"+sideNode2+"</sideNode2>";
		szXML += "<flowindex>"+ECF.getIndexOfFlow(flow)+"</flowindex>";
		return szXML;
	} // listXML

	/** List XML code for this flowpath.
	* @return XML code. */
	public String listXMLLabelPos(String szBlanks) {
		String szXML = "";
		szXML += "<xOffsetLabel>"+xOffsetLabel+"</xOffsetLabel>";
		szXML += "<yOffsetLabel>"+yOffsetLabel+"</yOffsetLabel>";
		return szXML;
	} // listXMLLabelPos

	/**
	* Draws a line from end (x2,y2) to end (x1,y1) in graphics context g, <br>
	* the line may be strait or broken into horizontal and vertical sections
	* that connect in strait angle to the sides of the connected nodes.
	* @param g the graphics context.
	*/
	private void drawLine(Graphics g) {
		int xm, ym;
		int xm1, ym1;
		int xm2, ym2;
		// Offset of middle line of a U-shaped flowpath.
		int offsetULine = 40;

		if (sideNode1==ECFNode.NONE  || sideNode2==ECFNode.NONE) {
			g.drawLine(x1,y1,x2,y2); // Strait line
			xLabel = (x1+x2)/2; yLabel = (y1+y2)/2;
		}
		else if ((sideNode1==ECFNode.EAST  || sideNode1==ECFNode.WEST) &&
			(sideNode2==ECFNode.NORTH  || sideNode2==ECFNode.SOUTH)) {
			xm = x2; ym = y1;        // L-line starting horizontal
			g.drawLine(x1,y1,xm,ym); // Horizontal line
			g.drawLine(xm,ym,x2,y2); // Vertical line
			xLabel = x2; yLabel = y1;
		}
		else if ((sideNode1==ECFNode.NORTH  || sideNode1==ECFNode.SOUTH) &&
			(sideNode2==ECFNode.EAST  || sideNode2==ECFNode.WEST)) {
			xm = x1; ym = y2;   // L-line starting vertical
			g.drawLine(x1,y1,xm,ym); // Vertical line
			g.drawLine(xm,ym,x2,y2); // Horizontal line
			xLabel = x1; yLabel = y2;
		}
		else if (sideNode1==ECFNode.EAST  && sideNode2==ECFNode.EAST) {
			xm = Math.max(x1,x2) + offsetULine; // U-line east starting horizontal
			xm1 = xm+xOffsetLabel; ym1 = y1;
			xm2 = xm+xOffsetLabel; ym2 = y2;
			g.drawLine(x1,y1,xm1,ym1);   // Horizontal line
			g.drawLine(xm1,ym1,xm2,ym2); // Vertical line
			g.drawLine(xm2,ym2,x2,y2);   // Horizontal line
			xLabel = xm; yLabel = (y1+y2)/2;
		}
		else if (sideNode1==ECFNode.WEST  && sideNode2==ECFNode.WEST) {
			xm = Math.min(x1,x2) - offsetULine;  // U-line west starting horizontal
			xm1 = xm+xOffsetLabel; ym1 = y1;
			xm2 = xm+xOffsetLabel; ym2 = y2;
			g.drawLine(x1,y1,xm1,ym1);   // Horizontal line
			g.drawLine(xm1,ym1,xm2,ym2); // Vertical line
			g.drawLine(xm2,ym2,x2,y2);   // Horizontal line
			xLabel = xm; yLabel = (y1+y2)/2;
		}
		else if (sideNode1==ECFNode.NORTH  && sideNode2==ECFNode.NORTH) {
			ym = Math.min(y1,y2) - offsetULine;  // U-line north starting vertical
			xm1 = x1; ym1 = ym+yOffsetLabel;
			xm2 = x2; ym2 = ym+yOffsetLabel;
			g.drawLine(x1,y1,xm1,ym1);   // Vertical line
			g.drawLine(xm1,ym1,xm2,ym2); // Horizontal line
			g.drawLine(xm2,ym2,x2,y2);   // Vertical line
			xLabel = (x1+x2)/2; yLabel = ym;
		}
		else if (sideNode1==ECFNode.SOUTH  && sideNode2==ECFNode.SOUTH) {
			ym = Math.max(y1,y2) + offsetULine;   // U-line south starting vertical
			xm1 = x1; ym1 = ym+yOffsetLabel;
			xm2 = x2; ym2 = ym+yOffsetLabel;
			g.drawLine(x1,y1,xm1,ym1);   // Vertical line
			g.drawLine(xm1,ym1,xm2,ym2); // Horizontal line
			g.drawLine(xm2,ym2,x2,y2);   // Vertical line
			xLabel = (x1+x2)/2; yLabel = ym;
		}
		else if ((sideNode1==ECFNode.EAST  && sideNode2==ECFNode.WEST) ||
				(sideNode1==ECFNode.WEST  && sideNode2==ECFNode.EAST)) {
			xm = (x1+x2)/2;                   // Z-line horizontal
			xm1 = xm+xOffsetLabel; ym1 = y1;
			xm2 = xm+xOffsetLabel; ym2 = y2;
			g.drawLine(x1,y1,xm1,ym1);   // Horizontal line
			g.drawLine(xm1,ym1,xm2,ym2); // Vertical line
			g.drawLine(xm2,ym2,x2,y2);   // Horizontal line
			xLabel = xm; yLabel = (y1+y2)/2;
		}
		else if ((sideNode1==ECFNode.NORTH  && sideNode2==ECFNode.SOUTH) ||
				(sideNode1==ECFNode.SOUTH  && sideNode2==ECFNode.NORTH)) {
			ym = (y1+y2)/2;                   // Z-line vertical
			xm1 = x1; ym1 = ym+yOffsetLabel;
			xm2 = x2; ym2 = ym+yOffsetLabel;
			g.drawLine(x1,y1,xm1,ym1);   // Vertical line
			g.drawLine(xm1,ym1,xm2,ym2); // Horizontal line
			g.drawLine(xm2,ym2,x2,y2);   // Vertical line
			xLabel = (x1+x2)/2; yLabel = ym;
		}
	} // drawLine

	/** Get the flowconnector at end iend of this FlowPath.
	* @param iend The end of this FlowPath.
	* @return The FlowConnector. */
	public FlowConnector getFlowConnector(int iend) {
		if (iend==1) return conn1; else return conn2;
	} // getFlowConnector
	/** Gets the index of the layer to which the flowpath belongs.
	* @return the index of the layer to which the flowpath belongs. */
	public int getLayer() {return ilayer;}
	/** Gets the name of the flowpath.
	* @return the name of the flowpath. */
	public String getName() { return flow.getName(); } // szName;}
	/** Gets the selected flow connector.
	* @return the connector at the end found by the contains() metod. */
	public FlowConnector getSelectedConnector() {return connSelected;}

	/** Open a dialog for setting properties of the flow path. */
	public void openDialog() {
		new FlowDialog(flow);
	} // openDialog

	/**
	* Paint the flowpath on layer with index ilayer.
	* @param g the graphics context.
	* @param ilayer index of the layer to which the flowpath belongs.
	*/
	public void paint(Graphics g,int ilayer) {
		if (this.ilayer==ilayer) {
			drawLine(g);
			paintArrow(g);
			conn1.paint(g);
			conn2.paint(g);
			if (fLabelShown) {
				// Flow name label
				label.setText(flow.getNameForLabel());
//				label.setText(flow.getName());
				Color colorBg = ECFModel.COLOR_DEFAULT;
				if (flow.isPrescribed()) colorBg = ECFModel.COLOR_PRESCRIBED;
				label.setBackgroundDefault(colorBg);
				int x = xLabel - label.getWidth()/2 + xOffsetLabel;
				int y = yLabel + yOffsetLabel;
				label.setPos(x,y);
				label.paint(g);
				// Flow quantity
				GraphLabel labelQuant = new GraphLabel(flow.getStringQuant(),x,y+14,Color.white);
				labelQuant.paint(g);
			} // label
		} // if
	} // paint

	/** Set side of node to which the head end of the flowpath is connected.
	* @param sideNode side of node, see <A href="ECFNode.html">ECFNode</A>.
	*/
	public void setSide1(int sideNode) {this.sideNode1 = sideNode;}

	/** Set side of node to which end 2 of the flowpath is connected.
	* @param sideNode side of node, see <A href="ECFNode.html">ECFNode</A>.
	 */
	public void setSide2(int sideNode) {this.sideNode2 = sideNode;}

	/**
	* Set start position (before translating) of the flowpath and
	* selected flow connector to current positions.
	*/
	public void setStartPos() {
		xStart = xOffsetLabel;
		yStart = yOffsetLabel;
		if (connSelected!=null) connSelected.setStartPos();
	} // setStartPos

	/** Set X-coordinate of the head end.
	* @param x1 X-coordinate. */
	public void setX1(int x1) {this.x1 = x1;}
	/** Set X-coordinate of end 2.
	* @param x2 X-coordinate. */
	public void setX2(int x2) {this.x2 = x2;}
	/** Set Y-coordinate of the head end.
	* @param y1 Y-coordinate. */
	public void setY1(int y1) {this.y1 = y1;}
	/** Set Y-coordinate of the tail end.
	* @param y2 Y-coordinate. */
	public void setY2(int y2) {this.y2 = y2;}
	/** Set flowpath as deselected. */
	public void setDeselected() { label.setDeselected(); }
	/** Set the name of the flowpath. */
	public void setName(String szName) {
//		this.szName = szName;
		flow.setName(szName);
	}
	/** Set selected flow connector.
	* @param iend Index of selected end = FlowPath.END_HEAD or FlowPath.END_TAIL. */
	public void setSelectedConnector(int iend) {
		if (iend==FlowPath.END_HEAD) connSelected = conn1;
		if (iend==FlowPath.END_TAIL) connSelected = conn2;
	} // setSelectedConnector

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(int ilayer,Units units) {
		if (this.ilayer==ilayer) flow.setUnitsIfNotSet(units);
	} // setUnitsIfNotSet

	/** Show properties of this FlowPath
	* @return A string telling the properties.*/
	public String toString() {
		if (flow==null) return "FlowPath: flow==null";
		return "FlowPath: szName="+flow.getName() // szName
				+ " ilayer="+ilayer +" x1="+x1 +" y1="+y1
				+" x2="+x2 +" y2="+y2
				+" sideNode1="+sideNode1 +" sideNode2="+sideNode2;
	} // toString

	/** Translate label or selected end by (dx,dy),
	* see method <code>contains</code> above.
	* @param dx translation in x-direction.
	* @param dy translation in y-direction. */
	public void  translate(int dx,int dy) {
		// Translate label
		if (fLabelSelected) {
			xOffsetLabel = xStart + dx;
			yOffsetLabel = yStart + dy;
		} // if
		// Translate connector
		if (connSelected!=null) {
			connSelected.translate(dx,dy);
			x1 = (int)conn1.getRefPoint().getX();
			y1 = (int)conn1.getRefPoint().getY();
			x2 = (int)conn2.getRefPoint().getX();
			y2 = (int)conn2.getRefPoint().getY();
		} // if
	} // translate

	/** Update the line of the flow path to current positions of ends
	* and current sides of connected nodes. */
	public void updateLine() {
		x1 = (int)conn1.getRefPoint().getX();
		y1 = (int)conn1.getRefPoint().getY();
		sideNode1 = conn1.getSide();
		x2 = (int)conn2.getRefPoint().getX();
		y2 = (int)conn2.getRefPoint().getY();
		sideNode2 = conn2.getSide();
	} // updateLine

	private void paintArrow(Graphics g) {
		// Empty arrow flow direction
		int dl = 20; //10; // Lenght of arrov head
		int dw = 8;  //4;  // Width of arrow head
		double pi = 3.14159;
		double angle = 0; // Clockwise from x-axis.
		if (sideNode1==ECFNode.NORTH) {
			g.drawLine(x1,y1,x1-dw,y1-dl);
			g.drawLine(x1,y1,x1+dw,y1-dl);
			angle = pi*1/2;
		}
		else if (sideNode1==ECFNode.EAST) {
			g.drawLine(x1,y1,x1+dl,y1-dw);
			g.drawLine(x1,y1,x1+dl,y1+dw);
			angle = pi;
		}
		else if (sideNode1==ECFNode.WEST) {
			g.drawLine(x1,y1,x1-dl,y1-dw);
			g.drawLine(x1,y1,x1-dl,y1+dw);
			angle = 0.0;
		}
		else if (sideNode1==ECFNode.SOUTH) {
			g.drawLine(x1,y1,x1-dw,y1+dl);
			g.drawLine(x1,y1,x1+dw,y1+dl);
			angle = pi*3/2;
		} else {
			try {
			angle = Math.atan2((y1-y2),(x1-x2)); // clockwise from x-axis to the right
			g.drawLine(x1,y1,(int)(x1-dl*Math.cos(angle)+dw*Math.sin(angle)),
							 (int)(y1-dl*Math.sin(angle)-dw*Math.cos(angle)));
			g.drawLine(x1,y1,(int)(x1-dl*Math.cos(angle)-dw*Math.sin(angle)),
							 (int)(y1-dl*Math.sin(angle)+dw*Math.cos(angle)));
			} catch(Exception e) {
				System.out.println(e);
			}
		}
		// Draw flow arrow
		if (flow.isPrescribed())
			// flowArrow(g,angle,flow.getFlowPrescribed(),ECFModel.COLOR_PRESCRIBED);
			flowArrow(g,angle,flow.getPixelValue(cpixelMaxArrowWidth),ECFModel.COLOR_PRESCRIBED);

		else
			// flowArrow(g,angle,flow.getFlowCalculated(),ECFModel.COLOR_CALCULATED);
			flowArrow(g,angle,flow.getPixelValue(cpixelMaxArrowWidth),ECFModel.COLOR_CALCULATED);
		// if
	} // paintArrow

	/** Paint a filled arrow with width propotional to flow rate.
	* @param g Graphics context.
	* @param theta Angle of flow direction.
	* @param cpixelArrowWidth Number of pixels at actual arrow width.
	* @param color The color of the filling.*/
	private void flowArrow(Graphics g, double theta, double cpixelArrowWidth, Color color) {
		int dl = 10; // Lenght of arrow head
		int dw =  4; // Width of arrow head at dl length
		// Filled arrow for flow magnitude
		int widthArrow = (int)(Math.abs(cpixelArrowWidth)+0.5);
		if (widthArrow>=cpixelMaxArrowWidth*1.2) {
			color = ECFModel.COLOR_WARNING;
			widthArrow = (int)(cpixelMaxArrowWidth*1.2);
		} // if
		int lengthArrow = 40;
		int dla = (int)dl* widthArrow/dw;
		int dwa = widthArrow;
		// Draw flow arrow
		double xArrow[] = new double[5];
		double yArrow[] = new double[5];
		if (cpixelArrowWidth>0) {
			// Arrow pointing in x-direction, points clockwise.
			xArrow[0]=x1;             yArrow[0]=y1;  // Arrow point
			xArrow[1]=x1-dla;         yArrow[1]=y1+dwa;
			xArrow[2]=x1-lengthArrow; yArrow[2]=y1+dwa;
			xArrow[3]=x1-lengthArrow; yArrow[3]=y1-dwa;
			xArrow[4]=x1-dla;         yArrow[4]=y1-dwa;
		} else {
			// Arrow pointing in minus x-direction, points clockwise.
			xArrow[0]=x1-lengthArrow; yArrow[0]=y1;  // Arrow point;
			xArrow[1]=x1-lengthArrow+dla; yArrow[1]=y1-dwa;
			xArrow[2]=x1;             yArrow[2]=y1-dwa;
			xArrow[3]=x1;             yArrow[3]=y1+dwa;
			xArrow[4]=x1-lengthArrow+dla; yArrow[4]=y1+dwa;
		} // if/else
		g.setColor(color);
		Polygon poly= new Polygon();
		double cos = Math.cos(theta);
		double sin = Math.sin(theta);
		for (int ipoint=0; ipoint<5; ipoint++) {
			// Make own transform because AffineTransform does not work with Applet
			int x = x1+(int)(cos*(xArrow[ipoint]-x1)-sin*(yArrow[ipoint]-y1));
			int y = y1+(int)(sin*(xArrow[ipoint]-x1)+cos*(yArrow[ipoint]-y1));
			poly.addPoint(x,y);
		}// while
		g.fillPolygon(poly);
	} // flowArrow

} // FlowPath