import java.util.*;

/**
* A <code>FlowRatio</code> between an outflow and an inflow of resources
* of the same kind (units).<p>
* Prescribed and calculated ratios are stored as vectors in
* <A href="Variable.html">Variables</A>.
* The <code>get</code> and <code>set</code> methods operate on the
* the Variable value that corresponds to the current time increment.
*/
public class FlowRatio extends Variable {

	/** */
	protected Splitter splitter;
	/** Index of ratio. */
	protected int iratio = 0;
	/** Flow out */
	protected Flow flowOut = null;

	/** Create a ratio of resources with ilayer = 0.
	* @param timeaxis The time axis against which the FlowRatio is plotted.
	*/
	public FlowRatio(Splitter splitter) {
		super();
		this.splitter = splitter;
		szName = "FlowRatio";
		Calculator.addFlowRatio(this);
	} // FlowRatio

	/** Create a ratio of resources.
	* @param timeaxis The time axis against which the FlowRatio is plotted.
	* @param szName Name of ratio
	*/
	public FlowRatio(String szName) {
		super();
		this.szName = szName;
		Calculator.addFlowRatio(this);
	} // FlowRatio

	/** Calculate flow ratios from flows. */
	public void calculate() {
		double ratio;
		double sumofFlowRatesIn = 0.0;

		Vector vrflowIn = splitter.getFlowsIn();
		if (flowOut==null || vrflowIn.size()==0) return;
		// Calculate sum of inflows
		Iterator iter = vrflowIn.iterator();
		while (iter.hasNext()) sumofFlowRatesIn += ((Flow)iter.next()).getBaseValueCalculated();
//		System.out.println("FlowRatio.calculate: sumofFlowRatesIn=" + sumofFlowRatesIn +
//			" flow out=" + flowOut.getBaseValueCalculated());
		// Calculate outflow ratio
		try  { //(Math.abs(sumofFlowRatesIn)>1.0e-4) {
			ratio = flowOut.getBaseValueCalculated()/sumofFlowRatesIn;
		}
		catch(Exception e) {
			System.out.println("FlowRatio.calculate:" + e);
			ratio = 0.0;
		}
//		System.out.println("FlowRatio.calculate: flowOut=" + flowOut.getName() + " ratio=" + ratio);
		setBaseValueCalculated(ratio);
	} // calculate

	/**
	* Decode flow ratio from XML code.
	* @param szXMLCode The xml description of the flow ratio.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("flowindex")) {
					int iflow = Integer.parseInt(szSub);
					flowOut = Calculator.getFlow(iflow);
				} // if
				if (szTag.equals("Variable")) super.decodeXML(szSub);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println(e);
		} // try/catch
	} // decodeXML

	/** Export FlowRatio values.
	* @return A string with  label and values separated by semicolon (;).
	*/
	public String export() {
		String szExport = ECFUtil.removeLineFeeds(szName) + ";";
		szExport = Quantity.padToLength(szExport,20);
		double[] vrvalueDisplay = getDisplayValuesCalculated();
		for (int ivalue=0; ivalue<vrvalueDisplay.length; ivalue++)
			szExport += " " + Quantity.formatDouble(vrvalueDisplay[ivalue]) + ";";
		szExport += "\n";
		return szExport;
	} // export

	/** Get Flow out.
	* @return Flow out. */
	public Flow getFlowOut() { return flowOut; }

	/** List XML code for the flow ratio.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<flowindex>"+ECF.calculator.getIndexOfFlow(flowOut)+"</flowindex>";
		szXML += szBlanks +"<Variable>"+ super.listXML("")
		 		+szBlanks +"</Variable>";
		return szXML;
	} // listXML

	/** Set Flow out.
	* @param Flow out. */
	public void setFlowOut(Flow flowOut) { this.flowOut = flowOut; }


	/** Returns a string representation of the FlowRatio.*/
/*	public String toString() {
		String sz = "FlowRatio: flowOut=";
		if (flowOut!=null) sz += flowOut.getName();
		else sz += "not defined";
		sz += ", " + super.toString();
		return sz;
	} // toString
*/
	/** Show properties of the FlowRatio
	* @return A string telling the properties.*/
	public String toString() {
		String szNameNoLineFeeds = ECFUtil.removeLineFeeds(szName);
		String sz = "\tFlowRatio: ";
		sz += Quantity.padToLength(szNameNoLineFeeds,25) + " ";
		if (super.state==DEFAULT)    sz += "DEFAULT=0  ";
		if (super.state==CALCULATED) sz += "CALCULATED= " + getStringQuantCalculated();
		if (super.state==PRESCRIBED) sz += "PRESCRIBED= " + getStringQuantPrescribed();
		return sz;
	} // toString

} // FlowRatio