import java.awt.*;   // Graphics
import java.util.*;  // Vector

/** Frame connectors of the frame of a submodel. Each frame connector connects to a
* twin flow connector <a href="TwinFlowConnector.html">TwinFlowConnector</a>
* which in turn connects to one flow path outside the
* frame and one flowpath inside the frame.
*/
public class FrameConnectors {

	/** Index of model layer. */
	protected int ilayer=0;
	/** The model frame to which the frame connectors belongs. */
	protected ModelFrame modelframe = null;
	/** Vector of twin flow connectors attached to west side. */
	protected Vector vrconnWest = new Vector();
	/** Vector of twin flow connectors attached to east side. */
	protected Vector vrconnEast = new Vector();
	/** Vector of twin flow connectors attached to north side. */
	protected Vector vrconnNorth = new Vector();
	/** Vector of twin flow connectors attached to south side. */
	protected Vector vrconnSouth = new Vector();

	/** Constructs frame connectors.
	* @param ilayer Index of the layer to which the frame connectors belongs.
	* @param modelframe The model frame to which the frame connectors belongs.*/
	public FrameConnectors(int ilayer, ModelFrame modelframe) {
		this.ilayer = ilayer;
		this.modelframe = modelframe;
	} // FrameConnectors

	/** Add flow path to a vector of frame connectors on one side.
	* @param flowpath The flow path to be added.
	* @param vrconn Vector of of flow paths before addition.
	* @param iconnNew Index of added connector.  */
	protected void addFlowPath(FlowPath flowpath, Vector vrconn, int iconnNew) {
		TwinFlowConnector twinflowconn = new TwinFlowConnector(flowpath,modelframe);
		if (vrconn.size()==0)
			vrconn.add(twinflowconn);
		else if (vrconn.size()>iconnNew)
			vrconn.insertElementAt(twinflowconn,iconnNew);
		else
			vrconn.add(twinflowconn);
	} // addFlowPath

	/**
	* Connect the <code>flowpath</code> to this frame connectors.
	* @param flowpath The flowpath to be connected.
	* @param iconnNew Index of added connector.
	*/
	public void connectFlowPath(FlowPath flowpath,int iconnNew) {
		// Side of connector
		FlowConnector flowconn = flowpath.getSelectedConnector();
		if (flowconn==null) return;
		int iside = flowconn.getSide();
		if (iside==ECFNode.WEST) addFlowPath(flowpath,vrconnWest,iconnNew);
		if (iside==ECFNode.EAST) addFlowPath(flowpath,vrconnEast,iconnNew);
		if (iside==ECFNode.NORTH) addFlowPath(flowpath,vrconnNorth,iconnNew);
		if (iside==ECFNode.SOUTH) addFlowPath(flowpath,vrconnSouth,iconnNew);
		setPositions();
	} // connectFlowPath

	/**
	* Decode frame connectors from XML code.
	* @param szXMLCode The xml description of the frame connectors .
	*/
	public void decodeXML(String szXMLCode) {
//		System.out.println("FrameConnectors.decodeXML: ilayer=" +
//			ilayer + "\n    " +szXMLCode);
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
//				System.out.println("FrameConnectors.decodeXML: szTag= " + szTag);
				if (szTag.equals("West")) decodeXMLOneSide(szSub,vrconnWest);
				if (szTag.equals("East")) decodeXMLOneSide(szSub,vrconnEast);
				if (szTag.equals("North")) decodeXMLOneSide(szSub,vrconnNorth);
				if (szTag.equals("South")) decodeXMLOneSide(szSub,vrconnSouth);
				szXMLCode = xml.remainder();
			} // while szXMLCode
		} catch(Exception e) {
			System.out.println("FrameConnectors.decodeXML: " + e +
			"\n  " + szXMLCode);
		} // try/catch
//		System.out.println("FrameConnectors.decodeXML: end");
	} // decodeXML

	/**
	* Decode frame connectors on one side from XML code.
	* @param szXMLCode The xml description of the frame connectors on one side.
	*/
	public void decodeXMLOneSide(String szXMLCode, Vector vrconn) {
//		System.out.println("FrameConnectors.decodeXMLOneSide: " + szXMLCode);
		XMLParser xml = null;
		int iconn = 0;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
//				System.out.println("FrameConnectors.decodeXMLOneSide: szTag= " + szTag);
				if (szTag.equals("TwinFlowConnector")) {
//					System.out.println("****** FrameConnectors.decodeXMLOneSide: szSub= " + szSub);
					TwinFlowConnector twinflowconn = (TwinFlowConnector)vrconn.elementAt(iconn);
					twinflowconn.decodeXML(szSub);
					iconn++;
				} //
				szXMLCode = xml.remainder();
			} // while szXMLCode
		} catch(Exception e) {
			System.out.println("FrameConnectors.decodeXMLOneSide: " + e +
			"\n  " + szXMLCode);
		} // try/catch
//		System.out.println("FrameConnectors.decodeXMLOneSide: end");
	} // decodeXMLOneSide


	/**
	* Finds a flowpath of the submodel. An <code>FlowPath</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return FlowPath if a flowpath was found, else <code>null</code>.
	*/
	public FlowPath findFlowPath(int ilayer,int x,int y) {
		Iterator iter = new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
	    while (iter.hasNext()) {
			TwinFlowConnector twinflowconn = (TwinFlowConnector)iter.next();
			FlowPath flowpath = twinflowconn.findFlowPath(ilayer,x,y);
			if (flowpath!=null) return flowpath;
		} // while
		return null;
	} // findFlowPath

	/** Get the flow connector with index iflowpath from the side iside.
	* @param iside The side of the frame. ECFNode.WEST, ECFNode.EAST ...
	* @param iflowpath The index of the flow path. */
	public TwinFlowConnector getTwinFlowConnector(int iside, int iflowpath) {
		if (iside==ECFNode.WEST) return (TwinFlowConnector)vrconnWest.elementAt(iflowpath);
		if (iside==ECFNode.EAST) return (TwinFlowConnector)vrconnEast.elementAt(iflowpath);
		if (iside==ECFNode.NORTH) return (TwinFlowConnector)vrconnNorth.elementAt(iflowpath);
		if (iside==ECFNode.SOUTH) return (TwinFlowConnector)vrconnSouth.elementAt(iflowpath);
		return null;
	} // getTwinFlowConnector

	/** Get index of node layer for this connectors.
	* @return Index of the model layer to which this
	* <code>NodeConnectors</code> belongs.*/
	public int getLayerIndex() {return ilayer;}

	/** List XML code for the frame connectors.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<West>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnWest);
		szXML += "</West>";
		szXML += szBlanks +"<East>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnEast);
		szXML += "</East>";
		szXML += szBlanks +"<North>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnNorth);
		szXML += "</North>";
		szXML += szBlanks +"<South>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnSouth);
		szXML += "</South>";
		return szXML;
	} // listXML

	/** List XML code for the flow connectors on one side this nodeconnectors.
	* @return XML code. */
	private String listXMLOneSide(String szBlanks,Vector vrconn) {
		String szXML = "";
		Iterator iter = vrconn.iterator();
		while (iter.hasNext()) {
			TwinFlowConnector twinflowconn = (TwinFlowConnector)iter.next();
			szXML += "<TwinFlowConnector>";
			szXML += twinflowconn.listXML("");
			szXML += "</TwinFlowConnector>";
		} // while
		return szXML;
	} // listXMLOneSide

	/**
	* Paint the flowpaths of the layer with index ilayer.
	* @param g the graphics context.
	* @param ilayer index of the layer to which the flowpath belongs.
	*/
	public void paint(Graphics g,int ilayer) {
		Iterator iterFlowConnector = new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
		while (iterFlowConnector.hasNext()) {
			TwinFlowConnector twinflowconn = (TwinFlowConnector)iterFlowConnector.next();
			twinflowconn.paint(g,ilayer);
		} // while
	} // paint

	/** Remove flow from the model frame.
	* @param flow The flow to be removed; */
	public void removeFlow(Flow flow) {
		// Remove flowconnector from any side,
		// side that does not contain flowconn is unaffected.
		removeFlowOnOneSide(vrconnEast,flow);
		removeFlowOnOneSide(vrconnWest,flow);
		removeFlowOnOneSide(vrconnNorth,flow);
		removeFlowOnOneSide(vrconnSouth,flow);
		setPositions();
	} // removeFlow

	/** Remove flow from one side of the model frame.
	* @param vrflowconn Vector of flow connectors on one side of frame.
	* @param flow The flow to be removed; */
	public void removeFlowOnOneSide(Vector vrflowconn, Flow flow) {
		// Remove flowconnector from one side.
		Iterator iter = vrflowconn.iterator();
		TwinFlowConnector twinconn = null;
		while (iter.hasNext()) {
			twinconn = (TwinFlowConnector)iter.next();
			if (twinconn.hasFlow(flow)) {
				twinconn.removeFlowPaths();
				vrflowconn.remove(twinconn);
				break;
			} // if
		} // while
	} // removeFlow

	/**
	* Set positions of connected <A href="TwinFlowConnector.html">TwinFlowConnector</A>s.
	*/
	public void setPositions() {
		// Position of node symbol
		Rectangle rect = modelframe.getRect();
		int xPos = (int)rect.getX();  // Position of upper left corner
		int yPos = (int)rect.getY();
		int height = (int)rect.getHeight();
		int width = (int)rect.getWidth();

		// East connectors
		int cconn = vrconnEast.size();
		if (cconn>0) {
			int dy = height/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				TwinFlowConnector conn = (TwinFlowConnector)vrconnEast.elementAt(iconn);
				conn.setPos(xPos+width,yPos+dy/2+iconn*dy,height,width);
				conn.setFlowPathIndex(iconn);
			}  // for
		} // if
		// West connectors
		cconn = vrconnWest.size();
		if (cconn>0) {
			int dy = height/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				TwinFlowConnector conn = (TwinFlowConnector)vrconnWest.elementAt(iconn);
				conn.setPos(xPos,yPos+dy/2+iconn*dy,height,width);
				conn.setFlowPathIndex(iconn);
			}  // for
		} // if
		// North connectors
		cconn = vrconnNorth.size();
		if (cconn>0) {
			int dx = width/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				TwinFlowConnector conn = (TwinFlowConnector)vrconnNorth.elementAt(iconn);
				conn.setPos(xPos+dx/2+iconn*dx,yPos,height,width);
				conn.setFlowPathIndex(iconn);
			}  // for
		} // if
		// South connectors
		cconn = vrconnSouth.size();
		if (cconn>0) {
			int dx = width/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				TwinFlowConnector conn = (TwinFlowConnector)vrconnSouth.elementAt(iconn);
				conn.setPos(xPos+dx/2+iconn*dx,yPos+height,height,width);
				conn.setFlowPathIndex(iconn);
			}  // for
		} // if
	} // setPositions

} // FrameConnectors