import java.awt.*;
import javax.swing.*;
import java.util.*;

/** A label that can be positioned in any place in a graphics context.
*/
public class GraphLabel {

	/** The text displayed on th label. */
	protected String szText;
	/** X-coordinate, pixels, for the lower left corner of the enclosing rectangle.*/
	protected int x = 0;
	/** Y-coordinate, pixels, for the lower left corner of the enclosing rectangle.*/
	protected int y = 0;
	/** Height per line, pixels, of the enclosing rectangle.*/
	protected int heightLine = 10;
	/** Height, pixels, of the enclosing rectangle.*/
	protected int height = heightLine;
	/** Width, pixels, of the enclosing rectangle.*/
	protected int width = 40;
	/** Enclosing rectangle, top-left point (x,y-height).*/
	protected Rectangle rectLabel = new Rectangle(x,y-height,width,height);
	/** Paints the interior, edges and name of the node symbol. */
	protected Painter painter;
	/** Indicator if label was selected. */
	boolean fselected = false;
	/** The default fill color for the node symbol. */
	protected Color colorFillDefault = ECFModel.COLOR_DEFAULT;
	/** The fill color for the node symbol. */
	protected Color colorFill = colorFillDefault;
	/** Width of border of node symbol (pixels). */
	protected int widthStroke = 0;

	/** Creates a graphical label.
	* @param szText The text of the label.
	*/
	public GraphLabel(String szText) {
		this.szText = szText;
		makePainter();
	} // GraphLabel

	/** Creates a graphical label.
	* @param szText The text of the label.
	* @param x X-coordinate, pixels, for the lower left corner.
	* @param y Y-coordinate, pixels, for the lower left corner.
	*/
	public GraphLabel(String szText,int x, int y) {
		this(szText);
		this.x = x;
		this.y = y;
		width = 6*szText.length();
		makeRectBackgnd();
		makePainter();
	} // GraphLabel

	/** Creates a graphical label.
	* @param szText The text of the label.
	* @param x X-coordinate, pixels, for the lower left corner.
	* @param y Y-coordinate, pixels, for the lower left corner.
	* @param colorFillDefault Default fill color.
	*/
	public GraphLabel(String szText,int x, int y, Color colorFillDefault) {
		this(szText,x,y);
		this.colorFillDefault = colorFillDefault;
		colorFill = colorFillDefault;
		makePainter();
	} // GraphLabel

	/** Checks if the label contains the coordiantes x,y.
	* @return true if x,y contained, else false. */
	public boolean	contains(int x,int y) {
		fselected = getShape().contains(x,y);
		makePainter();
		return fselected;
	} // contains

	/**
	* Decode label from XML code.
	* @param szXMLCode The xml description of the label.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("szText")) szText = szSub;
				if (szTag.equals("x")) x = Integer.parseInt(szSub);
				if (szTag.equals("y")) y = Integer.parseInt(szSub);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("ExplanatoryLabel.decodeXML: \n" + e);
		} // try/catch
		width = 6*szText.length();
		makeRectBackgnd();
	} // decodeXML

	/* Get the shape of the model frame.
	* @return the Shape of the frame
	*/
	public Shape getShape() {
		return rectLabel;
	} // getShape

	/** Get heigth of enclosing rectangle.
	* @return The height of enclosing rectangle (pixels). */
	public int getHeight() { return height; }

	/** Get the text of the label.
	* @return The text of the label. */
	public String getText() { return szText; }

	/** Get width of enclosing rectangle.
	* @return The width of enclosing rectangle (pixels). */
	public int getWidth() { return width; }

	/** List XML code for the label.
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = "";
		szXML += szBlanks +"<szText>" + szText + "</szText>";
		szXML += szBlanks +"<x>" + x + "</x>";
		szXML += szBlanks +"<y>" + y + "</y>";
		return szXML;
	} // listXML

	/** Make a background rectangle for the text
	* with (x,y) = lower left corner of rectangle.
	*/
	protected void makeRectBackgnd() {
		StringTokenizer tok = new StringTokenizer(szText,"\n");
		int ctoken = tok.countTokens();
		width = 10;
		while (tok.hasMoreTokens()) {
			String szLine = tok.nextToken();
			width = Math.max(width,6*szLine.length());
			height = Math.max(1,ctoken)*heightLine;
		} // while
		rectLabel = new Rectangle(x,y-height,width,height);
	} // makeRectBackgnd

	/**
	* Make a painter to paint the symbol of the node.
	*/
	protected void makePainter() {
		if (fselected) {
			colorFill = ECFModel.COLOR_SELECTED;
		} else {
			colorFill = colorFillDefault;
		} // if/else
		Painter painterFill = new FillPainter(colorFill);
		Stroke stroke = new BasicStroke(widthStroke);
		Painter painterDraw = new DrawPainter(Color.black,stroke);
		painter = new MultiPainter(painterFill,painterDraw);
	} // makePainter

	/** Paints the label.
	* @param g graphics context.
	*/
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;
		paint(g2);
	} // Paint

	/** Paints the label.
	* @param g2 graphics context.
	*/
	public void paint(Graphics2D g2) {
		if (g2==null) return;
		makePainter();
		painter.paint(g2,rectLabel);
		// Paint text, (x,y) = lower left corner of text
		// (x,y0) = lower left corner of first line of text
		int y0 = y-height+heightLine;
		StringTokenizer tok = new StringTokenizer(szText,"\n");
		while (tok.hasMoreTokens()) {
			String szLine = tok.nextToken();
			g2.drawString(szLine,x+1,y0-1);
			y0 += heightLine;
		} // while
	} // Paint

	/** Set the background color of the label. */
	public void setBackgroundDefault(Color color) {
		colorFillDefault = color;
	} // setBackgroundDefault

	/** Set label deselected. */
	public void setDeselected() {
		fselected = false;
	} // setDeselected

	/** Set the position of the lable.
	* @param x,y The coordinates of the upper left corner to be set.
	*/
	public void setPos(int x, int y) {
		this.x = x;
		this.y = y;
		makeRectBackgnd();
	} // setPos

	/** Set the text of the lable.
	* @param szText The text to be set.
	*/
	public void setText(String szText) {
		this.szText = szText;
		width = 6*szText.length();
		makeRectBackgnd();
	} // setText

} // GraphLabel