import java.awt.*;
import java.util.*;
import javax.swing.*;

/**
* An input-output module of the economic system. An input-output module
* is a node that redistributes a number of input flows into a number of
* output flows.<br>
* There are three kinds of input-output nodes:
* <ol>
* <li>Primary inputs to process inputs. <br>Defines how much production factors
* of different kinds that are needed for the listed processes. </li>
* <li>Process inputs to process outputs. <br>Defines how much resources from
* other processes that are needed for the listed processes. </li>
* <li>Process outputs to final uses. <br>Defines how much output is consumed
* by different kinds of users. </li>
* </ol>
*/
public class InputOutput extends ECFNode {

	/** Tolerance for cursor postion inside edges of input-output symbol for resizing*/
	protected int tol = 10;
	/** Start height of rectangle enclosing node symbol before resizing. */
	protected int heightStart;
	/** Start width of rectangle enclosing node symbol before resizing. */
	protected int widthStart;
	/** Side to move when resizing. */
	protected int szToMove = NONE; // EAST, WEST, NORTH, SOUTH

	/** Index of input-output module. */
	static int iinputoutput = 1;
	/** Reset input-output module index. */
	public static void resetIndex() { iinputoutput = 1; }

	/** Index of selected money layer */
	protected int ilayerMoneySelected = 0;
	/** Index of selected commodity layer */
	protected int ilayerCommoditySelected = 1;
	/** Index of added connector when connecting a new flow path
	* by the connectFlowPath method. */
	protected int iconnNew = 0;

	/** The time axis along which the prices are varied. */
	protected TimeAxis timeaxis;
	/** Vector of input flow paths */
	protected Vector vrFlowPathInput = new Vector();
	/** Vector of output flow paths */
	protected Vector vrFlowPathOutput = new Vector();
	/** A vector of branches or users of inputs.*/
	protected Vector vrBranch = new Vector();
	/** A vector of price relations. */
	protected Vector vrprice = new Vector();
	/** A vector of price equations. */
	protected Vector vrequationPrice = new Vector();

	/**
	* Creates a new input-output module.
	* @param model The parent model to which the input-output module belongs.
	* @param x,y Coordinates of the upper left corner of the input-output module symbol
	*/
	public InputOutput(ECFModel model, int x, int y) {
		super(model,x,y);
		this.timeaxis = ECF.getTimeAxis();
		this.clayer = ECF.getNoOfLayers();
		height = 80;
		width = 80;
//		setName("InputOutput"+iinputoutput);
		setName("InputOutput not yet implenented");
		for (int ilayer=0; ilayer<clayer; ilayer++)
			addLayer(ilayer);
		iinputoutput++;
	} // InputOutput

	/** Add a layer with index ilayer to this input-output module.
	* @param ilayer Index of the layer to be added. */
	public void addLayer(int ilayer) {
		vrNodeConnectors.add(new NodeConnectors(ilayer,this));
	} // addLayer

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		// Clear units of sub model of input-output module
	} // clearUnits

	/** Connect flow path to this node.
	* @param ilayer Index of the layer of the flow path
	* @param flowpath the flow path to be connected. */
	public void connectFlowPath(int ilayer, FlowPath flowpath) {
		super.connectFlowPath(ilayer,flowpath);
		// Add input or output flow
		if (flowpath.getSelectedConnector().isHeadEnd()) {
			vrFlowPathInput.add(flowpath);
		} else {
			vrFlowPathOutput.add(flowpath);
		} // if
		// Update fill color and recalculate
		updateNodeContext();
	} // connectFlowPath

	/**
	* Check if the input-output module symbol cointains the point (x,y).
	* @param ilayer not used, shown in all layers
	* @param x x-coordanate to be contained by the input-output module symbol
	* @param y y-coordanate to be contained by the input-output module symbol
	* @return true if the point (x,y) is contained, else false
	*/
	public boolean contains(int ilayer,int x,int y) {
		return getShape().contains(x,y);
	} // contains

	/**
	* Decode junction items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		// Decode ECFNode items
		super.decodeXMLItems(szTag,szSub);
		// Decode special input-output module items
		if (szTag.equals("ilayerMoneySelected")) {
			ilayerMoneySelected = Integer.parseInt(szSub);}
		if (szTag.equals("ilayerCommoditySelected")) {
			ilayerCommoditySelected = Integer.parseInt(szSub);}
		if (szTag.equals("Price")||szTag.equals("price")) {
			System.out.println("InputOutput.decodeXMLItems: Price not adopted for InputOutput!!!");
//			Price price = new Price(this);
//			price.decodeXML(szSub);
//			vrprice.add(price);
		} // if
	} // decodeXMLItems

	/**
	* Decode node connectors from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLNodeConnectors(String szXMLCode) {
		// Decode external node connectors of the input-output module (ECFNode)
		super.decodeXMLNodeConnectors(szXMLCode);
	} // decodeXML

	/**
	* Release associations to flow paths before deleting node.
	*/
	public void	disconnectThis() {
		super.disconnectThis();
		System.out.println("InputOutput.disconnectThis: What more to be done ???");
	} // deleteThis

	/** Get index of selected money layer.
	* @return Index of selected money layer. */
	public int getIndexOfSelectedMoneyLayer() { return ilayerMoneySelected; }

	/** Get index of selected commodity layer.
	* @return Index of selected commodity layer. */
	public int getIndexOfSelectedCommodityLayer() { return ilayerCommoditySelected; }

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors(int ilayer) {
		try {
			return (NodeConnectors)vrNodeConnectors.elementAt(ilayer);
		} catch(Exception e) { return null;}
	} // getNodeConnectors

	/** Get node specific equations for price relations,.
	* @return A vector of <a href="LinearEquation.html">
	*         <code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() throws Exception {
		Vector vrequations = new Vector();
		// Price relations
		Iterator iterPrice = vrprice.iterator();
		while (iterPrice.hasNext()) {
			Price price = (Price)iterPrice.next();
			if (price.isPrescribed()) {
				LinearEquation equ = price.getPriceEquation();
				if (equ!=null) vrequations.add(equ);
			} // if
		} // while
		return vrequations;
	} // getNodeSpecificEquations

	/** Get price relations.
	* @return A vector of Price objects. */
	public Vector getPriceRelations() { return vrprice; }

	/** Get a double arrow resize cursor directed up/down for horizontal sides
	* and left/right for vertical sides.
	* @param xP,yP Pointer coordinates (pixels). */
	public Cursor getResizeCursor(int xP, int yP) {
		Cursor cursor = super.getResizeCursor(xP,yP);
		int x0 = x; int x1 = x0 + width; // x-coord for left and right sides
		int y0 = y; int y1 = y0 + height; // y-coord for top and botton sides
		if (xP>=x0 && xP<x0+tol)      szToMove = WEST;    // Rev.1
		else if (xP>x1-tol && xP<=x1) szToMove = EAST;    // Rev.1
		else if (yP>=y0 && yP<y0+tol) szToMove = NORTH;
		else if (yP>y1-tol && yP<=y1) szToMove = SOUTH;
		else szToMove = NONE;
		if ((szToMove==WEST)  || szToMove==EAST)  cursor = new Cursor(Cursor.E_RESIZE_CURSOR);
		if ((szToMove==NORTH) || szToMove==SOUTH) cursor = new Cursor(Cursor.S_RESIZE_CURSOR);
		return cursor;
	} // showResizeCursor

	/**
	* Get the shape of the input-output module symbol.
	* @return the Shape of the symbol
	*/
	public Shape getShape() {
		return new Rectangle(x,y,width,height);
	} // getShape

	/** Check if node has a flow balance for this layer.
	* @return true if has flow balance, else false.*/
	public boolean hasFlowBalance(int ilayer) {
		try { return ((NodeConnectors)vrNodeConnectors.elementAt(ilayer)).hasFlowBalance();
		} catch(Exception e) { return false; }
	} // hasFlowBalance

	/** List XML code for the input-output module.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = super.listXML(szBlanks);
		szXML += szBlanks +"<ilayerMoneySelected>" +
			ilayerMoneySelected +"</ilayerMoneySelected>";
		szXML += szBlanks +"<ilayerCommoditySelected>" +
			ilayerCommoditySelected +"</ilayerCommoditySelected>";
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			System.out.println("InputOutput.listXML: Price not adopted for InputOutput!!!");
			Price price = (Price)iter.next();
			szXML += szBlanks +"<price>";
//			szXML += price.listXML("  "+szBlanks);
			szXML += szBlanks +"</price>";
		} // while
		return szXML;
	} // listXML

	/** Set resize side to NONE (applies to InputOutput). */
	public void quitResize() {
		szToMove=NONE;
	}

	/** Open a dialog window for node properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		JFrame mainframe = ECF.frameMain;
		int x = mainframe.getX();
		int y = mainframe.getY();
		InputOutputDialog dialog = new InputOutputDialog(this, new Point(x+20,y+20));
		dialog.show();
	} // openDialog

	/**
	* Paint the input-output module with painter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		if (g2==null) return;
		super.paint(g2,ilayer);
		textName.paint(g2, x+10, y+20);
		paintPriceLabels(g2,ilayer);
	} // paint

	/**
	* Paint price labels at connectors for flow paths.
	* @param g2 graphics context.
	* @param ilayer index of layer to be painted.
	*/
	private void paintPriceLabels(Graphics2D g2,int ilayer) {
		NodeConnectors conns = getNodeConnectors(ilayer);
		Iterator iterConn = conns.getNodeConnectorsIterator();
		while (iterConn.hasNext()) {
			FlowConnector conn = (FlowConnector)iterConn.next();
			Flow flow = conn.getFlowPath().getFlow();
			// Find price that has this flow
			Iterator iterprice = vrprice.iterator();
			while (iterprice.hasNext()) {
				Price price = (Price)iterprice.next();
				if (price.hasFlow(flow)) {conn.paintPriceMarker(g2,price.isPrescribed());}
			} // while price
		} // while connector
	} // paintPriceLabels

	/** Receive a message that a new flow path connector has been added
	* from NodeConnectors when connecting a new flow path by the
	* connectFlowPath method above.
	* @param iconnNew Index of added connector. */
	public void registerNewConnector(int iconnNew) {
		this.iconnNew = iconnNew;
	} // registerNewConnector

	/** Remove flow from the input-output module and its corresponding price.
	* @param flowpath The flow path which price shall be removed. */
	public void removeFlowReference(FlowPath flowpath) {

		// Remove flow path
		if (flowpath.getSelectedConnector().isHeadEnd()) {
			vrFlowPathInput.remove(flowpath);
		} else {
			vrFlowPathOutput.remove(flowpath);
		} // if

		// Find and remove price reference
		Flow flow = flowpath.getFlow();
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.hasFlow(flow)) {
				timeaxis.remove(price);
				vrprice.remove(price);
				Calculator.removePrice(price);
				break;
			} // if
		} // while

		Calculator.calculate();
	} // removeFlowReference

	/** Set if input-output module has a flow balance for this layer.
	* @param f true if has flow balance, else false.*/
	public void setHasFlowBalance(int ilayer, boolean f) {
		try { ((NodeConnectors)vrNodeConnectors.elementAt(ilayer)).setHasFlowBalance(f); }
		catch(Exception e) {;}
	} // setHasFlowBalance

	/** Set index of selected money layer.
	* @param ilayerMoneySelected Index of selected money layer. */
	public void setIndexOfSelectedMoneyLayer(int ilayerMoneySelected) {
		this.ilayerMoneySelected = ilayerMoneySelected; }

	/** Set index of selected commodity layer.
	* @param ilayerCommoditySelected Index of selected commodity layer. */
	public void setIndexOfSelectedCommodityLayer(int ilayerCommoditySelected) {
		this.ilayerCommoditySelected = ilayerCommoditySelected; }

    /**
    * Translate or resize the node symbol.
    * @param dx translation or resizing in x direction (pixels)
    * @param dy translation or resizing in y direction (pixels)
    */
	public void translate(int dx, int dy) {
		if      (szToMove==NONE)  {super.translate(dx,dy);}
		else if (szToMove==WEST && widthStart-dx>=3*tol)  {x=xStart+dx; width=widthStart-dx;}
		else if (szToMove==EAST && widthStart+dx>=3*tol)  {width=widthStart+dx;}
		else if (szToMove==NORTH && heightStart-dy>=3*tol) {y=yStart+dy; height=heightStart-dy;}
		else if (szToMove==SOUTH && heightStart+dy>=3*tol) {height=heightStart+dy;}
		super.repositionConnectors();
	} // translate

	/** Set the name of the node.
	* @param szName The name to be set.
	*/
	public void setName(String szName) {
		super.setName(szName);
	} // setName

	/**
	* Set start position (before translating) of the node to is current
	* position.
	*/
	public void setStartPos() {
		super.setStartPos();
		// Set start size of input-output module symbol before resizing
		heightStart = height; widthStart = width;
	} // setStartPos

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
/*	public void setUnitsIfNotSet(int ilayer,Units units) {
		// Set units of sub model of input-output module
		if (submodel!=null) submodel.setUnitsIfNotSet(ilayer,units);
	} // setUnitsIfNotSet
*/
	/** Show properties of this InputOutput
	* @return A string telling the properties.*/
	public String toString() {
		return "InputOutput:   szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag for this input-output module.
	* @return XML tag. */
	public String xmltag() {return "inputoutput";}

} // InputOutput