/** A linear equation with <code>cvar</code> number of variables. <br>
*  The equation is represented by it's coefficients and a right member: <br>
*  <code>[a1 a2 a3 ... an | rm]</code> where
*  <code>ai</code> = coefficents, <code>rm</code> = right member
*/

public class LinearEquation implements Comparable {

	double[] equation;
	int cvar;
	String szType = "LinearEquation";
	String szNameNode = "Eq. Node Name";
	ECFNode node = null;
	int ilayer = 0;

	LinearEquation(int cvar) {
		this.cvar = cvar;
		equation = new double[cvar+1];
		for (int ivar=0; ivar<cvar+1; ivar++) {
			equation[ivar]=0.0;
		} // for
	} // LinearEquation

	/** Compare the min indices of flows in two equations
	* by <code>equ1.compareTo((Object)equ2)</code>
	* @return -1 if equ1 has the lowest min flow index, <br>
	* 0 if min flow indices are equal, <br>
	* 1 if equ1 has the highest flow index.
	*/
	public int compareTo(Object equ2) {
		if (this.getMinFlowIndex()<((LinearEquation)equ2).getMinFlowIndex()) return -1;
		else if (this.getMinFlowIndex()==((LinearEquation)equ2).getMinFlowIndex()) return 0;
		else { return 1; }
	} // compareTo

	/** Get the coefficients of the equation, including right member.
	* @return A double array of coefficiens.*/
	public double[] getCoefficients() {return equation;}

	/** Get the description of the equation,
	* for example */
	public String getDescription() {
		String szNameLayer = ECF.getLayer(getLayerIndex()).getName();
		String sz = getType() + " of node " + getNodeName() + " in layer " +
		szNameLayer;
		return sz;
	} // getDescription

	/** Get an array of indices of the flows that are included
	*   in the equation.
	* @return An array of flow indices.
	*/
	public int[] getFlowIndices() {
		// Count number of coeff different from zero
		int cflow = 0;
		for (int ivar=0; ivar<cvar; ivar++) {
			if (Math.abs(equation[ivar])>0.000001) cflow++;
		} // for
		int[] vri = new int[cflow];
		// Set indices
		int iflow =0;
		for (int ivar=0; ivar<cvar; ivar++) {
			if (Math.abs(equation[ivar])>0.000001) { vri[iflow] = ivar; iflow++; }
		} // for
		return vri;
	} // getFlowIndices

	/** Get the minimum index of the flows in the equation.
	* @return Min flow index.*/
	protected int getMinFlowIndex() {
		int[] vri = getFlowIndices();
		if (vri.length==0) return 0;
		else return vri[0];
	} // getMinFlowIndex

	/** Get the maximum index of the flows in the equation.
	* @return Max flow index.*/
	protected int getMaxFlowIndex() {
		int[] vri = getFlowIndices();
		if (vri.length==0) return 0;
		else return vri[vri.length-1];
	} // getMaxFlowIndex

	/** Get layer index.
	* @return Layer index. */
	public int getLayerIndex() { return ilayer; }

	/** Get node name corresponding to this equation.
	* @return Node name. */
	public String getNodeName() {
		if (node!=null)  {
			String szNodeName = node.getName();
			ECFModel parentmodel = node.getParentModel();
			if (parentmodel!=null) szNodeName = parentmodel.getName() + "/" + szNodeName;
			return szNodeName;
		}
		else return szNameNode; }

	/** Get number of variables in the left member of the equation.
	* @return Number of variables.*/
	public int getNoVar() {return cvar;}

	/** Get a description of equation type.
	* @return Description of equation type. */
	public String getType() { return szType; }

	/** Increment coefficient.
	* @param icoeff Index of coefficent.
	* @param n Increment added to coefficent ivar. */
	public void incCoefficient(int icoeff, double n) {
		equation[icoeff]+=n;
	} // setCoefficient

	public void setCoefficient(int ivar, double n) {
		equation[ivar]=n;
	} // setCoefficient

	/** Set layer index.
	* @param Layer index. */
	public void setLayerIndex(int ilayer) { this.ilayer = ilayer; }

	/** Set node name corresponding to this equation.
	* @param sxNameNode Node name. */
	public void setNode(ECFNode node) { this.node = node; }

	/** Set node name corresponding to this equation.
	* @param sxNameNode Node name. */
	public void setNodeName(String szNameNode) { this.szNameNode = szNameNode; }

	/** Set a description of equation type.
	* @param Description of equation type. */
	public void setType(String szType) { this.szType = szType; }

	public void setRightMember(double n) {
		equation[cvar]=n;
	} // setCoefficient

	public String toString() {
		String sz;
		sz = "Equ: [";
		if (cvar>0) {
			for (int ivar=0; ivar<cvar-1; ivar++) {
				sz += equation[ivar] + ", ";
			} // for
			sz += equation[cvar-1];
		} // if
		sz += " | ";
		sz += equation[cvar] + "]";
		return sz;
	} // toString

} // LinearEquation