import javax.swing.*;     // JFileChooser
import java.io.*;         // File
import java.util.*;		  // Date
import java.text.*;       // DateFormat

/** Exports model data to file in maple format.*/
public class MapleExporter extends Exporter {

	protected String szCommandStart = "> ";
	protected String szCommentSign = "# ";
	protected int cflow = 0;

	/** Creates a MapleExporter for writing model data to file.
	* @param frame The frame to which the dialog belongs.
	* @param szNameXMLFile The name of the XML model file.
	* @param szExtension File extension of exported file. */
	public MapleExporter(String szNameXMLFile, String szExtension) {
		super(szNameXMLFile,szExtension);
	} // MapleExporter

	/** Returns the comment sign for the specific file format.
	* @return Comment sign for Maple.*/
	protected String commentSign() { return szCommentSign;}

	/** Returns the index offset for vectors (0 or 1),
	* should be replaced in subclass. */
	protected int indexOffset() { return 1;}

	/** List the heading.
	* @return A string with the heading.*/
	public String listHeading() {
		String szExport = listHeading(szCommentSign);
		// Heading
		// Preamble
		szExport += szCommandStart + "restart;\n";
		szExport += szCommandStart + "with(LinearAlgebra);\n\n";
		return szExport;
	} // listHeading

	/** List the time histories of variables.
	* @return A string listing time histories.*/
	public String listTimeHistories() {
		// Heading
		String szExport = "";
		// Time axis
		if (timeaxis!=null)	szExport += timeaxis.exportMaple(szCommandStart);
		// Data vectors
		szExport += listVariableTimeHistories(vrflow,"Flow");
		szExport += listVariableTimeHistories(vrflowratio,"Ratio");
		szExport += listVariableTimeHistories(vrprice,"Price");
		szExport += listVariableTimeHistories(vrproductivity,"Productivity");
		return szExport;
	} // listTimeHistories

	/** List variable time histories.
	* @return A string with values and line feeds between each line. */
	private String listVariableTimeHistories(Vector vrvariable, String szType) {
		String szVariables = "";
		if (vrvariable==null)	return szVariables;
		int cvar = vrvariable.size();
		if (cvar==0)	return szVariables;

		// Vector sizes
		szVariables += "\n" + szCommandStart + "vrName" + szType + ":=Vector(" + cvar + ");\n";
		szVariables += szCommandStart + "vr" + szType + ":=Vector(" + cvar + ");\n";

		// Variables as a function of time
		int ivar = 1;
		Iterator iter = vrvariable.iterator();
		while (iter.hasNext()) {
			szVariables += ((Variable)iter.next()).exportMaple(szType,ivar,szCommandStart);
			ivar++;
		} // while
		return szVariables;
	} // listVariableTimeHistories

	/** List equations that were not applied.
	* @return A string with flow equations.*/
	protected String listEquationsNotApplied() {
		return listEquationsNotApplied(szCommentSign);
	} // listEquationsNotApplied

	public String listAtSelectedTime() {
		return listAtSelectedTime(szCommentSign);
	} // listAtSelectedTime

	/** Calculation formulas for Maple. */
	public String listCalculationFormulas() {
		System.out.println("MapleExporter.listCalculationFormulas: vrflow=" + vrflow);
		if (vrflow!=null) cflow = vrflow.size(); else return "";
		String szExport = "";

		// Maple calculations:
		// Print equation system
		szExport += "\n" + szCommentSign + "Print equation system:\n";
		szExport += szCommandStart + "cflow:="+cflow+";\n";
		szExport += szCommandStart +
					"for iflow from 1 to cflow do print(vrbRM[iflow]) od;\n";
		szExport += szCommandStart + "print(mxaSys);\n";

		// Calculate scaled flows
		szExport += "\n" + szCommentSign + "Calculate scaled flows:\n";
		szExport += szCommandStart + "mxaTemp:=Matrix(cflow,cflow);\n";
		szExport += szCommandStart +
					"for irow from 1 to cflow do for icol from 1 to cflow" +
					" do mxaTemp[irow,icol]:=(mxaSys[irow,icol]) od od;\n";
		szExport += szCommandStart + "mxaInv:=MatrixInverse(mxaTemp);\n";
		szExport += szCommandStart + "vrflowSol:=mxaInv.vrbRM;\n";
		szExport += szCommandStart + "for iflow from 1 to cflow do print(vrflowSol[iflow]);od;\n";

		// List flow scales
		Flow flow;
		int cdim = vrflow.size();
		szExport += "\n" + szCommentSign +"Flow scales:\n" +
			szCommandStart + "vrscaleFlow:=Vector(" + cdim + ",[";
		Iterator iter =vrflow.iterator();
		if (iter.hasNext()) {
			flow = (Flow)iter.next();
			szExport += flow.getScaleForCalc();
		} // if
		while (iter.hasNext()) {
			flow = (Flow)iter.next();
			szExport += "," + flow.getScaleForCalc();
		} // while

		// Calculate Flows in base units
		szExport += "]);\n" +
			szCommentSign + "Flows in base units:\n";
		szExport += szCommandStart + "for iflow from 1 to cflow " +
			"do vrflowBase[iflow]:=vrflowSol[iflow]*vrscaleFlow[iflow];od;\n";

		return szExport;
	} // listCalculationFormulas

	/** List constants and equation system in Maple format.
	* @return A string with Maple assignments for constants
	* and line feeds between each line. */
	protected String listEquationSystem() {
		String szEqSys = szCommentSign + "Equation system:\n";
		szEqSys += szCommentSign + "----------------\n\n";
		szEqSys += szCommentSign +"Assignments for constants, scaled values:\n";

		// Flow ratios as a function of time
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			if (flowratio.isPrescribed())
				szEqSys += flowratio.toStringCalcValueAndScaleMaple(szCommandStart);
		} // while

		// Prices as a function of time
		iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.isPrescribed())
				szEqSys += price.toStringCalcValueAndScaleMaple(szCommandStart);
		} // while

		// Productivities as a function of time
		iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
			if (productivity.isPrescribed())
				szEqSys += productivity.toStringCalcValueAndScaleMaple(szCommandStart);
		} // while

		// Maple format of the equation system
		if (eqsys!=null)
			szEqSys += "\n\n" + eqsys.toStringMaple(szCommentSign,szCommandStart);

		return szEqSys;
	} // listEquationSystem

} // MapleExporter