import javax.swing.*;     // JFileChooser
import java.io.*;         // File
import java.util.*;		  // Date
import java.text.*;       // DateFormat

/** Exports model data to file in Matlab format.
* applicable to Matlab 5.3 and higher.*/
public class MatlabExporter extends Exporter {

	protected String szCommandStart = "";
	protected String szCommentSign = "% ";

	/** Creates a MatlabExporter for writing model data to file.
	* @param frame The frame to which the dialog belongs.
	* @param szNameXMLFile The name of the XML model file.
	* @param szExtension File extension. */
	public MatlabExporter(String szNameXMLFile, String szExtension) {
		super(szNameXMLFile,szExtension);
	} // MatlabExporter

	/** Returns the comment sign for the specific file format,
	* should be replaced in subclass. */
//	protected String commentSign() { return "% ";}

	/** Returns the index offset for vectors (0 or 1),
	* should be replaced in subclass. */
	protected int indexOffset() { return 1;}

	/** List the time histories of variables.
	* @return A string listing time histories.*/
	public String listTimeHistories() {
		String szExport = "";
		// Time axis
		if (timeaxis!=null)	szExport += timeaxis.exportMatlab();
		// Data vectors
		szExport += listVariableTimeHistories(vrflow,"Flow");
		szExport += listVariableTimeHistories(vrflowratio,"Ratio");
		szExport += listVariableTimeHistories(vrprice,"Price");
		szExport += listVariableTimeHistories(vrproductivity,"Productivity");
		return szExport;
	} // listTimeHistories

	/** List variable time histories.
	* @return A string with values and line feeds between each line. */
	private String listVariableTimeHistories(Vector vrvariable, String szType) {
		// Variables as a function of time
		if (vrvariable==null) return "";
		String szVariables = "";
		int ivar = 1;
		Iterator iter = vrvariable.iterator();
		while (iter.hasNext()) {
			szVariables += ((Variable)iter.next()).exportMatlab(szType,ivar);
			ivar++;
		} // while
		return szVariables;
	} // listVariableTimeHistories

	/** List equations that were not applied.
	* @return A string with flow equations.*/
	protected String listEquationsNotApplied() {
		return listEquationsNotApplied(szCommentSign);
	} // listEquationsNotApplied

	public String listAtSelectedTime() {
		return listAtSelectedTime(szCommentSign);
	} // listAtSelectedTime

	/** Calculation formulas for Matlab. */
	public String listCalculationFormulas() {
		int cflow = 0;
//		System.out.println("MatlabExporter.listCalculationFormulas: vrflow=" + vrflow);
		if (vrflow!=null) cflow = vrflow.size(); else return "";
		String szExport = "";

		// Matlab calculations
		szExport += "\n"+ szCommentSign + "Flows:";
		szExport += "\nvrflowSol=mxaSys\\vrbRM';";

		// List flow scales
//		System.out.println("MatlabExporter.listCalculationFormulas: List flow scales");
		Flow flow;
		szExport += "\n" + szCommentSign +"Flow scales:\n" + "vrscaleFlow=[";
		Iterator iter =vrflow.iterator();
		if (iter.hasNext()) {
			flow = (Flow)iter.next();
			szExport += flow.getScaleForCalc();
		} // if
		while (iter.hasNext()) {
			flow = (Flow)iter.next();
			szExport += "," + flow.getScaleForCalc();
		} // while
		szExport += "];\n" + szCommentSign + "Flows in base units:\n";
		szExport += "vrflowBase = vrflowSol.*vrscaleFlow';";

//		System.out.println("MatlabExporter.listCalculationFormulas: end");
		return szExport;
	} //listCalculationFormulas


	/** List constants and equation system in Matlab format.
	* @return A string with Matlab assignments for constants
	* and line feeds between each line. */
	protected String listEquationSystem() {
		String szEqSys = szCommentSign + "Equation system:\n";
		szEqSys += szCommentSign + "----------------\n\n";
		szEqSys += szCommentSign +"Assignments for constants, scaled values:\n";

		// Flow ratios and scales as a function of time
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			if (flowratio.isPrescribed())
				szEqSys += flowratio.toStringCalcValueAndScaleMatlab();
		} // while

		// Prices and scales as a function of time
		iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.isPrescribed())
				szEqSys += price.toStringCalcValueAndScaleMatlab();
		} // while

		// Productivities and scales as a function of time
		iter = vrproductivity.iterator();
		while (iter.hasNext()) {
			Productivity productivity = (Productivity)iter.next();
			if (productivity.isPrescribed())
				szEqSys += productivity.toStringCalcValueAndScaleMatlab();
		} // while

		// Matlab format of the equation system
		if (eqsys!=null)
			szEqSys += "\n\n" + eqsys.toStringMatlab(szCommentSign);

		return szEqSys;
	} // listEquationSystem

} // MatlabExporter