import java.awt.*;
import java.util.*;
import javax.swing.*;

/**
* A model frame encloses a submodel of an economic system.
*/
public class ModelFrame extends ECFNode implements FlowPathHolder {

	/** The parent sector to which the submodel belongs. */
	private Sector sector;
	/** The heigth of the model frame.*/
	protected int heightFrame = 600;
	/** The width of the model frame.*/
	protected int widthFrame = 700;
	/** Label with submodel name. */
	protected GraphLabel lblNameSubModel;
	/** Label for closing (hiding) submodel. */
	protected GraphLabel lblClose;
	/** Frame connectors for each layer. */
	protected Vector vrFrameConnectors = new Vector();
	/** Vector of flow paths of frame. */
	protected Vector vrflowpath = new Vector();

	/**
	* Creates a new model frame.
	* @param model The submodel to which the frame belongs.
	* @param x,y coordinates of the upper left corner of the model frame.
	*/
	public ModelFrame(ECFModel model, Sector sector, int x, int y) {
		super(model,x,y,ECFModel.COLOR_TRANSPERANT);
		this.sector = sector;
		height = heightFrame;
		width = widthFrame;
		lblNameSubModel = new GraphLabel(model.getName(),x,y);
		lblClose = new GraphLabel("Close",x+width,y);
		lblClose.setBackgroundDefault(Color.white);
	} // ModelFrame

	/** Add external flows to the frame of the submodel.
	* @param vrNodeConnectors A vector of NodeConnectors,
	*        one item for each layer.*/
	public void addFlowsToTheFrame(Vector vrNodeConnectors) {
		for (int ilayer=0; ilayer<vrNodeConnectors.size(); ilayer++) {
			NodeConnectors conns = (NodeConnectors)vrNodeConnectors.elementAt(ilayer);
			Iterator iterFlowConn = conns.getFlowConnectorsIterator();
			while (iterFlowConn.hasNext()) {
				FlowConnector flowconn = (FlowConnector)iterFlowConn.next();
				FlowPath flowpath = flowconn.getFlowPath();
				flowpath.setSelectedConnector(flowconn.getEndOfFlow());
				connectFlowPath(ilayer,flowpath,100);
			} // while
		} // for
	} // addFlowsToTheFrame

	/** Add a layer with index ilayer to the model frame.
	* @param ilayer Index of the layer to be added. */
	public void addLayer(int ilayer) { ;
		vrFrameConnectors.add(new FrameConnectors(ilayer,this));
	} // addLayer

	/** Connect flow path to the frame.
	* @param ilayer Index of the layer of the flow path.
	* @param flowpath The flow path to be connected.
	* @param iconnNew Index of added connector.
	*/
	public void connectFlowPath(int ilayer, FlowPath flowpath, int iconnNew) {
		// Find node connectors
		Iterator iter = vrFrameConnectors.iterator();
		while (iter.hasNext()) {
			FrameConnectors conns = (FrameConnectors)iter.next();
			if (ilayer==conns.getLayerIndex()) {
				conns.connectFlowPath(flowpath,iconnNew);
				break;
			} // if
		} // while
	} // connectFlowPath

	/**
	* Check if the frame cointains the point (x,y).
	* @param ilayer not used, shown in all layers
	* @param x x-coordanate to be contained by the frame
	* @param y y-coordanate to be contained by the frame
	* @return true if the point (x,y) is contained, else false
	*/
	public boolean contains(int ilayer,int x,int y) {
		return getShape().contains(x,y);
	} // contains

	/**
	* Check if the Close label of the frame cointains the point (x,y).
	* @param ilayer not used, shown in all layers
	* @param x x-coordanate to be contained by the label
	* @param y y-coordanate to be contained by the label
	* @return true if the point (x,y) is contained, else false
	*/
	public boolean containsInCloseLabel(int ilayer,int x,int y) {
		return lblClose.getShape().contains(x,y);
	} // containsInCloseLabel

	/** Get the surrounding rectangle.
	* @return The surrounding rectangle. */
	public Rectangle getRect() {return new Rectangle(x,y,width,height); }

	/** Update fill color of node symbol and recalculate flows. */
	protected void updateNodeContext() {
		colorFill = ECFModel.COLOR_TRANSPERANT;
		makePainter();
		Calculator.calculate();
	} // updateNodeContext

	/**
	* Decode model frame from XML code.
	* @param szXMLCode The xml description of the model frame.
	*/
	public void decodeXML(String szXMLCode) {
//		System.out.println("ModelFrame.decodeXML: 1");
		XMLParser xml = null;
		int ilayer = 0;
		int clayer = vrFrameConnectors.size();
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				super.decodeXMLItems(szTag,szSub);
				// Override previous values
				height = heightFrame;
				width = widthFrame;
				if (szTag.equals("FrameConnectors") && ilayer<clayer) {
					FrameConnectors frameconns = (FrameConnectors)vrFrameConnectors.elementAt(ilayer);
					frameconns.decodeXML(szSub);
					ilayer++;
				} //
				szXMLCode = xml.remainder();
			} // while
			lblNameSubModel.setPos(x,y);
			lblClose.setPos(x+width,y);
		} catch(Exception e) {
			System.out.println("ModelFrame.decodeXML: " + e +
			"\n  " + szXMLCode);
		} // try/catch
//		System.out.println("ModelFrame.decodeXML: end");
	} // decodeXML

	/** Delete the model frame. */
	public void deleteModelFrame() {
//		System.out.println("ModelFrame.deleteModelFrame:");
	} // deleteModelFrame

	/**
	* Finds a flowpath of the submodel. An <code>FlowPath</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return FlowPath if a flowpath was found, else <code>null</code>.
	*/
	public FlowPath findFlowPath(int ilayer,int x,int y) {
		FrameConnectors frameconns = (FrameConnectors)vrFrameConnectors.elementAt(ilayer);
		FlowPath flowpath = frameconns.findFlowPath(ilayer,x,y);
		return flowpath;
	} // findFlowPath

	/** Get the flow path with index iflowpath from the side iside.
	* @param ilayer The index of the layer
	* @param iside The side of the frame. ECFNode.WEST, ECFNode.EAST ...
	* @param iflowpath The index of the flow path. */
	public FlowPath getFlowPath(int ilayer, int iside, int iflowpath) {
		FrameConnectors frameconns = (FrameConnectors)vrFrameConnectors.elementAt(ilayer);
		TwinFlowConnector twinconn = frameconns.getTwinFlowConnector(iside,iflowpath);
		return twinconn.getFlowPathInside();
	} // getFlowPath

	/** Get the index of a flowpath.
	* @param flowpath The flowpath for which the index is wanted.
	* @return The index of the flowpath. */
	public int getFlowPathIndex(FlowPath flowpath) {
		int iflowpath = vrflowpath.indexOf(flowpath);
		return iflowpath;
	} // getFlowPathIndex

	/**
	* Get the shape of the model frame.
	* @return the Shape of the frame
	*/
	public Shape getShape() {
		return new Rectangle(x,y,width,height);
	}

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors(int ilayer) {
		return null;
	} // getNodeConnectors

	/** Check if frame has a flow balance for this layer.
	* The total flow balance is set by the sector to which the frame
	* and submodel belongs.
	* @return false.*/
	public boolean hasFlowBalance(int ilayer) {
		return false;
	} // hasFlowBalance

	/** List XML code for the modelframe.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = super.listXML(szBlanks);
		Iterator iter = vrFrameConnectors.iterator();
		while (iter.hasNext()) {
			FrameConnectors frameconns = (FrameConnectors)iter.next();
			szXML += szBlanks + "<FrameConnectors>";
			szXML += frameconns.listXML("  " + szBlanks);
			szXML += szBlanks + "</FrameConnectors>";
		} // while
		return szXML;
	} // listXML

	/** Paint the model frame with connecting flows.
	* @param g2 The graphics context.
	* @param ilayer Index of the layer to be drawn. */
	public void paint(Graphics2D g2, int ilayer) {
		if (g2==null) return;
		super.paint(g2,ilayer);
		lblNameSubModel.paint(g2);
		lblClose.paint(g2);
		Iterator iter = vrFrameConnectors.iterator();
		while (iter.hasNext()) {
			FrameConnectors frameconns = (FrameConnectors)iter.next();
			frameconns.paint(g2,ilayer);
		} // while
	} // paint

	public void registerFlowPathOfFrame(FlowPath flowpath) {
		vrflowpath.add(flowpath);
	} // registerFlowPathOfFrame

	/** Remove outside flow from the model frame.
	* @param flow The flow to be removed; */
	public void removeFlow(Flow flow) {
		int ilayer = flow.getLayerIndex();
		if (vrFrameConnectors.size()>0) {
			FrameConnectors frameconns =
				(FrameConnectors)vrFrameConnectors.elementAt(ilayer);
			frameconns.removeFlow(flow);
		} // if
	} // removeFlow

	/** Set the name of the submodel.
	* @param szName The name to be set.
	*/
	public void setName(String szName) {
		super.setName(szName);
		lblNameSubModel.setText(szName);
	} // setName

	/** Show properties of the model frame.
	* @return A string telling the properties.*/
	public String toString() {
		return "ModelFrame: szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

} // ModelFrame
