import java.util.*;
import java.awt.*;
import java.awt.geom.*;

/** Connectors to FlowConnectors for one model layer.
* Each side of a node has a vector of connectors to flow connectors.
* Connector vectors are orientated in positive x and y directions
* = higher index means larger x or y values.
* <p>
* Flowpaths are connected to a node via this <code>NodeConnectors</code>.
* The metod <code>connectFlowPath(flowpath)</code> adds the selected
* <A href="FlowConnector.html">FlowConnector</A> to the closest side of
* the node and in the correct order before, between or after previous
* connectors. The reverse association is established by the
* <code>setAssociation(this)</code> of the <code>FlowConnector</code> class.
* <p>
* All flowpaths are disconnected by the <code>disconnectAllFlows()</code> method.
* A single flowpath is disconnected from a <code>FlowConnector</code> by calling
* the <code>undoAssociation(FlowConnector flowconn)</code> method of this class.
*/

public class NodeConnectors {

	/** Index of model layer. */
	protected int ilayer;
	/** Node to which this <code>NodeConnectors</code> belongs. */
	protected ECFNode node;
	/** Flag indicating if node has a flow balance in this layer.*/
	protected boolean fHasFlowBalance = false;
	/** Vector of flowconnectors attached to east side. */
	protected Vector vrconnEast;
	/** Vector of flowconnectors attached to west side. */
	protected Vector vrconnWest;
	/** Vector of flowconnectors attached to north side. */
	protected Vector vrconnNorth;
	/** Vector of flowconnectors attached to south side. */
	protected Vector vrconnSouth;

	/**
	* Creates <code>NodeConnectors</code> for layer <code>ilayer</code>
	* that connect the node <code>node</code> to flowpaths.
	* @param ilayer Index of the model layer to which this
	* <code>NodeConnectors</code> belongs.
	* @param node The node to which this <code>NodeConnectors</code> belongs.
	*/
	public NodeConnectors(int ilayer, ECFNode node) {
		this.ilayer = ilayer;
		this.node = node;
		vrconnEast = new Vector();
		vrconnWest = new Vector();
		vrconnNorth = new Vector();
		vrconnSouth = new Vector();
	} // NodeConnectors

	/**
	* Connect the <code>flowpath</code> to this flowconnectors.
	* @param flowpath The flowpath to be connected.
	*/
	public void connectFlowPath(FlowPath flowpath) {
		// Position of node symbol
		Rectangle rect = node.getRect();
		int xPos = (int)rect.getX();  // Position of upper left corner
		int yPos = (int)rect.getY();
		int height = (int)rect.getHeight();
		int width = (int)rect.getWidth();
		// Position of connector
		FlowConnector flowconn = flowpath.getSelectedConnector();
		if (flowconn==null) return;
		Point2D pnt = flowconn.getRefPoint();
		int x = (int)pnt.getX();
		int y = (int)pnt.getY();
		flowconn.setAssociation(this);

		// Diagonal lines from all corners,
		// x+y+c=0 (c=0 omitted because cancels when compared to flow
		// connection coordinates.
		int xyTopLeft = xPos - yPos;  // Diagonal line NW-SE
		int xyTopRight = xPos + width + yPos;  // Diagonal line NE-SW
		int xyBottomLeft = xPos + yPos + height;  // Diagonal line NE-SW
		int xyBottomRight = xPos + width - yPos - height;  // Diagonal line NW-SE

		if (x-y<=xyTopLeft && x+y<=xyBottomLeft && x<=xPos+width/2) {
			// add to West
			flowconn.setSide(ECFNode.WEST);
			addInYDirection(vrconnWest,flowconn);
		}
		else if (x-y>=xyBottomRight && x+y>=xyTopRight && x>xPos+width/2) {
			// add to East
			flowconn.setSide(ECFNode.EAST);  // Rev.1
			addInYDirection(vrconnEast,flowconn);  // Rev.1
		}
		else if (x-y>=xyTopLeft && x+y<=xyTopRight && y<=yPos+height/2) {
			// add to North
			flowconn.setSide(ECFNode.NORTH);
			addInXDirection(vrconnNorth,flowconn);
		}
		else {
			// add to South
			flowconn.setSide(ECFNode.SOUTH);
			addInXDirection(vrconnSouth,flowconn);
		}
		setPositions();
	} // connectFlowPath

	/**
	* Decode nodeconnectors from XML code.
	* @param szXMLCode The xml description of this nodeconnectors.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("ilayer")) ilayer = Integer.parseInt(szSub);
				if (szTag.equals("fHasFlowBalance")) {fHasFlowBalance = (0==szSub.compareTo("true"));}
				if (szTag.equals("West")) decodeXMLOneside(szSub,vrconnWest,ECFNode.WEST);
				if (szTag.equals("East")) decodeXMLOneside(szSub,vrconnEast,ECFNode.EAST);
				if (szTag.equals("North")) decodeXMLOneside(szSub,vrconnNorth,ECFNode.NORTH);
				if (szTag.equals("South")) decodeXMLOneside(szSub,vrconnSouth,ECFNode.SOUTH);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("NodeConnectors.decodeXML: " + e);
		} // try/catch
		setPositions();
	} // decodeXML

	/**
	* Decode nodeconnectors on one side from XML code.
	* @param szXMLCode The xml description of this nodeconnectors of this side.
	*/
	private void decodeXMLOneside(String szXMLCode, Vector vrconn, int iside) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("FlowConnector")) {
//					System.out.println("NodeConnectors.decodeXMLOneside: FlowConnector" +
//						" ilayer= " + ilayer + " iside=" + iside);
					decodeXMLFlowConnector(szSub,vrconn,iside);
				} // FlowConnector
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("NodeConnectors.decodeXMLOneside: " + e);
		} // try/catch
	} // decodeXMLOneside

	/**
	* Decode flowconnector from XML code.
	* @param szXMLCode The xml description of this flowconnector.
	* @param vrconn Vector of flow connectors on one side.
	* @param iside The side to which the the connector shall be connected.
	*/
	private void decodeXMLFlowConnector(String szXMLCode, Vector vrconn, int iside) {
		XMLParser xml = null;
		int iflowpath = 0;
		FlowPath flowpath = null;
		int isideOfFrame = 0;
		int iend = 1;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();

				if (szTag.equals("iend")) iend = Integer.parseInt(szSub);

				if (szTag.equals("flowpath")) {
					iflowpath = Integer.parseInt(szSub);
					flowpath = node.getFlowPath(iflowpath);
				} // if flowpath

				if (szTag.equals("FlowPathOfFrame")) {
					XMLParser xml2 = null;
					while (szSub.length()>0) {
						xml2 = new XMLParser(szSub);
							String szTag2 = xml2.tag();
							String szSub2 = xml2.sub();
							if (szTag2.equals("iflowpath")) iflowpath = Integer.parseInt(szSub2);
							if (szTag2.equals("isideOfFrame")) isideOfFrame = Integer.parseInt(szSub2);
						szSub = xml2.remainder();
					} // while
					SubModel modelParent = (SubModel)node.getParentModel();
					ModelFrame modelframe = modelParent.getModelFrame();
					flowpath = modelframe.getFlowPath(ilayer,isideOfFrame,iflowpath);
					if (iend==FlowPath.END_HEAD) {
						flowpath.setSide1(iside);
						flowpath.getFlowConnector(iend).setSide(iside);
					} // if HEAD
					else  { // TAIL
						flowpath.setSide2(iside);
						flowpath.getFlowConnector(iend).setSide(iside);
					} // else TAIL
				} // if FlowPathOfFrame

				szXMLCode = xml.remainder();
			} // while
			try {
				FlowConnector flowconn = flowpath.getFlowConnector(iend);
				vrconn.add(flowconn);
				flowconn.setAssociation(this);
			} catch(ArrayIndexOutOfBoundsException aie) {
				System.out.println("NodeConnectors.decodeXMLFlowConnector 1:" + aie);
			}
		} catch(Exception e) {
			System.out.println("NodeConnectors.decodeXMLFlowConnector 2:" + e);
		} // try/catch
	} // decodeXML

	/**
	* Disconnect all connected flows from the node.
	*/
	public void disconnectAllFlows() {
		// Disconnect all flow lines before deleting sector
		Iterator iter = new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			conn.undoAssociation();
		} // while
	} // disconnectAllFlowLines

	/** Get an iterator for all <a href="LinearEquation.html">
	* <code>LinearEquations</code></a> of the node and layer.
	* @return An iterator for all equations
	*/
	public Iterator getEquationsIterator() {
		Vector vrequations =new Vector();
		// Flow balance
		if (hasFlowBalance()) vrequations.add(getFlowBalanceEquation());
		return vrequations.iterator();
	} // getLayerEquationsIterator

	/** Get a flow balance equation for for the node and layer.
	* @return An equation for flow balance.
	*/
	public LinearEquation getFlowBalanceEquation() {
//		int cflows = Calculator.getNumberOfFlows();
		int cflows = ECF.vrflow.size();
		LinearEquation equation= new LinearEquation(cflows);  // Flows + right member
		Iterator iter = getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			setFlowCoefficient(equation,conn);
		} // while
		equation.setType("Flow Balance");
//		equation.setNodeName(node.getName());
		equation.setNode(node);
		equation.setLayerIndex(ilayer);
		return equation;
	} // getFlowBalanceEquation

	/** Get index of node layer for this connectors.
	* @return Index of the model layer to which this
	* <code>NodeConnectors</code> belongs.*/
	public int getLayerIndex() {return ilayer;}

	/** Get an iterator for all <a href="FlowConnector.html">
	* flow connectors (on all four sides).
	* @return A <a href="NodeConnectorsIterator.html">
	* NodeConnectorsIterator</a>.
	*/
	public NodeConnectorsIterator getFlowConnectorsIterator() {
		return new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
	} // getFlowConnectorsIterator

	/** Get an iterator for all node connectors (on all four sides).
	* @return A <a href="NodeConnectorsIterator.html">
	* NodeConnectorsIterator</a>.
	*/
	public NodeConnectorsIterator getNodeConnectorsIterator() {
		return new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
	} // getNodeConnectorsIterator

	/** Get number of flows that are connected with only one end.
	* @return An integer vector [cflows,cNotPrescr,cDefault] where <br>
	*  - cflows is number of connected flows, <br>
	*  - cNotPrescr is the number of these flows with not PRESCRIBED state. <br>
	*  - cDefault is the number of these flows with DEFAULT state.*/
	public int[] getNumberOfFlows() {
		LinearEquation eq = getFlowBalanceEquation();
		int[] vriflow = eq.getFlowIndices();
		int cflow = vriflow.length;
		int cNotPrescr = 0;
		int cDef = 0;
		for (int iflow=0; iflow<vriflow.length; iflow++) {
			if (!(Calculator.getFlow(vriflow[iflow])).isPrescribed()) cNotPrescr+=1;
			if ((Calculator.getFlow(vriflow[iflow])).isDefault()) cDef+=1;
		}
		int vrcFlow[] = new int[3];
		vrcFlow[0] = cflow; vrcFlow[1] = cNotPrescr; vrcFlow[2] = cDef;
		return vrcFlow;
	} // getNumberOfFlows

	/** Get a flow balance equation for for the node and layer.
	* @return An equation for flow balance.
	*/
	public Vector getNodeTypeEquations() {
		Vector vrequationNodeType = new Vector();
		int cflows = Calculator.getNumberOfFlows();
		LinearEquation equation= new LinearEquation(cflows);  // Flows + right member
		Iterator iter = new NodeConnectorsIterator(vrconnEast,vrconnWest,
	                       vrconnNorth,vrconnSouth);
/*		while (iter.hasNext()) {
			FlowCoefficient coeff = (FlowCoefficient)iter.next();
			setFlowCoefficient(equation,coeff);
		} // while
*/		equation.setType("Splitter Ratio");
		equation.setLayerIndex(ilayer);
		vrequationNodeType.add(equation);
		return vrequationNodeType;
	} // getNodeTypeEquations

	/** Check if node has a flow balance for this layer.
	* @return true if has flow balance, else false.*/
	public boolean hasFlowBalance() { return fHasFlowBalance; }

	/** List XML code for this nodeconnectors.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<ilayer>"+ilayer+"</ilayer>";
		szXML += "<fHasFlowBalance>"+fHasFlowBalance+"</fHasFlowBalance>";
		szXML += szBlanks +"<East>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnEast);
		szXML += "</East>";
		szXML += szBlanks +"<West>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnWest);
		szXML += "</West>";
		szXML += szBlanks +"<North>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnNorth);
		szXML += "</North>";
		szXML += szBlanks +"<South>";
		szXML += listXMLOneSide("  "+szBlanks,vrconnSouth);
		szXML += "</South>";
		return szXML;
	} // listXML

	/** List XML code for the flow connectors on one side this nodeconnectors.
	* @return XML code. */
	private String listXMLOneSide(String szBlanks1,Vector vrconn) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		Iterator iter = vrconn.iterator();
		while (iter.hasNext()) {
			FlowConnector flowconn = (FlowConnector)iter.next();
			szXML += "<FlowConnector>";
			szXML += flowconn.listXML("");
			szXML += "</FlowConnector>";
		} // while
		return szXML;
	} // listXMLOneSide

	/** Check if flow balance equation may be applied for this structure.
	* @return true if at least two flows are connected with one end only and
	* at least one of these flows has NOT PRESCRIBED state. */
	public boolean mayApplyFlowBalance() {
		int[] vrcFlow = getNumberOfFlows();
		return vrcFlow[0]>=2 && vrcFlow[1]>=1;
	} // mayApplyFlowBalance

	/** Check if flow balance equation may be applied at this stage when building
	* an equation system.
	* @return true if at least two flows are connected with one end only and
	* at least one of these flows has DEFAULT state. */
	public boolean mayApplyFlowBalanceNow() {
		int[] vrcFlow = getNumberOfFlows();
		return vrcFlow[0]>=2 && vrcFlow[2]>=1;
	} // mayApplyFlowBalance

	/** Paint flow paths.
	* @param g the graphics context.
	* @param ilayer index of the layer to which the flow paths belongs.
	*/
	public void paintFlowPaths(Graphics g ,int ilayer) {
		Iterator iterFlowConnector = getNodeConnectorsIterator();
		while (iterFlowConnector.hasNext()) {
			FlowConnector flowconn = (FlowConnector)iterFlowConnector.next();
			FlowPath flowpath = flowconn.getFlowPath();
			flowpath.paint(g,ilayer);
		} // while
	} // paintFlowPaths

	/** Set a coefficient in a flow balance equation. Inflow has the coefficent +1
	* and outflow has the coefficent -1.
	* @param equation The equation to be modified.
	* @param conn The flow connector of the flow.
	*/
	public void setFlowCoefficient(LinearEquation equation, FlowConnector conn) {
		int iflow = conn.getFlowPath().getFlow().getIndexOfFlow();
		int iendFlow = conn.getEndOfFlow();
		// Use increments to cancel flows that are connected with both ends to the same node.
		if (iendFlow==1) equation.incCoefficient(iflow,1); else equation.incCoefficient(iflow,-1);
	} // setFlowCoefficient

	/** Set if node has a flow balance for this layer.
	* @param f true if has flow balance, else false.*/
	public void setHasFlowBalance(boolean f) {fHasFlowBalance=f; }

	/**
	* Set positions of connected <A href="FlowConnector.html">FlowConnector</A>s.
	*/
	public void setPositions() {
		// Position of node symbol
		Rectangle rect = node.getRect();
		int xPos = (int)rect.getX();  // Position of upper left corner
		int yPos = (int)rect.getY();
		int height = (int)rect.getHeight();
		int width = (int)rect.getWidth();

		// East connectors
		int cconn = vrconnEast.size();
		if (cconn>0) {
			int dy = height/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				FlowConnector conn = (FlowConnector)vrconnEast.elementAt(iconn);
				conn.setPos(xPos+width,yPos+dy/2+iconn*dy);
			}  // for
		} // if
		// West connectors
		cconn = vrconnWest.size();
		if (cconn>0) {
			int dy = height/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				FlowConnector conn = (FlowConnector)vrconnWest.elementAt(iconn);
				conn.setPos(xPos,yPos+dy/2+iconn*dy);  // Rev.1
			}  // for
		} // if
		// North connectors
		cconn = vrconnNorth.size();
		if (cconn>0) {
			int dx = width/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				FlowConnector conn = (FlowConnector)vrconnNorth.elementAt(iconn);
				conn.setPos(xPos+dx/2+iconn*dx,yPos);
			}  // for
		} // if
		// South connectors
		cconn = vrconnSouth.size();
		if (cconn>0) {
			int dx = width/cconn;
			for (int iconn=0; iconn<cconn; iconn++) {
				FlowConnector conn = (FlowConnector)vrconnSouth.elementAt(iconn);
				conn.setPos(xPos+dx/2+iconn*dx,yPos+height);
			}  // for
		} // if
	} // setPositions

	/**
	* Undo association whith the <code>FlowConnector</code> flowconn.
	* Should only be called by the <code>disconnectNode()</code> method of
	* the <code>FlowConnector</code> class.
	* @param flowconn The <code>FlowConnector</code> to be disconnected.
	*/
	public void undoAssociation(FlowConnector flowconn) {
		// Remove flowconnector from any side,
		// side that does not contain flowconn is unaffected.
		vrconnEast.remove(flowconn);
		vrconnWest.remove(flowconn);
		vrconnNorth.remove(flowconn);
		vrconnSouth.remove(flowconn);
		setPositions();
		// Remove flow from node specific equations
		node.removeFlowReference(flowconn.getFlowPath());
		// Calculate new configuration
		Calculator.calculate();
	}; // undoAssociation

	private void addInXDirection(Vector vrconn,FlowConnector connNew) {
		// Insert new connector last if x >= x of all old connectors
		// else insert at position of first old connector that has x
		// greater than x of new connector
		if (vrconn.size()==0) {   // No old connectors
			vrconn.add(connNew);
			node.registerNewConnector(0);
		} else {				  // At least one old connector
			int xConnNew = (int)connNew.getRefPoint().getX();
			FlowConnector conn = (FlowConnector)vrconn.lastElement();
			int xConn = (int)conn.getRefPoint().getX();
			if (xConnNew>=xConn) {
				vrconn.add(connNew);
				int iconnNew = vrconn.size();
				node.registerNewConnector(iconnNew);
			} else {
				Iterator iter = vrconn.iterator();
				while (iter.hasNext()) {
					conn = (FlowConnector)iter.next();
					xConn = (int)conn.getRefPoint().getX();
					if (xConn>xConnNew) {
						int iconnNew = vrconn.indexOf(conn);
						vrconn.insertElementAt(connNew,iconnNew);
						node.registerNewConnector(iconnNew);
						break;
					} // if
				} // while
			} // if-else
		} // if
	} // addInXDirection

	private void addInYDirection(Vector vrconn,FlowConnector connNew) {
		// Insert new connector last if y >= y of all old connectors
		// else insert at position of first old connector that has y
		// greater than y of new connector
		if (vrconn.size()==0) {    // No old connector
			vrconn.add(connNew);
			node.registerNewConnector(0);
		} else {				  // At least one old connector
			int yConnNew = (int)connNew.getRefPoint().getY();
			FlowConnector conn = (FlowConnector)vrconn.lastElement();
			int yConn = (int)conn.getRefPoint().getY();
			if (yConnNew>=yConn) {
				vrconn.add(connNew);
				int iconnNew = vrconn.size();
				node.registerNewConnector(iconnNew);
			} else {
				Iterator iter = vrconn.iterator();
				while (iter.hasNext()) {
					conn = (FlowConnector)iter.next();
					yConn = (int)conn.getRefPoint().getY();
					if (yConn>yConnNew) {
						int iconnNew = vrconn.indexOf(conn);
						vrconn.insertElementAt(connNew,iconnNew);
						node.registerNewConnector(iconnNew);
						break;
					} // if
				} // while
			} // if-else
		} // if
	} // addInYDirection

} // NodeConnectors