import java.awt.*;
import java.util.*;

/**
* A <code>Price</code> belongs to a sector. The price is a
* proportionality constant between a flow of money and a flow
* of commodities. The two flows lie in different model layers
* and go in opposite directions. <p>
* Prescribed and calculated prices are stored as vectors in
* <A href="Variable.html">Variables</A>.
* The <code>get</code> and <code>set</code> methods operate on the
* the Variable value that corresponds to the current time increment.
*/
public class Price extends Variable implements FractionVariable {

	/** The sector to which the price belongs. */
	protected Sector sector;
	/** Commodity flow */
	protected Flow flowCommodity = null;
	/** Money flow */
	protected Flow flowMoney = null;
	/** Commodity unit. */
	protected Unit unitCommodity = null;
	/** Money unit. */
	protected Unit unitMoney = null;
	/** Vector of price units. */
	protected Units unitsPrice;

	/** Create an empty price object with no associated flows.
	* @param sector The sector to which the Price belongs.
	*/
	public Price(Sector sector) {
		super();
		this.sector = sector;
		szName = "Price";
		Calculator.addPrice(this);
	} // Price

	/** Create an empty price object with the name szName.
	* @param szName Name of ratio
	*/
	public Price(String szName) {
		super();
		this.szName = szName;
		Calculator.addPrice(this);
	} // Price

	/** Calculate price from flows. */
	public void calculate() {
		// Get flow rates
		try {
			double rateMoney = flowMoney.getBaseValueCalculated();
			double rateCommodity = flowCommodity.getBaseValueCalculated();
			// Calculate price
			double price = rateMoney/rateCommodity;
			setBaseValueCalculated(price);
		} catch(Exception e) {
			// System.out.println("Price.calculate: " + e);
			setBaseValueCalculated(0.0);
		} // try/catch
	} // calculate

	/**
	* Decode price from XML code.
	* @param szXMLCode The xml description of the price.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("flowindexMoney")) {
					int iflow = Integer.parseInt(szSub);
					setFlowMoney(Calculator.getFlow(iflow));
				} // if
				if (szTag.equals("flowindexCommodity")) {
					int iflow = Integer.parseInt(szSub);
					setFlowCommodity(Calculator.getFlow(iflow));
				} // if
				if (szTag.equals("Variable")) super.decodeXML(szSub);
				//if (szTag.equals("unitsPrice")) unitsPrice.decodeXML(szSub);
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println(e);
		} // try/catch
	} // decodeXML

	/** Export price values.
	* @return A string with label and values separated by semicolon (;).
	*/
	public String export() {
		String szExport = ECFUtil.removeLineFeeds(szName) + ";";
		szExport = Quantity.padToLength(szExport,20);
		double[] vrvalueDisplay = getDisplayValuesCalculated();
		for (int ivalue=0; ivalue<vrvalueDisplay.length; ivalue++)
			szExport += " " + Quantity.formatDouble(vrvalueDisplay[ivalue]) + ";";
		szExport += "\n";
		return szExport;
	} // export

	/** Get prescribed price value for equations.
	* @return Prescribed value scaled for calculations. */
	public double getCalcValuePrescribed() {
		return getBaseValuePrescribed()/getScaleForCalc();
	} // getCalcValuePrescribed

	/** Get Commodity Flow.
	* @return Commodity Flow object. */
	public Flow getFlowDenominator() { return flowCommodity; }
	public Flow getFlowCommodity() { return flowCommodity; }

	/** Get Money Flow.
	* @return Money Flow object. */
	public Flow getFlowNominator() { return flowMoney; }
	public Flow getFlowMoney() { return flowMoney; }

	/** Get Commodity Unit.
	* @return Commodity Unit object. */
	public Unit getUnitNominator() { return unitCommodity; }
	public Unit getUnitCommodity() { return unitCommodity; }

	/** Get Money Unit.
	* @return Money Unit object. */
	public Unit getUnitDenominator() { return unitMoney; }
	public Unit getUnitMoney() { return unitMoney; }

	/** Get a price equation.
	* @return An equation for money flow in relation to commodity flow.
	*/
	public LinearEquation getPriceEquation() throws Exception {
		// System.out.println("Price.getPriceEquation: ");
		if (flowMoney==null || flowCommodity==null) return null;
		// Make empty equation
		int cflows = Calculator.getNumberOfFlows();
		LinearEquation equation= new LinearEquation(cflows);  // Flows + right member
		// Set coefficient for commodity flow.
		double price = getCalcValuePrescribed();
		if (price==0.0) throw new Exception("Price " + getName() +
					" of sector " + sector.getName() + " is zero");
		equation.setCoefficient(flowCommodity.getIndexOfFlow(),price);
		// Set coefficient for money flow
		equation.setCoefficient(flowMoney.getIndexOfFlow(),-1);
		// Set equation descriptive data.
		equation.setType("Price '" + getName() + "'");
//		equation.setNodeName(sector.getName());
		equation.setNode(sector);
		equation.setLayerIndex(flowCommodity.getLayerIndex());
		//
		return equation;
	} // getSplitterRatioEquation

	/** Get the scaling factor for setting prices in equation systems.
	* Price in calculations = Base price / scale.
	* Base price = Price in calculations * scale.
	* @return Scaling factor for prices.*/
	public double getScaleForCalc() {
		double scale;
		try { scale=flowMoney.getUnits().getScaleForCalc()/
				flowCommodity.getUnits().getScaleForCalc();
		} catch(Exception e) { scale =1.0; }
		return scale;
	} // getScaleForCalc

	/** Check if the price has the <code>flow</code> as eiter money
	* flow or commodity flow.
	* @param flow The flow to be checked.
	* @return true if the <code>flow</code> belongs to the Price, else false.*/
	public boolean hasFlow(Flow flow) {
		return flow==flowMoney || flow==flowCommodity;
	} // hasFlow

	/** List XML code for the price.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<flowindexMoney>"+flowMoney.getIndexOfFlow()+"</flowindexMoney>";
		szXML += szBlanks +"<flowindexCommodity>"+flowCommodity.getIndexOfFlow()+"</flowindexCommodity>";
		szXML += szBlanks +"<Variable>"+ super.listXML("")
		 		+"</Variable>";
		szXML += szBlanks +"<unitsPrice>"+unitsPrice.listXML("  "+szBlanks)
				+szBlanks +"</unitsPrice>";
		return szXML;
	} // listXML

	/**
	* Paint price label at connector for flow path. The label consists
	* of the letter P and a background rectangle if the price has
	* prescribed values.
	* @param g2 graphics context.
	* @param ilayer index of layer to be painted.
	*/
	public void paintLabel(Graphics2D g2,int ilayer) {
		if (flowMoney.getIndexOfFlow()==ilayer) {;}
		if (flowCommodity.getIndexOfFlow()==ilayer) {;}

	} // paintLabel

	/** Set Commodity Flow.
	* @param Commodity Flow object. */
	public void setFlowCommodity(Flow flowCommodity) throws NoUnitException {
		this.flowCommodity = flowCommodity;
		Units units = flowCommodity.getUnits();
		if (units==null) throw new NoUnitException(flowCommodity.getName());
		unitCommodity = units.getDisplayUnit();
		setUnitPrice();
	} // setFlowCommodity

	/** Set Money Flow.
	* @param Money Flow object. */
	public void setFlowMoney(Flow flowMoney) throws NoUnitException {
		this.flowMoney = flowMoney;
		Units units = flowMoney.getUnits();
		if (units==null) throw new NoUnitException(flowMoney.getName());
		unitMoney = units.getDisplayUnit();
		setUnitPrice();
	} // setFlowMoney

	/** Set <a href="Unit.html">Unit</a> of the price. */
	protected void setUnitPrice() {
		String szUnitPrice;
		double factor;
		if (unitMoney!=null && unitCommodity!=null) {
			szUnitPrice =  unitMoney.getName() + "/" + unitCommodity.getName();
			factor = unitMoney.getFactor() / unitCommodity.getFactor();
		} else {
			szUnitPrice =  "No price unit set";
			factor = 1.0;
		} // if/else
		Unit unitPrice = new Unit(szUnitPrice,factor,"Price unit");
		unitsPrice = new Units ("Price", unitPrice);
		setUnits(unitsPrice);
	} // setUnitPrice

	/** Set Money Unit.
	* @param Index of money unit. */
	public void setUnitMoney(int iunitMoney) {setUnitDenominator(iunitMoney);}
	public void setUnitDenominator(int iunitMoney) {
		Vector vrunit = flowMoney.getUnits().getUnits();
		unitMoney = (Unit)vrunit.elementAt(iunitMoney);
		setUnitPrice();
	} // setUnitDenominator

	/** Set Commodity Unit.
	* @param Index of commodity unit. */
	public void setUnitCommodity(int iunitCommodity) {setUnitNominator(iunitCommodity);}
	public void setUnitNominator(int iunitCommodity) {
		Vector vrunit = flowCommodity.getUnits().getUnits();
		unitCommodity = (Unit)vrunit.elementAt(iunitCommodity);
		setUnitPrice();
	} // setUnitNominator

	/** Returns a string representation of the price.*/
	public String toString() {
		String sz = "Price: "+ szName;
		if (flowMoney!=null) sz += " flowMoney=" + flowMoney.getName();
		if (flowCommodity!=null) sz += " flowCommodity=" + flowCommodity.getName();
		sz += ", " + super.toString();
		return sz;
	} // toString

} // Price