import java.util.*;

/**
* A <code>Productivity</code> between an outflow and an inflow of resources
* of different kinds (units).<p>
* Prescribed and calculated ratios are stored as vectors in
* <A href="Variable.html">Variables</A>.
* The <code>get</code> and <code>set</code> methods operate on the
* the Variable value that corresponds to the current time increment.
*/
public class Productivity extends Variable implements FractionVariable {

	/** */
	protected ECFProcess process;
	/** Index of productivity factor. */
	protected Flow flowIn = null;
	/** Flow out */
	protected Flow flowOut = null;
	/** Flow in unit. */
	protected Unit unitFlowIn = null;
	/** Flow out unit. */
	protected Unit unitFlowOut = null;
	/** Vector of productivity units. */
	protected Units unitsProductivity;

	/** Create a ratio of resources with ilayer = 0.
	* @param timeaxis The time axis against which the Productivity is plotted.
	*/
	public Productivity(ECFProcess process) {
		super();
		this.process = process;
		szName = "Productivity";
		Calculator.addProductivity(this);
	} // Productivity

	/** Calculate productivities from flows. */
	public void calculate() {
		double factor;

		if (flowIn==null || flowOut==null) return;
		// Calculate productivity factor
		try  {
			factor = flowOut.getBaseValueCalculated()/flowIn.getBaseValueCalculated();
		}
		catch(Exception e) {
			System.out.println("Productivity.calculate:" + e);
			factor = 0.0;
		}
		setBaseValueCalculated(factor);
	} // calculate

	/**
	* Decode productivity from XML code.
	* @param szXMLCode The xml description of the produvtivity.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("flowindexIn")) {
					int iflow = Integer.parseInt(szSub);
					setFlowIn(Calculator.getFlow(iflow));
				} // if
				if (szTag.equals("flowindexOut")) {
					int iflow = Integer.parseInt(szSub);
					setFlowOut(Calculator.getFlow(iflow));
				} // if
				if (szTag.equals("Variable")) super.decodeXML(szSub);
				szXMLCode = xml.remainder();
			} // while
			setUnitProductivity();
		} catch(Exception e) {
			System.out.println(e);
		} // try/catch
	} // decodeXML

	/** Export Productivity values.
	* @return A string with  label and values separated by semicolon (;).
	*/
	public String export() {
		String szExport = ECFUtil.removeLineFeeds(szName) + ";";
		szExport = Quantity.padToLength(szExport,20);
		double[] vrvalueDisplay = getDisplayValuesCalculated();
		for (int ivalue=0; ivalue<vrvalueDisplay.length; ivalue++)
			szExport += " " + Quantity.formatDouble(vrvalueDisplay[ivalue]) + ";";
		szExport += "\n";
		return szExport;
	} // export

	/** Get prescribed productivity value for equations.
	* @return Prescribed value scaled for calculations. */
	public double getCalcValuePrescribed() {
		return getBaseValuePrescribed()/getScaleForCalc();
	} // getCalcValuePrescribed

	/** Get Flow in.
	* @return Flow in. */
	public Flow getFlowDenominator() { return flowIn; }

	/** Get Flow out.
	* @return Flow out object . */
	public Flow getFlowNominator() { return flowOut; }

	/** Get the scaling factor for setting productivities in equation systems.
	* Productivity in calculations = Base productivity / scale.
	* Base productivity = Productivity in calculations * scale.
	* @return Scaling factor for productivities.*/
	public double getScaleForCalc() {
		double scale;
		try { scale=flowOut.getUnits().getScaleForCalc()/
				flowIn.getUnits().getScaleForCalc();
		} catch(Exception e) { scale =1.0; }
		return scale;
	} // getScaleForCalc

	/** Get Flow in Unit.
	* @return Flow in Unit object. */
	public Unit getUnitDenominator() { return unitFlowIn; }

	/** Get Flow out Unit.
	* @return Flow out Unit object. */
	public Unit getUnitNominator() { return unitFlowOut; }

	/** List XML code for the produvtivity.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<flowindexIn>"+flowIn.getIndexOfFlow()+"</flowindexIn>";
		szXML += szBlanks +"<flowindexOut>"+flowOut.getIndexOfFlow()+"</flowindexOut>";
		szXML += szBlanks +"<Variable>"+ super.listXML("  "+szBlanks)
		 		+szBlanks +"</Variable>";
		return szXML;
	} // listXML

	/** Set Flow in
	* @param Flow in object. */
	public void setFlowIn(Flow flowIn) throws NoUnitException {
		this.flowIn = flowIn;
		Units units = flowIn.getUnits();
		if (units==null) throw new NoUnitException(flowIn.getName());
		unitFlowIn = units.getDisplayUnit();
		setUnitProductivity();
	} // setFlowIn

	/** Set Flow out.
	* @param Flow out object. */
	public void setFlowOut(Flow flowOut) throws NoUnitException {
		this.flowOut = flowOut;
		Units units = flowOut.getUnits();
		if (units==null) throw new NoUnitException(flowOut.getName());
		unitFlowOut = units.getDisplayUnit();
		setUnitProductivity();
	} // setFlowOut

	/** Set <a href="Unit.html">Unit</a> of the productivity. */
	protected void setUnitProductivity() {
		String szUnitProductivity;
		double factor;
		if (unitFlowIn!=null && unitFlowOut!=null) {
			szUnitProductivity =  unitFlowOut.getName() + "/" + unitFlowIn.getName();
			factor = unitFlowIn.getFactor() / unitFlowOut.getFactor();
		} else {
			szUnitProductivity =  "No productivity unit set";
			factor = 1.0;
		} // if/else
		Unit unitProductivity = new Unit(szUnitProductivity,factor,"Productivity unit");
		unitsProductivity = new Units ("Productivity", unitProductivity);
		setUnits(unitsProductivity);
	} // setUnitProductivity


	/** Set Flow in Unit.
	* @param Index of flow in unit. */
	public void setUnitDenominator(int iunitFlowIn) {
		Vector vrunit = flowIn.getUnits().getUnits();
		unitFlowIn = (Unit)vrunit.elementAt(iunitFlowIn);
		setUnitProductivity();
	} // setUnitDenominator

	/** Set Flow out Unit.
	* @param Index of flow out unit. */
	public void setUnitNominator(int iunitFlowOut) {
		Vector vrunit = flowOut.getUnits().getUnits();
		unitFlowOut = (Unit)vrunit.elementAt(iunitFlowOut);
		setUnitProductivity();
	} // setUnitNominator

	/** Returns a string representation of the Productivity.*/
	public String toString() {
		String sz = "Productivity: " + getName() + " flowIn=";
		if (flowIn!=null) sz += flowIn.getName();
		else sz += "not defined";
		sz += " flowOut=";
		if (flowOut!=null) sz += flowOut.getName();
		else sz += "not defined";
		sz += ", " + super.toString();
		return sz;
	} // toString

} // Productivity