import java.util.*;
import java.text.*;   // NumberFormat

/**
* A Quantity has a value and a set of units. The base value
* the value used in calculations. The display value is the
* value used in text and diagrams and depends upon which units
* from the set of units that was selected.
*/
public class Quantity {

	public static String formatDouble(double value) {
		NumberFormat numfmt = NumberFormat.getInstance(Locale.UK);
		int cdec = 0;
		double valueAbs = Math.abs(value);
		if (valueAbs<=1000.0) cdec = 1;
		if (valueAbs<=100.0) cdec = 2;
		if (valueAbs<=10.0) cdec = 3;
		if (valueAbs<=1) cdec = 4;
		if (valueAbs<=0) cdec = 5;
		if (valueAbs<=0.1) cdec = 6;
		if (valueAbs<=0.01) cdec = 7;
		numfmt.setMaximumFractionDigits(cdec);
		return numfmt.format(value);
	} // DoubleFormat

	public static double parseDouble(String szDouble) throws ParseException {
		NumberFormat numfmt = NumberFormat.getInstance(Locale.UK);
		Number num = numfmt.parse(szDouble);
		double value = num.doubleValue();
		return value;
	} // parseDouble

	/** Pad a string with blanks to a desired length.
	* @param sz The string to be padded.
	* @param cchar The final number of characters.
	* @return A string with trailing blanks and desired length.*/
	public static String padToLength(String sz, int cchar) {
		while (sz.length()<cchar) sz += " ";
		return sz;
	} // padToLength

	/** The Units of the Quantity.*/
	protected Units units = null;
	/** The base value. */
	protected double value = 0;
	/** The name. */
	protected String szName = "quant";

	/** Creates a Quantity object with units.
	* @param units The Units of the quantity type.*/
	public Quantity() {
	} // Quantity

	/** Creates a Quantity object with units.
	* @param units The Units of the quantity type.*/
	public Quantity(Units units) {
		this.units = units;
		szName = "quant";
	} // Quantity

	/** Remove units from flows etc. */
	public void clearUnits() { units = null; }

	/** Get the base value of the Quantity. */
	private double getBaseValue() { return value; }

	/** Get the pixel value of the Quantity. */
	public int getPixelValue(int cpixel) {
		if (units==null) return 0;
		return (int)(getBaseValue()*units.getScaleToPixels(cpixel));
	} // getPixelValue

	/** Get the name. */
	public String getName() { return szName; }

	/** Get the string with value and display unit.
	* @return A string with decimal number and unit, e.g. "1.3 MCU"
	*/
	public String getStringQuant() {
		if (units==null) return "";
		return formatDouble(getValue()) + " " +
				units.getNameOfDisplayUnit();
		} // getStringQuant

	/** Get the units of the quantity.
	* @return The units of the quantity.*/
	public Units getUnits() {
		return units;
	}

	/** Get the display value. */
	private double getValue() {
		if (units!=null)
			return units.getFactor()*value;
		else
			return 0.0;
		}

	/** Set the base value. */
	private void setBaseValue(double value) { this.value =value; }

	/** Set the display unit.
	* @param szUnit The name of the display unit.*/
	public void setDisplayUnit(String szUnit) {
		units.setDisplayUnit(szUnit);
	} // setDisplayUnit

	/** Set the name of the quantity.
	* @param The name of the quantity. */
	public void setName(String szName) { this.szName = szName; }

	/** Set the units of the quantity.
	* @param units The units of the quantity.*/
	public void setUnits(Units units) { this.units = units;	}

	/** Set the units if not previously set.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(Units units) {
			if (this.units==null) this.units=units;
	} // setUnitsIfNotSet

	/** Set value depending upon selected unit. */
	private void setValue(double value) {
		this.value = value/units.getFactor();
	} //setValue

	/** Get string representation of the Quantity. */
	public String toString() {
		return "Quantity using " + units.getName();
	} // toString

	/** Test the Quantity, Units and Unit classes. */
	public static void main(String[] args) {

		// Make units
		Vector vrunits = Units.generateStandardUnits();
		Units units = (Units)vrunits.elementAt(0);

		// Make Quantity objects
		Vector vrquant = new Vector();
		Quantity quant = new Quantity(units);
		quant.setName("Housing");
		quant.setValue(1400.0);
		vrquant.add(quant);
		quant = new Quantity(units);
		quant.setName("Transportation");
		quant.setValue(230.0);
		vrquant.add(quant);
		quant = new Quantity(units);
		quant.setName("Food");
		quant.setValue(370.0);
		vrquant.add(quant);

		System.out.println("Test Quantity, Units and Unit");
		QuantityFrame frame = new QuantityFrame("Test Quantity",units,vrquant);

	} // main

} // Quantity