import java.awt.*;
import java.util.*;
import javax.swing.*;

/**
* A sector of the economic system. A sector is a node that represents an actor
* of some kind in the system and may have complex inner properties.
*/
public class Sector extends ECFNode {

	/** Tolerance for cursor postion inside edges of sector symbol for resizing*/
	protected int tol = 10;
	/** Start height of rectangle enclosing node symbol before resizing. */
	protected int heightStart;
	/** Start width of rectangle enclosing node symbol before resizing. */
	protected int widthStart;
	/** Side to move when resizing. */
	protected int szToMove = NONE; // EAST, WEST, NORTH, SOUTH
	/** Index of selected money layer */
	protected int ilayerMoneySelected = 0;
	/** Index of selected commodity layer */
	protected int ilayerCommoditySelected = 1;
	/** A vector of prive relations. */
	protected Vector vrprice = new Vector();
	/** The time axis along which the prices are varied. */
	protected TimeAxis timeaxis;
	/** A vector of price equations. */
	protected Vector vrequationPrice = new Vector();
	/** Label for opening (showing) submodel. */
	protected GraphLabel lblOpen;
	/** Flag indicating if open label (lblOpen) was hit. */
	protected boolean fOpenLabelHit = false;
	/** Submodel inside the node.*/
	protected SubModel submodel = null;
	/** Index of added connector when connecting a new flow path
	* by the connectFlowPath method. */
	protected int iconnNew = 0;
	/** Index of sector. */
	static int isector = 1;
	/** Reset sector index. */
	public static void resetIndex() { isector = 1; }

	/**
	* Creates a new sector.
	* @param model The parent model to which the sector belongs.
	* @param x,y Coordinates of the upper left corner of the sector symbol
	*/
	public Sector(ECFModel model, int x, int y) {
		super(model,x,y);
		this.timeaxis = ECF.getTimeAxis();
		this.clayer = ECF.getNoOfLayers();
		setName("Sector"+isector);
		lblOpen = new GraphLabel("Open",x+width,y);
		lblOpen.setBackgroundDefault(Color.white);
		submodel = new SubModel(this);
		for (int ilayer=0; ilayer<clayer; ilayer++)
			addLayer(ilayer);
		isector++;
	} // Sector

	/** Add a layer with index ilayer to this sector.
	* @param ilayer Index of the layer to be added. */
	public void addLayer(int ilayer) {
		vrNodeConnectors.add(new NodeConnectors(ilayer,this));
		if (submodel!=null) submodel.addLayer(ilayer);
	} // addLayer

	/** Remove units from flows etc. of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		// Clear units of sub model of sector
		if (submodel!=null) submodel.clearUnits(ilayer);
	} // clearUnits

	/** Connect flow path to this node.
	* @param ilayer Index of the layer of the flow path
	* @param flowpath the flow path to be connected. */
	public void connectFlowPath(int ilayer, FlowPath flowpath) {
		super.connectFlowPath(ilayer,flowpath);
		submodel.connectFlowPath(ilayer,flowpath,iconnNew);
		// Update fill color and recalculate
		updateNodeContext();
	} // connectFlowPath

	/**
	* Check if the sector symbol cointains the point (x,y).
	* @param ilayer not used, shown in all layers
	* @param x x-coordanate to be contained by the sector symbol
	* @param y y-coordanate to be contained by the sector symbol
	* @return true if the point (x,y) is contained, else false
	*/
	public boolean contains(int ilayer,int x,int y) {
		boolean fLabel = containsInOpenLabel(ilayer,x,y);
		if (fLabel) fOpenLabelHit = true;
		else fOpenLabelHit = false;
		return getShape().contains(x,y) || fLabel;
	} // contains

	/**
	* Check if the Open label of the sector cointains the point (x,y).
	* @param ilayer not used, shown in all layers
	* @param x x-coordanate to be contained by the label
	* @param y y-coordanate to be contained by the label
	* @return true if the point (x,y) is contained, else false
	*/
	public boolean containsInOpenLabel(int ilayer,int x,int y) {
		return hasSubModel() && lblOpen.getShape().contains(x,y);
	} // containsInOpenLabel

	/**
	* Decode junction items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		// Decode ECFNode items
		super.decodeXMLItems(szTag,szSub);
		// Decode special sector items
		if (szTag.equals("ilayerMoneySelected")) {
			ilayerMoneySelected = Integer.parseInt(szSub);}
		if (szTag.equals("ilayerCommoditySelected")) {
			ilayerCommoditySelected = Integer.parseInt(szSub);}
		if (szTag.equals("Price")||szTag.equals("price")) {
			Price price = new Price(this);
			price.decodeXML(szSub);
			vrprice.add(price);
		} // if
		if (szTag.equals("submodel")) {
			ECF.modelOpen = submodel;
			submodel.decodeXML(szSub);
			ECF.modelOpen = modelParent;
		} // if
		lblOpen.setPos(x+width,y);
	} // decodeXMLItems

	/**
	* Decode node connectors from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLNodeConnectors(String szXMLCode) {
		// Decode external node connectors of the sector (ECFNode)
		super.decodeXMLNodeConnectors(szXMLCode);
		// Add the same flows to the frame of the submodel.
		if (submodel!=null) {
			submodel.addFlowsToTheFrame(vrNodeConnectors);
		} // if submodel
	} // decodeXML

	/** Delete the submodel of the node. */
	public void deleteSubmodel() {
		// Find and remove price references
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			timeaxis.remove(price);
			Calculator.removePrice(price);
		} // while
		vrprice = new Vector();
		submodel.deleteSubmodel();
		submodel = null;
	} // deleteSubmodel

	/** Get index of selected money layer.
	* @return Index of selected money layer. */
	public int getIndexOfSelectedMoneyLayer() { return ilayerMoneySelected; }

	/** Get index of selected commodity layer.
	* @return Index of selected commodity layer. */
	public int getIndexOfSelectedCommodityLayer() { return ilayerCommoditySelected; }

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors(int ilayer) {
		try {
			return (NodeConnectors)vrNodeConnectors.elementAt(ilayer);
		} catch(Exception e) { return null;}
	} // getNodeConnectors

	/** Get node specific equations for price relations,.
	* @return A vector of <a href="LinearEquation.html">
	*         <code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() throws Exception {
		Vector vrequations = new Vector();
		// Price relations
		Iterator iterPrice = vrprice.iterator();
		while (iterPrice.hasNext()) {
			Price price = (Price)iterPrice.next();
			if (price.isPrescribed()) {
				LinearEquation equ = price.getPriceEquation();
				if (equ!=null) vrequations.add(equ);
			} // if
		} // while
		return vrequations;
	} // getNodeSpecificEquations

	/** Get price relations.
	* @return A vector of Price objects. */
	public Vector getPriceRelations() { return vrprice; }

	/** Get a double arrow resize cursor directed up/down for horizontal sides
	* and left/right for vertical sides.
	* @param xP,yP Pointer coordinates (pixels). */
	public Cursor getResizeCursor(int xP, int yP) {
		Cursor cursor = super.getResizeCursor(xP,yP);
		int x0 = x; int x1 = x0 + width; // x-coord for left and right sides
		int y0 = y; int y1 = y0 + height; // y-coord for top and botton sides
		if (xP>=x0 && xP<x0+tol)      szToMove = WEST;    // Rev.1
		else if (xP>x1-tol && xP<=x1) szToMove = EAST;    // Rev.1
		else if (yP>=y0 && yP<y0+tol) szToMove = NORTH;
		else if (yP>y1-tol && yP<=y1) szToMove = SOUTH;
		else szToMove = NONE;
		if ((szToMove==WEST)  || szToMove==EAST)  cursor = new Cursor(Cursor.E_RESIZE_CURSOR);
		if ((szToMove==NORTH) || szToMove==SOUTH) cursor = new Cursor(Cursor.S_RESIZE_CURSOR);
		return cursor;
	} // showResizeCursor

	/**
	* Get the shape of the sector symbol.
	* @return the Shape of the symbol
	*/
	public Shape getShape() {
		return new Rectangle(x,y,width,height);
	} // getShape

	/**
	* Get the submodel of the sector or null.
	* @return the SubModel of the sector if any, else null.
	*/
	public SubModel getSubModel() {
		return submodel;
	}

	/** Check if node has a flow balance for this layer.
	* @return true if has flow balance, else false.*/
	public boolean hasFlowBalance(int ilayer) {
		try { return ((NodeConnectors)vrNodeConnectors.elementAt(ilayer)).hasFlowBalance();
		} catch(Exception e) { return false; }
	} // hasFlowBalance

	/** Check if node has a sub model.
	* @return true if has a sub model, else false.*/
	public boolean hasSubModel() {
		if (submodel.vrnode.size()==0) return false;
		else return true;
	} // hasSubModel

	/** List XML code for the sector.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = super.listXML(szBlanks);
		szXML += szBlanks +"<ilayerMoneySelected>" +
			ilayerMoneySelected +"</ilayerMoneySelected>";
		szXML += szBlanks +"<ilayerCommoditySelected>" +
			ilayerCommoditySelected +"</ilayerCommoditySelected>";
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			szXML += szBlanks +"<price>";
			szXML += price.listXML("  "+szBlanks);
			szXML += szBlanks +"</price>";
		} // while
		if (submodel!=null) {
			szXML += szBlanks +"<submodel>";
			szXML += submodel.listXML("  "+szBlanks);
			szXML += szBlanks +"</submodel>";
		} // if submodel
		return szXML;
	} // listXML

	/** Set resize side to NONE (applies to Sector). */
	public void quitResize() {
		szToMove=NONE;
	}

	/** Open a dialog window for node properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		JFrame mainframe = ECF.frameMain;
		int x = mainframe.getX();
		int y = mainframe.getY();
		SectorDialog dialog = new SectorDialog(this, new Point(x+20,y+20));
		dialog.show();
	} // openDialog

	/** Open submodel.*/
	public void openSubModel() {
		ECF.openSubModel(submodel);
	} // openSubModel

	/**
	* Paint the sector or its submodel with painter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		if (g2==null) return;
//		System.out.println("Sector.paint: " + g2.getTransform());
		super.paint(g2,ilayer);
		textName.paint(g2, x+10, y+20);
		paintPriceLabels(g2,ilayer);
		if (hasSubModel()) lblOpen.paint(g2);
	} // paint

	/**
	* Paint price labels at connectors for flow paths.
	* @param g2 graphics context.
	* @param ilayer index of layer to be painted.
	*/
	private void paintPriceLabels(Graphics2D g2,int ilayer) {
		NodeConnectors conns = getNodeConnectors(ilayer);
		Iterator iterConn = conns.getNodeConnectorsIterator();
		while (iterConn.hasNext()) {
			FlowConnector conn = (FlowConnector)iterConn.next();
			Flow flow = conn.getFlowPath().getFlow();
			// Find price that has this flow
			Iterator iterprice = vrprice.iterator();
			while (iterprice.hasNext()) {
				Price price = (Price)iterprice.next();
				if (price.hasFlow(flow)) {conn.paintPriceMarker(g2,price.isPrescribed());}
			} // while price
		} // while connector
	} // paintPriceLabels

	/** Receive a message that a new flow path connector has been added
	* from NodeConnectors when connecting a new flow path by the
	* connectFlowPath method above.
	* @param iconnNew Index of added connector. */
	public void registerNewConnector(int iconnNew) {
		this.iconnNew = iconnNew;
	} // registerNewConnector

	/** Remove flow from the sector and its corresponding price.
	* @param flowpath The flow path which price shall be removed. */
	public void removeFlowReference(FlowPath flowpath) {

		// Find and remove price reference
		Flow flow = flowpath.getFlow();
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.hasFlow(flow)) {
				timeaxis.remove(price);
				vrprice.remove(price);
				Calculator.removePrice(price);
				break;
			} // if
		} // while

		// Remove flow from submodel
		if (submodel!=null) {
			submodel.removeFlow(flow);
		} // if

		Calculator.calculate();
	} // removeFlowReference

	/** Set if sector has a flow balance for this layer.
	* @param f true if has flow balance, else false.*/
	public void setHasFlowBalance(int ilayer, boolean f) {
		try { ((NodeConnectors)vrNodeConnectors.elementAt(ilayer)).setHasFlowBalance(f); }
		catch(Exception e) {;}
	} // setHasFlowBalance

	/** Set index of selected money layer.
	* @param ilayerMoneySelected Index of selected money layer. */
	public void setIndexOfSelectedMoneyLayer(int ilayerMoneySelected) {
		this.ilayerMoneySelected = ilayerMoneySelected; }

	/** Set index of selected commodity layer.
	* @param ilayerCommoditySelected Index of selected commodity layer. */
	public void setIndexOfSelectedCommodityLayer(int ilayerCommoditySelected) {
		this.ilayerCommoditySelected = ilayerCommoditySelected; }

    /**
    * Translate or resize the node symbol.
    * @param dx translation or resizing in x direction (pixels)
    * @param dy translation or resizing in y direction (pixels)
    */
	public void translate(int dx, int dy) {
		if      (szToMove==NONE)  {super.translate(dx,dy);}
		else if (szToMove==WEST && widthStart-dx>=3*tol)  {x=xStart+dx; width=widthStart-dx;}
		else if (szToMove==EAST && widthStart+dx>=3*tol)  {width=widthStart+dx;}
		else if (szToMove==NORTH && heightStart-dy>=3*tol) {y=yStart+dy; height=heightStart-dy;}
		else if (szToMove==SOUTH && heightStart+dy>=3*tol) {height=heightStart+dy;}
		lblOpen.setPos(x+width,y);
		super.repositionConnectors();
	} // translate

	/** Set the name of the node.
	* @param szName The name to be set.
	*/
	public void setName(String szName) {
		super.setName(szName);
		if (submodel!=null) submodel.setName(szName);
	} // setName

	/**
	* Set start position (before translating) of the node to is current
	* position.
	*/
	public void setStartPos() {
		super.setStartPos();
		// Set start size of sector symbol before resizing
		heightStart = height; widthStart = width;
		lblOpen.setPos(x+width,y);
		if (fOpenLabelHit) openSubModel();
	} // setStartPos

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(int ilayer,Units units) {
		// Set units of sub model of sector
		if (submodel!=null) submodel.setUnitsIfNotSet(ilayer,units);
	} // setUnitsIfNotSet

	/** Show properties of this Sector
	* @return A string telling the properties.*/
	public String toString() {
		return "Sector:   szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag for this sector.
	* @return XML tag. */
	public String xmltag() {return "sector";}

} // Sector