import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/** The sector dialog set the properties of a sector.
* I has a number of functions:
* <dir>
* <li>Set or change the name of the sector.</li>
* <li>Mark if flows in a layer obey a flow balance.</li>
* <li>Add a <a href="Price.html">price</a> relation between flows.<br>
*    Flows have to be in a money layer and a commodity layer.<br>
*    Prices can not be added if they already are used in a price relation.<br>
*    There are combo boxes to choose layers and flows.</li>
* <li>Set names of the prices.</li>
* <li>Open an edit dialog to set prescribed prices.</li>
* <li>Remove price relations.
* </dir>
* <p>
* The OK button closes the dialog. Settings can not be canceled, only redone.
* <p>
* <img src="SectorDialog.gif">
*
*/
public class SectorDialog extends JDialog implements ActionListener,FocusListener {

	// Status variables
	/** The sector to be edited. */
	protected Sector sector;
	/** The time axis for (price) time histories.*/
	protected TimeAxis timeaxis;
	/** The name of the sector. */
	protected String szName;

	/** Number of model layers. */
	private int clayer =1;
	// Price relations
	private Vector vrprice;
	/** Text fields for price names. */
	private Vector vrtxtfldPrice;
	private Vector vrbtnEdit;
	private Vector vrbtnRemove;
	private Vector vrflowMoney;
	private Vector vrflowComdty;

	// Controls
	private Container contentpane;
	private JTextField txtfldName;
	private JPanel panelPriceList;
	private JPanel panelButtons;
	private Checkbox vrchkboxPrescr[];
	private JComboBox comboboxMoneyLayer;
	private JComboBox comboboxComdtyLayer;
	private JComboBox comboboxMoneyFlow;
	private JComboBox comboboxComdtyFlow;
	private JLabel labelWarnMsg;
	private JButton btnAddPrice;
	private JButton btnOpenSubModel;
	private JButton btnOK;

	/** Creates the dialog for editing sector properties.
	* @param owner The owner of the dialog.
	* @param sector The sector which properties shall be set.
	* @param pos The position of the upper left corner of the dialog frame.
	*/
	public SectorDialog(Sector sector, Point pos) {
		super(ECF.frameMain, "Sector properties", true);
		this.sector = sector;
		this.vrprice = sector.getPriceRelations();
		timeaxis = ECF.getTimeAxis();
		setLocation(pos);
		szName = sector.getName();
		ECFModel modelParent = sector.getParentModel();
		// Get content pane
		contentpane = getContentPane();
		// Make check boxes for layers.
		clayer = sector.getNoOfLayers();
		vrchkboxPrescr = new Checkbox[clayer];
		// Make text fields and buttons for prices
		int cprice = vrprice.size();
		vrtxtfldPrice = new Vector();
		vrbtnEdit = new Vector();
		vrbtnRemove = new Vector();
		try { makeContentPane();
		} catch(Exception e) {System.out.println(e);}
		pack();
	} // SectorDialog

	/** Make the content pane ot the dialog. */
	private void makeContentPane() {

		contentpane.setLayout(new BoxLayout(contentpane,BoxLayout.Y_AXIS));

		// Add name input and flow balance check panel
		JPanel panelNameAndCheck = new JPanel(new GridLayout(1+clayer,2));
		addSectorNameInput(panelNameAndCheck);
		addFlowBalanceCheck(panelNameAndCheck);
		contentpane.add(panelNameAndCheck);

		// Add price relation (layer and flow selection) panel
		if (clayer>=2) {
			contentpane.add(makeAddPriceRelationPanel());
			panelPriceList = makePriceListPanel();
			contentpane.add(panelPriceList);
		} // if

		// Add buttons panel
		panelButtons = makeButtonsPanel();
		contentpane.add(panelButtons);

	} // makeContentPane

	/** Add label and text field for sector name.
	* @param panel A grid layout panel. */
	private void addSectorNameInput(JPanel panel) {
		// Sector name
		JLabel labelName = new JLabel("Sector Name:");
		panel.add(labelName,BorderLayout.WEST);
		txtfldName = new JTextField(sector.getName(),10);
		panel.add(txtfldName, BorderLayout.CENTER);
	} //makeNameAndCheckPanel

	/** Make panel for flow balance check boxes.
	* @param panel A grid layout panel. */
	private	void addFlowBalanceCheck(JPanel panel) {
			// Prescribed flow panel
			for (int ilayer=0; ilayer<clayer; ilayer++) {
				Layer layer = ECF.getLayer(ilayer);
				String szNameLayer;
				if (layer==null) szNameLayer = "Layer #"+ ilayer;
				else szNameLayer = layer.getName();
				panel.add(new JLabel(szNameLayer));
				vrchkboxPrescr[ilayer] = new Checkbox("Flow balance:", sector.hasFlowBalance(ilayer));
				panel.add(vrchkboxPrescr[ilayer],BorderLayout.WEST);
			} // for
	} // makeNameAndCheckPanel

	/** Make a panel for adding price relation based on selected layers and flows.
	* @return A panel with four comboboxes and one "Add price" button.
	*/
	private JPanel makeAddPriceRelationPanel() {
		JPanel panelSelectAndAddPrice = new JPanel(new BorderLayout());

		// Selection panel
		JPanel panelSelection = new JPanel(new GridLayout(4,2));

		// Comboboxes for money and commodity layer selection
		// --------------------------------------------------
		// Headings
		panelSelection.add(new JLabel("Money layer"));
		panelSelection.add(new JLabel("Commodity layer"));
		// List layers
		Vector vrlayer = ECF.getLayers();
		Vector vrszLayer = new Vector();
		Iterator iter = vrlayer.iterator();
		while (iter.hasNext()) {
			Layer layer = (Layer)iter.next();
			String szName = layer.getName();
			vrszLayer.add(szName);
		} // while
		// Combo box for Money layers
		comboboxMoneyLayer = new JComboBox(vrszLayer);
		comboboxMoneyLayer.setSelectedIndex(sector.getIndexOfSelectedMoneyLayer());
		panelSelection.add(comboboxMoneyLayer);
		// Combo box for Commodity layers
		comboboxComdtyLayer = new JComboBox(vrszLayer);
		comboboxComdtyLayer.setSelectedIndex(sector.getIndexOfSelectedCommodityLayer());
		panelSelection.add(comboboxComdtyLayer);

		// Combo boxes for flow selection
		// ------------------------------
		// Headings
		panelSelection.add(new JLabel("Money flow"));
		panelSelection.add(new JLabel("Commodity flow"));
		// Money flows combo box
		int ilayerMoney = comboboxMoneyLayer.getSelectedIndex();
		vrflowMoney = getFlows(ilayerMoney);
		Vector vrszFlow = new Vector();
		iter = vrflowMoney.iterator();
		while (iter.hasNext()) vrszFlow.add(((Flow)iter.next()).getName());
		comboboxMoneyFlow = new JComboBox(vrszFlow);
		panelSelection.add(comboboxMoneyFlow);
		// Commodity flows combo box
		int ilayerComdty = comboboxComdtyLayer.getSelectedIndex();
		vrflowComdty = getFlows(ilayerComdty);
		vrszFlow = new Vector();
		iter = vrflowComdty.iterator();
		while (iter.hasNext()) vrszFlow.add(((Flow)iter.next()).getName());
		comboboxComdtyFlow = new JComboBox(vrszFlow);
		panelSelection.add(comboboxComdtyFlow);

		// Add selection comboboxes
		panelSelectAndAddPrice.add(panelSelection,BorderLayout.NORTH);

		// Add warning message label
		labelWarnMsg = new JLabel(" ");
		labelWarnMsg.setForeground(Color.red);
		labelWarnMsg.setVisible(true);
		panelSelectAndAddPrice.add(labelWarnMsg,BorderLayout.CENTER);

		// Add add price button
		btnAddPrice = new JButton("Add price");
		btnAddPrice.setEnabled(true);  // Add price button enabled as default
		panelSelectAndAddPrice.add(btnAddPrice,BorderLayout.SOUTH);

		// Check if different layers were chosen
		checkLayerSelections();
		// Check if flows are used in prices
		disableAddpriceButtonIfFlowUsed();

		// Add action listeners
		//---------------------
		comboboxMoneyLayer.addActionListener(this);
		comboboxComdtyLayer.addActionListener(this);
		comboboxMoneyFlow.addActionListener(this);
		comboboxComdtyFlow.addActionListener(this);
		btnAddPrice.addActionListener(this);

        panelSelectAndAddPrice.setBorder(BorderFactory.createTitledBorder(
                                    "Add a price relation:"));

		return panelSelectAndAddPrice;
	} // makeAddPriceRelationPanel

	private JPanel makePriceListPanel() {
		// List of prices
		JLabel label;
		JButton btn;
		Font fontHeadings = new Font("Arial", Font.BOLD, 12);
		int cprice = vrprice.size();
		JPanel panelListOfPrices = new JPanel(new GridLayout(1+cprice,5,10,0));

		// Headings
		label = new JLabel("Money flow");
		label.setFont(fontHeadings);
		panelListOfPrices.add(label);
		label = new JLabel("Commodity flow");
		label.setFont(fontHeadings);
		panelListOfPrices.add(label);
		label = new JLabel("Name of price");
		label.setFont(fontHeadings);
		panelListOfPrices.add(label);
		// Fill remaining cells of heading line
		panelListOfPrices.add(new JLabel(""));
		panelListOfPrices.add(new JLabel(""));

		// Prices
		Iterator iter = vrprice.iterator();
		int iprice=0;
		vrtxtfldPrice.clear();
		vrbtnEdit.clear();
		vrbtnRemove.clear();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			// Names of money and commodity flows
			panelListOfPrices.add(new JLabel(price.getFlowMoney().getName()));
			panelListOfPrices.add(new JLabel(price.getFlowCommodity().getName()));
			// Name text field
			JTextField txtfld = new JTextField(price.getName());
			txtfld.setEnabled(true);
			vrtxtfldPrice.add(txtfld);
			panelListOfPrices.add(txtfld);
			// Edit price button
			btn = new JButton("Edit...");
			btn.addActionListener(this);
			vrbtnEdit.add(btn);
			panelListOfPrices.add(btn);
			// Remove price button
			btn = new JButton("Remove");
			btn.addActionListener(this);
			vrbtnRemove.add(btn);
			panelListOfPrices.add(btn);
			// Add focus listener
			txtfld.addFocusListener(this);
			// Add actionlisteners
			txtfld.addActionListener(this);
			iprice++;
		} // while
		// Create border
        panelListOfPrices.setBorder(BorderFactory.createTitledBorder(
                                    "Edit prices:"));
		return panelListOfPrices;
	} // makePriceListPanel

	/** Make a vector of flows of layer ilayer.
	* @param ilayer Index of the layer for which the flows are listed. */
	private Vector getFlows(int ilayer) {
		Vector vrflow = new Vector();
		NodeConnectors conns = sector.getNodeConnectors(ilayer);
		Iterator iter = conns.getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			Flow flow = conn.getFlowPath().getFlow();
			vrflow.add(flow);
		} // while
		return vrflow;
	} // getFlows

	/** Make panel for buttons OK and Cancel.
	* @return A buttons panel. */
	private JPanel makeButtonsPanel() {
		JPanel panelButtons = new JPanel();
		panelButtons.setLayout(new FlowLayout());

		btnOK = new JButton("OK");
		btnOK.addActionListener(this);
		panelButtons.add(btnOK);

		btnOpenSubModel = new JButton("Open submodel");
		btnOpenSubModel.addActionListener(this);
		panelButtons.add(btnOpenSubModel);

		return panelButtons;
	} //makeButtonsPanel

	/** Check if different layers were chosen,
	* If same layers: issue warning message and disable "Add price" button. */
	private void checkLayerSelections() {
		int ilayerMoney = comboboxMoneyLayer.getSelectedIndex();
		int ilayerComdty = comboboxComdtyLayer.getSelectedIndex();
		// Check if different layers were chosen
		if (ilayerMoney==ilayerComdty) {
			if (labelWarnMsg!=null) labelWarnMsg.setText("Different layers have to be chosen");
			if (btnAddPrice!=null) btnAddPrice.setEnabled(false);
		} // if different layers
	} // checkLayerSelections

	/** Check if flow was used in prices before.
	* @param vrflow Vector of flows in combo box.
	* @param comboboxFlow Combo box foe flow selection.
	* @return true if flow is used in a price relation. */
	private boolean checkIfFlowUsedInPrice(Vector vrflow,JComboBox comboboxFlow) {
		boolean fFlowUsed = false;
		if (vrflow.size()==0) return fFlowUsed;
		// Check flow
		Flow flow = (Flow)vrflow.elementAt(comboboxFlow.getSelectedIndex());
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.hasFlow(flow)) {
				fFlowUsed = true;
			} // if
		} // while money flow
		return fFlowUsed;
	} // checkIfFlowUsedInPrice

	/** Disable "Add price" button if either money flow or
	* commodity flow alredy is used in a price relation. */
	private void disableAddpriceButtonIfFlowUsed() {
		if (vrflowMoney.size()==0 || vrflowComdty.size()==0 ||
			checkIfFlowUsedInPrice(vrflowMoney,comboboxMoneyFlow) ||
			checkIfFlowUsedInPrice(vrflowComdty,comboboxComdtyFlow))
		btnAddPrice.setEnabled(false);
	} // disableAddpriceButtonIfFlowUsed

	/** Check if flow was used in prices before.
	* If used before: issue warning message..
	* @param vrflow Vector of flows in combo box.
	* @param comboboxFlow Combo box foe flow selection.
	* @return true if flow is used in a price relation. */
	private void warnIfFlowUsed(Vector vrflow,JComboBox comboboxFlow) {
		// Check flow
		Flow flow = (Flow)vrflow.elementAt(comboboxFlow.getSelectedIndex());
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			if (price.hasFlow(flow)) {
				labelWarnMsg.setText(flow.getName() +
					" is already used for price " + price.getName());
				btnAddPrice.setEnabled(false);
			} // if
		} // while money flow
	} // warnIfFlowUsed

	/** Refrech list of prices. */
	private void refreshListOfPrices() {
		// Remove the two last panels
		contentpane.remove(panelPriceList);
		contentpane.remove(panelButtons);
		// Remake and add the two last panels
		panelPriceList = makePriceListPanel();
		contentpane.add(panelPriceList);
		panelButtons = makeButtonsPanel();
		contentpane.add(panelButtons);
		pack();
	} // refreshListOfPrices

	/** Open a dialog for setting properties of a price.
	* @param price The Price which properties shall be edited. */
	public void openPriceDialog(Price price) {
		PriceDialog dialog = new PriceDialog(this,price);
	} // openDialog

	/** Invoked when a text field for setting a price name
	* gains the keyboard focus, does nothing.
	* @param e The event that activated the FocusListener.
	*/
	public void focusGained(FocusEvent e) {
	} // focusGained

	/** Invoked when a text field for setting a price name
	* gains the keyboard focus, sets the name of a <a href="Price.html">price</a>.
	* @param e The event that activated the FocusListener.
	*/
	public void focusLost(FocusEvent e) {
		// Find the invoking text field and set price name.
		int iprice = 0;
		Iterator iter = vrprice.iterator();
		while (iter.hasNext()) {
			Price price = (Price)iter.next();
			JTextField txtfld = (JTextField)vrtxtfldPrice.elementAt(iprice);
			if (e.getSource()==txtfld)    {
				price.setName(((JTextField)txtfld).getText());
			} // if
			iprice++;
		} // while
	} // focusLost

	/** Processes an action event fired by combo boxes and buttons.
	* @param e The event that activated the ActionListener.
	*/
	public void actionPerformed(ActionEvent e) {

		// No warning text at start
		if (labelWarnMsg!=null) labelWarnMsg.setText(" ");
		// Add price button enabled as default
		if (btnAddPrice!=null) btnAddPrice.setEnabled(true);

		// Update Money combo box
		if (e.getSource()==comboboxMoneyLayer) {
			comboboxMoneyFlow.removeActionListener(this);
			int ilayer = comboboxMoneyLayer.getSelectedIndex();
			sector.setIndexOfSelectedMoneyLayer(ilayer);
			comboboxMoneyFlow.removeAllItems();
			Vector vrflowMoney = getFlows(ilayer);
			Iterator iter = vrflowMoney.iterator();
			while (iter.hasNext()) comboboxMoneyFlow.addItem(((Flow)iter.next()).getName());
			comboboxMoneyFlow.addActionListener(this);
			// Check if different layers were chosen
			checkLayerSelections();
			disableAddpriceButtonIfFlowUsed();

		// Update Commodity combo box
		} else if (e.getSource()==comboboxComdtyLayer) {
			comboboxComdtyFlow.removeActionListener(this);
			int ilayer = comboboxComdtyLayer.getSelectedIndex();
			sector.setIndexOfSelectedCommodityLayer(ilayer);
			comboboxComdtyFlow.removeAllItems();
			Vector vrflowComdty = getFlows(ilayer);
			Iterator iter = vrflowComdty.iterator();
			while (iter.hasNext()) comboboxComdtyFlow.addItem(((Flow)iter.next()).getName());
			comboboxComdtyFlow.addActionListener(this);
			// Check if different layers were chosen
			checkLayerSelections();
			disableAddpriceButtonIfFlowUsed();

		// Check if flows are used in prices
		} else if (e.getSource()==comboboxMoneyFlow) {
			disableAddpriceButtonIfFlowUsed();
			warnIfFlowUsed(vrflowMoney,comboboxMoneyFlow);

		// Check if flows are used in prices
		} else if (e.getSource()==comboboxComdtyFlow) {
			disableAddpriceButtonIfFlowUsed();
			warnIfFlowUsed(vrflowComdty,comboboxComdtyFlow);

		// Add a new price relation
		} else if (e.getSource()==btnAddPrice) {
			Price price = new Price(sector);
			try {
				Flow flowMoney = (Flow)vrflowMoney.elementAt(comboboxMoneyFlow.getSelectedIndex());
				price.setFlowMoney(flowMoney);
				Flow flowComdty = (Flow)vrflowComdty.elementAt(comboboxComdtyFlow.getSelectedIndex());
				price.setFlowCommodity(flowComdty);
				vrprice.add(price);
				refreshListOfPrices();
				disableAddpriceButtonIfFlowUsed();
			} catch(NoUnitException excNoUnit) {
				if (labelWarnMsg!=null) labelWarnMsg.setText("Flow '" +
					excNoUnit.getMessage() + "' has no units. Assign units before creating a price.");
			} catch(Exception excAdd) {
				// The catch clause will find a variety of null pointer exceptions
				// and index out of bounds exceptions.
				// Instead of checking beforehand, it is easier with try and catch.
				System.out.println("SectorDialog.actionPerformed: \n" + excAdd);
			} // try/catch

		// OK button
		} else if (e.getSource()==btnOK) {
			sector.setName(txtfldName.getText());
			for (int ilayer=0; ilayer<clayer; ilayer++)
				sector.setHasFlowBalance(ilayer,vrchkboxPrescr[ilayer].getState());
			dispose();
			Calculator.calculate();

		// Open submodel button
		} else if (e.getSource()==btnOpenSubModel) {
			sector.setName(txtfldName.getText());
			for (int ilayer=0; ilayer<clayer; ilayer++)
				sector.setHasFlowBalance(ilayer,vrchkboxPrescr[ilayer].getState());
			sector.openSubModel();
			dispose();

		// Edit price and Remove price buttons
		} else {
			JButton btn;
			int iprice = 0;
			Iterator iter = vrprice.iterator();
			while (iter.hasNext()) {
				Price price = (Price)iter.next();
				btn = (JButton)vrbtnEdit.elementAt(iprice);
				if (e.getSource()==btn)    {
					openPriceDialog(price);
					break;
				} // if
				btn = (JButton)vrbtnRemove.elementAt(iprice);
				if (e.getSource()==btn)    {
					vrprice.remove(price);
					Calculator.removePrice(price);
					break;
				} // if
				iprice++;
			} // while
			try {
				refreshListOfPrices();
				// Check if flows are used in prices
				disableAddpriceButtonIfFlowUsed();
			} catch(Exception excAdd) {
				System.out.println("SectorDialog.actionPerformed: remove price\n" + excAdd);
			} // try/catch

		} // if

	} // actionPerformed

} // class SectorDialog