import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;

/**
* A source/sink of the economic system. A source is a node that is connected
* to one single flow. It belongs to one model layer only. A flow may come from the
* source or go to the node, which then is a sink.
*/
public class Source extends ECFNode {

	/** Index of the layer to which the source belongs. */
	protected int ilayer;
	/** Radius of the source symbol. */
	protected final int radius = 15;
	/** The source flow of the source/sink. */
	protected Flow flowSource = null;
	/** Index of source. */
	static int isource = 0;
	/** Reset source index. */
	public static void resetIndex() { isource = 0; }

	/**
	* Creates a new source/sink with no source flow.
	* @param model The model to which the source belongs.
	* @param ilayer index of the layer to which the source belongs
	* @param x,y coordinates of the upper left corner of the source symbol
	*/
	public Source(ECFModel model, int ilayer, int x, int y) {
		super(model,x,y);
		this.ilayer = ilayer;
		// Set symbol properties
		width = 2*radius;
		height = 2*radius;
		isource++;
		setName("Source"+isource);
		widthStroke = 3;
		makePainter();
		// Add node connectors for layer ilayer.
		NodeConnectors conns = new NodeConnectors(ilayer,this);
		conns.setHasFlowBalance(false);
		vrNodeConnectors.add(conns);
	} // Source

	/**
	* Creates a new source/sink with default source flow..
	* @param model The model to which the source belongs.
	* @param ilayer Index of the layer to which the source belongs
	* @param x,y coordinates of the upper left corner of the source symbol
	* @param sz Dummy argument.
	*/
	public Source(ECFModel model, int ilayer, int x, int y, String sz) {
		this(model,ilayer,x,y);
		// Set source flow.
		flowSource = new Flow(ilayer,"SourceFlow" + isource);
		ECF.vrflow.add(flowSource);
		Calculator.calculate();
	} // Source

	/** Remove units from source flow of layer ilayer
	* @param ilayer Index of layer from which units are removed.*/
	public void clearUnits(int ilayer) {
		if (this.ilayer==ilayer) flowSource.clearUnits(ilayer);
	} // clearUnits

	/**
	* Checks if the source belongs to layer <code>ilayer</code> and contains
	* the coordinates <code>x,y</code>.
	* @param ilayer index of the layer to which the source belongs
	* @param x,y coordinates to be contained within the source symbol
	*/
	public boolean contains(int ilayer,int x,int y) {
		if (this.ilayer != ilayer) return false;
		return getShape().contains(x,y);
	} // contains

	/**
	* Decode source items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		if (szTag.equals("ilayer")) ilayer = Integer.parseInt(szSub);
		if (szTag.equals("flowindex")) {
			int iflow = Integer.parseInt(szSub);
			Flow flowTemp = Calculator.getFlow(iflow);
			if (flowTemp!=null) flowSource = flowTemp;
			else {
				String szMsg = "Source.decodeXML could not find flow " +
					"with index=" + iflow;
				JOptionPane.showMessageDialog(ECF.frameMain,szMsg);
			} // if/else
		} // flowindex
		super.decodeXMLItems(szTag,szSub);
	} // decodeXML

	/**
	* Release associations to flow paths before deleting node.
	*/
	public void	disconnectThis() {
		super.disconnectThis();
		ECF.removeFlow(flowSource);
	} // deleteThis

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors(int ilayer) {
		if (ilayer==this.ilayer) return (NodeConnectors)vrNodeConnectors.elementAt(0);
		else return null;
	} // getNodeConnectors

	/**
	* Get the shape of this source symbol.
	*/
	public Shape getShape() {
		return new Ellipse2D.Float((float)x,(float)y, (float)2*radius, (float)2*radius);
	}

	/** Get node specific equations, to be specified in subtypes.
	* @return A vector of <a href="LinearEquation.html"><code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() {
		Vector vreq = new Vector();
		try {
			NodeConnectors conns = (NodeConnectors)vrNodeConnectors.elementAt(0);
			LinearEquation eq = conns.getFlowBalanceEquation();
			int ivar = flowSource.getIndexOfFlow();
			eq.setCoefficient(ivar,1);
			vreq.add(eq);
		} catch(Exception e) {
			JOptionPane.showMessageDialog(ECF.frameMain,
				"Error in Source.getNodeSpecificEquations(): \n" + e +
				"\nMail your model file to olert@swipnet.se",
				"ECFModel.decodeXML",JOptionPane.WARNING_MESSAGE);
			System.out.println("Source.getNodeSpecificEquations:" + e);
		} // try catch
		return vreq;
	} // getNodeSpecificEquations

	/** Get the source flow.
	* @return The flow from the source. */
	public Flow getSourceFlow() {return flowSource;}

	/** Check if source has a flow balance for this layer.
	* @return true if has flow balance, else false.*/
	public boolean hasFlowBalance(int ilayer) {
		try {
			if (ilayer == this.ilayer)
				return ((NodeConnectors)vrNodeConnectors.elementAt(0)).hasFlowBalance();
			else
				return false;
		} catch(Exception e) { return false;}
	} // hasFlowBalance

	/** List XML code for the junction.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<ilayer>"+ilayer+"</ilayer>";
		szXML += szBlanks +"<flowindex>"+ECF.calculator.getIndexOfFlow(flowSource)+"</flowindex>";
		szXML += super.listXML(szBlanks);
		return szXML;
	} // listXML

	/** Open a dialog window for source properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		Point pnt = new Point(ECF.frameMain.getX()+20,ECF.frameMain.getY()+20);
		SourceDialog dialog = new SourceDialog(ECF.frameMain,this, pnt);
		dialog.show();
	} // openDialog

	/**
	* Paint the source with painter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		if (g2==null || flowSource==null) return;
		if (this.ilayer==ilayer)  {
			super.paint(g2,ilayer);
			if (flowSource.isPrescribed()) {
				int xFB = x-4; int yFB = y-4;
				// Colored square for source flow condition
				g2.setColor(ECFModel.COLOR_CONDITION);
				g2.fillRect(xFB,yFB,15,10);
				// Text for source flow condition
				g2.setColor(Color.black);
				g2.drawString("SF",xFB,yFB+10);
			} // if
			// Text for source name
			textName.paint(g2, (int)(x+1.5*radius), y+2*radius+12);
		} // if
	} // paint

	/** Set the units of layer ilayer if not previously set.
	* @param ilayer layer index.
	* @param units The units to be set.*/
	public void setUnitsIfNotSet(int ilayer,Units units) {
		// Set units of source flow
		if (this.ilayer==ilayer && flowSource!=null) flowSource.setUnitsIfNotSet(units);
	} // setUnitsIfNotSet

	/** Show properties of this Sector
	* @return A string telling the properties.*/
	public String toString() {
		return "Source:   szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag fot this source.
	* @return XML tag. */
	public String xmltag() {return "source";}

} // Source