import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.util.*;

/**
* A splitter of the economic system. A splitter is a node that is connected
* to flows of the same kind. It belongs to one model layer only. The sum of
* flows (taken with sign) to and from the junction is zero. The flow rate
* of individual outflows can be specified as fractions (flow ratios) of the
* total inflow.
*/
public class Splitter extends AbstractJunction {

	/** A vector of <a href="FlowRatio.html">FlowRatios</a>*/
	protected Vector vrflowratio = new Vector();
	/** Index of splitter. */
	static int isplitter = 1;
	/** Reset splitter index. */
	public static void resetIndex() { isplitter = 1; }

	/**
	* Creates a new splitter.
	* @param time axis the time axis flor flow ratios.
	* @param ilayer index of the layer to which the splitter belongs
	* @param x,y coordinates of the upper left corner of the splitter symbol
	*/
	public Splitter(ECFModel model, int ilayer, int x, int y) {
		super(model,ilayer,x,y);
		setName("Splitter"+isplitter);
		isplitter++;
	} // Splitter

	/** Clear variables of nodes, for splitter remove flowratios. */
	public void clearVariables() {
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			Calculator.removeFlowRatio(flowratio);
		} // while
		vrflowratio.clear();
	} // clearVariables

	/** Connect a flow path to this splitter.
	* If the flow is an outflow, create a default flow ratio = 0
	* @param ilayer index of the layer of the flow path
	* @param flowpath the flowpath to be connected. */
	public void connectFlowPath(int ilayer, FlowPath flowpath) {
		if (ilayer==this.ilayer) {
			FlowConnector conn = flowpath.getSelectedConnector();
			if (conn.isTailEnd()) {
				// The splitter is the source of the flow.
				// = the flow is an outflow
				FlowRatio flowratio = new FlowRatio(this);
				flowratio.setFlowOut(flowpath.getFlow());
				vrflowratio.add(flowratio);
			} // if
		}
		super.connectFlowPath(ilayer,flowpath);
	} // connectFlowPath

	/**
	* Decode junction items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		if (szTag.equals("FlowRatio")) {
			FlowRatio flowratio = new FlowRatio(this);
			flowratio.decodeXML(szSub);
			vrflowratio.add(flowratio);
		} // if
		super.decodeXMLItems(szTag,szSub);
	} // decodeXMLItems

	public Vector getFlowsIn() {
		Vector vrflowIn = new Vector();

		// Check all connectors and find inflows
		NodeConnectors conns = getNodeConnectors();
		Iterator iter = conns.getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			if (conn.isHeadEnd()) {
				vrflowIn.add(conn.getFlowPath().getFlow());
			} // if
		} // while

		return vrflowIn;
	} // getFlowsIn

	/** Get a Vector of FlowRatios for the splitter.
	* @return A Vector of <a href="Flowratio.html">FlowRatios</a>.
	*/
	public Vector getFlowRatios() { return vrflowratio; }

	/** Get node specific equations for outflow ratios, one for each
	* prescribed flow ratio.
	* @return A vector of <a href="LinearEquation.html">
	*         <code>LinearEquations</code></a>. */
	protected Vector getNodeSpecificEquations() throws Exception {
		Vector vrequations = new Vector();
		// Flow splitter ratios
		Iterator iterFlowRatio = vrflowratio.iterator();
		while (iterFlowRatio.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iterFlowRatio.next();
			if (flowratio.isPrescribed())
				vrequations.add(getSplitterRatioEquation(flowratio));
		} // while
		return vrequations;
	} // getNodeSpecificEquations


	/** Get a flow ratio equation for for the splitter.
	* @param flowratio a flow ratio object.
	* @return An equation for flow ratio.
	*/
	public LinearEquation getSplitterRatioEquation(FlowRatio flowratio) throws Exception {

		int cflows = Calculator.getNumberOfFlows();
		LinearEquation equation= new LinearEquation(cflows);  // Flows + right member

		// Set cofficients for inflows
		NodeConnectors conns = getNodeConnectors();
		Iterator iter = conns.getNodeConnectorsIterator();
		while (iter.hasNext()) {
			FlowConnector conn = (FlowConnector)iter.next();
			int iend = conn.getEndOfFlow();
			FlowPath flowpath = conn.getFlowPath();
			Flow flow = flowpath.getFlow();
			double ratio = flowratio.getBaseValuePrescribed();
			if (conn.isHeadEnd()) {
				if (ratio==0.0) throw new Exception("Flow ratio " +
					flowratio.getName() +" of splitter " + model.getName() +
					"/" + szName +  " is zero");
				equation.setCoefficient(flow.getIndexOfFlow(),ratio);
			} // if
		} // while

		// Set ratio for outflow
		equation.setCoefficient(flowratio.getFlowOut().getIndexOfFlow(),-1);

		equation.setType("Splitter ratio");
//		equation.setNodeName(getName());
		equation.setNode(this);
		equation.setLayerIndex(ilayer);

		return equation;
	} // getSplitterRatioEquation

	/** Creates a <code>Shape</code> of a line from the center <code>p1</code>
	* of the splitter symbol to the connection point <code>p2</code> of an outflow.
	* @param p1 The center of the splitter symbol.
	* @param p2 The connection point of an outflow.
	* @return The <code>Shape</code> of a line from point <code>p1</code>
	*         to point <code>p2</code>.*/
	private Shape getShapeRatioLine(Point2D p1, Point2D p2) {
		Line2D line = new Line2D.Double(p1,p2);
		return line;
	} // getShapeRatioLine

	/** List XML code for the splitter.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = super.listXML(szBlanks);
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			szXML += szBlanks +"<FlowRatio>";
			szXML += flowratio.listXML("  "+szBlanks);
			szXML += szBlanks +"</FlowRatio>";
		} // while
		return szXML;
	} // listXML

	/**
	* Make a painter to paint the line to indicate that
	* a flow ratio has been set.
	* @param colorFill the color of the line.
	* @return A Painter.
	*/
	protected Painter makePainterRatioLine(Color colorFill) {
		Painter painterFill = new FillPainter(colorFill);
		int widthStroke = 5;
		Stroke stroke = new BasicStroke(widthStroke);
		Painter painterDraw = new DrawPainter(colorFill,stroke);
		return new MultiPainter(painterFill,painterDraw);
	} // makePainter

	/** Open a dialog window for splitter properties.
	* @param mainframe The parent frame.*/
	public void openDialog() {
		Point pnt = new Point(ECF.frameMain.getX()+20,ECF.frameMain.getY()+20);
		SplitterDialog dialog = new SplitterDialog(this, pnt);
		dialog.show();
	} // openDialog

	/**
	* Paint the splitter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {

		super.paint(g2,ilayer);
		if (g2==null) return;

		if (this.ilayer==ilayer)  {

			// Lines indicating directions that have flow ratios,
			// from center to flow connector.
			paintRatioLines(g2);

			// A capitol "S" for splitter in the center of the circle
			g2.setColor(Color.black);
			TextGeometry text = new TextGeometry(new Font("Times",Font.BOLD,18),"S");
			Rectangle rectBounds = text.getBounds();
			int h = (int)rectBounds.getHeight();
			int w = (int)rectBounds.getWidth();
			text.paint(g2, x+radius-w/2, y+radius+h/2);
		}
	} // paint

	/** Paint lines inside the splitter symbol that indicate the flows
	* for which ratios have been specified.
	* Lines run from center to flow connectors.
	* @param g2 graphics context. */
	private void paintRatioLines(Graphics2D g2) {

		Color colorFill = ECFModel.COLOR_PRESCRIBED;
		Painter painterRatioLine = makePainterRatioLine(colorFill);
		Point2D posCenter = new Point2D.Double(x+radius,y+radius);

		// Flow ratios
		g2.setColor(ECFModel.COLOR_PRESCRIBED);
		// Check all flowratios
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			// Flow ratio
			FlowRatio flowratio = (FlowRatio)iter.next();
			Flow flow = flowratio.getFlowOut();
			// Find corresponding flow connector
			NodeConnectors conns = getNodeConnectors();
			Iterator iterConns = conns.getNodeConnectorsIterator();
			while (iterConns.hasNext()) {
				FlowConnector conn = (FlowConnector)iterConns.next();
				FlowPath flowpath = conn.getFlowPath();
				Flow flowConn = flowpath.getFlow();
				if (flowConn==flow && flowratio.isPrescribed()) {
					// Draw a line
					Point2D posConn = conn.getRefPoint();
					Shape shapeRatioLine = getShapeRatioLine(posCenter,posConn);
					painterRatioLine.paint(g2,shapeRatioLine);
				} // if
			} // while connectors
		} // while ratios
	} // paintRatioLines

	/**
	* Paints the flow ratio curves with axes in a diagram.
	* @param g The graphical context.
	* @param xOrigin The x-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param yOrigin The y-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param widthXAxis The width (pixels) of the time axis.
	* @param heightYAxis The height (pixels) of the value axis.
	*/
	public void paintRatioCurves(Graphics g, int xOrigin, int yOrigin,
		int widthXAxis, int heightYAxis) {
		// Find maximum of ratio curves
		FlowRatio flowratio;
		double max = 0.0;
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			flowratio = (FlowRatio)iter.next();
			max = Math.max(max,flowratio.maxValue());
		} // while
		// Value axis
		ValueAxis valueaxis = new ValueAxis(max," ");
		valueaxis.paint(g, xOrigin, heightYAxis, yOrigin);
		// Time axis
		if (ECF.timeaxis.getTimeIncrements().size()==0) return;
		ECF.timeaxis.paint(g, xOrigin, xOrigin+widthXAxis, yOrigin);
		// Paint ratio curves
		iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			flowratio = (FlowRatio)iter.next();
			ValuesPlot valuesplot = new ValuesPlot(ECF.timeaxis.getTimeIncrements(),
				flowratio.getDisplayValues(),flowratio.getDisplayColors());
			valuesplot.paint(g,xOrigin,yOrigin,widthXAxis,heightYAxis,valueaxis.getScale(),
		           ECF.timeaxis.getSelectedIncrement());
		} // while
	} // paintRatioCurves

	/** Remove flow from the splitter and its corresponding flow ratio.
	* @param flowpath The flow path which flow ratio shall be removed. */
	public void removeFlowReference(FlowPath flowpath) {

		// Find and remove flow reference
		Flow flow = flowpath.getFlow();
		Iterator iter = vrflowratio.iterator();
		while (iter.hasNext()) {
			FlowRatio flowratio = (FlowRatio)iter.next();
			if (flowratio.getFlowOut()==flow) {
				ECF.timeaxis.remove(flowratio);
				vrflowratio.remove(flowratio);
				Calculator.removeFlowRatio(flowratio);
				break;
			} // if
		} // while

		Calculator.calculate();
	} // removeFlowReference

	/** Show properties of this Splitter
	* @return A string telling the properties.*/
	public String toString() {
		return "Splitter: szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag for this splitter.
	* @return XML tag. */
	public String xmltag() {return "splitter";}

} // Splitter