import java.awt.*;
import java.awt.geom.AffineTransform;
import java.util.*;  // Vector

/** A submodel is a model inside a sector. It has the same
* layers as its supermodel and the same units.
*/
public class SubModel extends ECFModel {

	/** The parent sector to which the submodel belongs. */
	private Sector sector;
	/** The frame that encloses the submodel.*/
	protected ModelFrame modelframe;

	/** Create a submodel inside a frame. */
	public SubModel(Sector sector) {
		this.sector = sector;
		int x= 100;
		int y = 50;
		modelSuper = sector.getParentModel();
		szName = sector.getName();
		modelframe = new ModelFrame(this,sector,x,y);
		modelframe.setName(sector.getName());
	} // SubModel

	/** Add external flows to the frame of the submodel.
	* @param vrNodeConnectors A vector of NodeConnectors,
	*        one item for each layer.*/
	public void addFlowsToTheFrame(Vector vrNodeConnectors) {
		modelframe.addFlowsToTheFrame(vrNodeConnectors);
	} // addFlowsToTheFrame

	/** Add a layer to submodel.
	* @param ilayer Index of the layer to be added. */
	public void addLayer(int ilayer) {
		super.addLayer();
		modelframe.addLayer(ilayer);
	} // addLayer

	/** Connect flow path to this node.
	* @param ilayer Index of the layer of the flow path
	* @param flowpath The flow path to be connected.
	* @param iconnNew Index of added connector.
	*/
	public void connectFlowPath(int ilayer, FlowPath flowpath, int iconnNew) {
		modelframe.connectFlowPath(ilayer,flowpath,iconnNew);
	} // connectFlowPath

	/**
	* Decode submodel from XML code.
	* @param szXMLCode The xml description of the submodel.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("modelframe")) {
					modelframe.decodeXML(szSub);
				} // if
				if (szTag.equals("model")) {
					super.decodeXML(szSub);
				} // if
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println("SubModel.decodeXML: " + e +
			"\n  " + szXMLCode);
		} // try/catch
	} // decodeXML

	/** Delete the submodel. */
	public void deleteSubmodel() {
		super.clearModel();
		modelframe.deleteModelFrame();
		modelframe = null;
	} // deleteSubmodel

	/**
	* Finds a flowpath of the submodel. An <code>FlowPath</code> is returned if
	* contained in layer <code>ilayer</code> and contains the mouse pointer
	* coordinates <code>(x,y)</code>.
	* @param ilayer the index of the layer pointed in by the mouse.
	* @param (x,y) the panel coordinates of the mose pointer.
	* @return FlowPath if a flowpath was found, else <code>null</code>.
	*/
	public FlowPath findFlowPath(int ilayer,int x,int y) {
		FlowPath flowpath = null;
		// Try to find the flowpath in model frame.
		flowpath = modelframe.findFlowPath(ilayer,x,y);
		// Try to find the flowpath in the enclosed model.
		if (flowpath==null)	flowpath = super.findFlowPath(ilayer,x,y);
		return flowpath;
	} // findFlowPath

	/** Get the model frame of the submodel.
	* @return The frame of the submodel. */
	public ModelFrame getModelFrame() {
		return modelframe;
	} // getModelFrame

	/** Get the name of the submodel = parent sector name.
	* @return The name of the parent sector. */
	public String getName() {
		return szName;
	} // getName

	/** List XML code for the submodel.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = szBlanks +"<modelframe>";
		szXML += modelframe.listXML("  "+szBlanks);
		szXML += szBlanks +"</modelframe>";
		szXML += szBlanks +"<model>";
		szXML += super.listXML("  "+szBlanks);
		szXML += szBlanks +"</model>";
		return szXML;
	} // listXML

	/**
	* Paint the sector or its submodel with painter.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		if (g2==null) return;
		modelframe.paint(g2,ilayer);
		super.paint(g2,ilayer);
	} // paint

	/** Remove outside flow from the submodel.
	* @param flow The flow to be removed; */
	public void removeFlow(Flow flow) {
//		System.out.println("SubModel.removeFlow:" + flow);
		modelframe.removeFlow(flow);
	} // removeFlow

	/**
	* Selects a node or flow path in layer <code>ilayer</code> if
	* the model element contains the coordinates <code>(x,y)</code>,
	* otherwise deselects all elements. sets the start point
	* <code>(xStart,yStart)</code> for dragging nodes or a flow path
	* label.
	* @param fAddSelection if true adds element to previous selections,
	*        else selects only this node or flow path.
	* @param fEditProperties if true opens a window for editing the
	*        properties of this element.
	* @param ilayer the index of the layer to contain the element.
	* @param (x,y) the coordinates to be contained by the element.
	*/
	public void selectModelElement(boolean fAddSelection,
				boolean fEditProperties, int ilayer, int x,int y) {
//		System.out.println("SubModel.selectModelElement: x=" + x + " y=" + y);
		if (modelframe.containsInCloseLabel(ilayer,x,y)) {
			ECF.closeSubModel(modelSuper);
		} else {
			super.selectModelElement(fAddSelection,
				fEditProperties,ilayer,x,y);
		} // if/else
	} // selectModelElement

	/** Set the name of the submodel.
	* @param szName The name to be set.
	*/
	public void setName(String szName) {
		super.setName(szName);
		modelframe.setName(sector.getName());
	} // setName

	public String toString() { return "SubModel " + szName; }

} // SubModel