import java.awt.*;
import java.awt.geom.*;

/**
* A terminal of the economic system. A terminal is a node that is connected
* to flows of the same kind. It belongs to one model layer only. The sum of
* flows (taken with sign) to and from the terminal is zero.
*/
public class Terminal extends ECFNode {

	/** Index of the layer to which the terminal belongs. */
	protected int ilayer;
	/** Radius of the terminal symbol. */
	protected int radius = 10;
	/** Node connectors. */
	protected NodeConnectors conn;
	/** Index of terminal. */
	static int iterminal = 1;
	/** Reset terminal index. */
	public static void resetIndex() { iterminal = 1; }


	/**
	* Creates a new terminal.
	* @param ilayer index of the layer to which the terminal belongs
	* @param x,y coordinates of the upper left corner of the terminal symbol
	*/
	public Terminal(ECFModel model, int ilayer, int x, int y) {
		super(model,x,y);
		this.ilayer = ilayer;
		width = 2*radius;
		height = 2*radius;
		setName("Terminal"+iterminal);
		textName = new TextGeometry(new Font("SansSerif",Font.PLAIN,12),szName);
		conn = new NodeConnectors(ilayer,this);
		vrNodeConnectors.add(conn);
		conn.setHasFlowBalance(false);
		iterminal++;
	} // Terminal

	/**
	* Checks if the terminal belongs to layer <code>ilayer</code> and contains
	* the coordinates <code>x,y</code>.
	* @param ilayer index of the layer to which the terminal belongs
	* @param x,y coordinates to be contained within the terminal symbol
	*/
	public boolean contains(int ilayer,int x,int y) {
		if (ilayer != this.ilayer) return false;
		return getShape().contains(x,y);
	} // contains

	/**
	* Decode terminal items from XML code.
	* @param szXMLCode The xml description of this node.
	*/
	public void decodeXMLItems(String szTag, String szSub) {
		if (szTag.equals("ilayer")) ilayer = Integer.parseInt(szSub);
		super.decodeXMLItems(szTag,szSub);
	} // decodeXMLItems

	/** Get the NodeConnectors of curent layer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors() {
		return getNodeConnectors(ilayer);
	} // getNodeConnectors

	/** Get the NodeConnectors of layer ilayer.
	* @param ilayer The index of the layer.
	* @return The NodeConnectors of layer ilayer.*/
	public NodeConnectors getNodeConnectors(int ilayer) {
		if (ilayer==this.ilayer) return (NodeConnectors)vrNodeConnectors.elementAt(0);
		else return null;
	} // getNodeConnectors

	/**
	* Get the shape of this node symbol.
	*/
	protected Shape getShape() {
		return new Ellipse2D.Float((float)x,(float)y, (float)2*radius, (float)2*radius);
	} // getShape

	/** Check if terminal has a flow balance for this layer.
	* @return false always.*/
	public boolean hasFlowBalance(int ilayer) {return false;}

	/** List XML code for the terminal.
	* @return XML code. */
	public String listXML(String szBlanks1) {
		String szBlanks = ""; // To give one compact line
		String szXML = "";
		szXML += szBlanks +"<ilayer>"+ilayer+"</ilayer>";
		szXML += super.listXML(szBlanks);
		return szXML;
	} // listXML

	/**
	* Paint the terminal.
	* @param g2 graphics context
	* @param ilayer index of layer to be painted
	*/
	public void paint(Graphics2D g2,int ilayer) {
		if (this.ilayer==ilayer)  {
			super.paint(g2,ilayer);
			if (g2==null) return;
			textName.paint(g2, x+radius+FlowConnector.SIDE/2+2, y+2*radius+10);
		}
	} // paint

	/** Show properties of this Terminal
	* @return A string telling the properties.*/
	public String toString() {
		return "Terminal: szName="+szName +" height="+height +" width="+width
		        +" x="+x +" y="+y;
	} // toString

	/** Return the XML code tag for this terminal.
	* @return XML tag. */
	public String xmltag() {return "terminal";}

} // Terminal