import java.awt.*;
import java.awt.font.*;

// See Lazlo:Object-Oriented Programming page 226.

/** Draws a string as glyphs. <p>
* <pre>
* TextGeometry text = new TextGeometry(szName);
* text.paint(g2, x, y);
* </pre>
*/
public class TextGeometry {

	// Default font if not supplied to constructor
	public static final Font DefaultFont =
		new Font("SansSerif",Font.BOLD,12);

	/** The font used for the text*/
	protected Font font;
	/** Font renderer context. */
	protected FontRenderContext frc;
	/** Glyphs for the text. */
	protected GlyphVector glyphs;
	/** The painter paints glyphs as filled black charaters. */
	protected Painter painter = new FillPainter(Color.black);
	/** Position of text baseline of first character. */
	protected int x,y;

	/**
	* Creates a TextGeometry object with font <code>font</code>
	* and the text <code>sz</code>.
	* @param font The font to be used.
	* @param sz The text.
	*/
	public TextGeometry(Font font, String sz)
		throws NullPointerException {
		// EFFECTS: If font or sz is null throws NullPointerException;
		//   else constructs this for string sz and given font
		this.font = font;
		frc = new FontRenderContext(null,true,false);
		glyphs = font.createGlyphVector(frc,sz);
		x = y = 0;
	} // TextGeometry

	/**
	* Creates a TextGeometry object with the font of
	* the graphics context g2 and with the text <code>sz</code>.
	* @param g2 The graphics context.
	* @param sz The text.
	*/
	public TextGeometry(Graphics2D g2, String sz)
		throws NullPointerException {
		// EFFECTS: If g2 or sz is null throws NullPointerException;
		//   else constructs this for string sz using the font and font
		//   rendering context in the graphics context g2.
		frc = g2.getFontRenderContext();
		font = g2.getFont();
		glyphs = font.createGlyphVector(frc,sz);
		x = y = 0;
	} // TextGeometry

	/**
	* Creates a TextGeometry object with current font
	* and with the text <code>sz</code>.
	* @param sz The text.
	*/
	public TextGeometry(String sz)
		throws NullPointerException {
		// EFFECTS: If sz is null throws NullPointerException;
		//   else constructs this for string sz using the default font.
		this(DefaultFont,sz);
	} // TextGeometry

	/** Sets the text to be painted. */
	public void setText(String sz) {
		glyphs = font.createGlyphVector(frc,sz);
	} // setName

	public Shape shape() {
		// EFFECTS: Returns the shape of this text
		return glyphs.getOutline(x,y);
	} // shape

	public Rectangle getBounds() {
		// EFFECTS: Returns an integer Rectangle that completely encloses the text.
		return shape().getBounds();
	} // shape

	public void translate(int dx, int dy) {
		// MODIFIES: this
		// EFFECTS: Translates this text by dx and dy.
		this.x += dx;
		this.y += dy;
	} // translate

	public void paint(Graphics2D g2, int x, int y) {
		this.x = x;
		this.y = y;
		painter.paint(g2,shape());
	} // paint

} // class TextGeometry