import java.util.*; // Vector, StringTokenizer
import java.awt.*;  // Graphics
import java.awt.geom.AffineTransform;

/**
* TimeAxis is a time scale used for plotting variables as a function of time.<p>
* <img src="TimeAxis.gif">
* <p>
* A new time axis can be given new durations of individual time increments.
* <p>
* An 'original' time axis has fixed dureations with given start times
* and lengths. It can be extended to the left (before start) and to the right
* (after end). The new time increments can be given new durations which
* means that the time increments before or after are translated accordingly.
*/
public class TimeAxis {

	/** The time axis has no time increments. */
	public static final int EMPTY=0;
	/** No times are set and the axis has only one increment fron 0.0 to 1.0. */
	public static final int SINGLE_NO_TIME_INCREMENT=1;
	/** The time axis has one time increment with a given start time. */
	public static final int SINGLE_INCREMENT=2;
	/** The time axis has a given start time and increment lengths = 1.0. */
	public static final int NEW=3;
	/** The time axis has been extended with new time increments before or
	* after original and locked increments. */
	public static final int EXTENDED=4;
	/** The time axis is locked and the length of time increments can not be changed. */
	public static final int LOCKED=5;
	/** Status of time axis = "EMPTY","SINGLE_NO_TIME_INCREMENT","SINGLE_INCREMENT"
	*,"NEW","EXTENDED" or "LOCKED".*/
	private String[] szStatus = {"EMPTY","SINGLE_NO_TIME_INCREMENT","SINGLE_INCREMENT","NEW","EXTENDED","LOCKED"};

	/** The status of the time axis. */
	protected int status = EMPTY;
	/** Index of start sample of original axis. */
	protected int iincStartOrig = 0;
	/** Length (increments) of original axis. */
	protected int cincOrig = 0;
	/** Index of selected increment. */
	protected int iincSelected = 0;
	/** Tolerance for finding start and end times.*/
	private final double eps = 0.0001;

	/** A Vector of <a href="TimeIncrement.html">TimeIncrements</a> telling the times for the start of the time
	* increments. The last element is the end time for the last increment. */
	protected Vector vrtimeInc  = new Vector();

	/** A Vector of variables which were created with the TimeAxis. */
	protected Vector vrvariable = new Vector();

	/** Create a time axis with one time increment from time 0.0 to 1.0. */
	public TimeAxis() {
		vrtimeInc = new Vector();
		vrvariable = new Vector();
		status = EMPTY;
		makeSingleNoTimeIncrement();
	} // TimeAxis

	/**
	* Register a Variable for later updating when the size of the TimeAxis
	* is extended.
	* @param variable The Variable to be registered.
	*/
	public void add(Variable variable) { vrvariable.add(variable); }

	/** Adjust the length of all registered Variables to the leng of the TimeAxis. */
	public void adjustVariablesToTimeAxis() {
		Iterator iter = vrvariable.iterator();
		while (iter.hasNext()) {
			Variable variable = (Variable)iter.next();
			variable.adjustToTimeAxis();
		} // while
	} // adjustVariablesToTimeAxis

	/**
	* Reset time axis.
	* The time axis will have no time specified and unit length.
	*/
	public void clearTimeAxis() {
		vrtimeInc = new Vector();
		vrvariable = new Vector();
		status = EMPTY;
		makeSingleNoTimeIncrement();
	} // clearTimeAxis

	/**
	* Reset time axis and associated variables.
	* The time axis will have no time specified and unit length.
	* Variables will have all magnitudes erased.
	*/
	public void clearVariableData() {

		// Clear time axis
		// System.out.println("TimeAxis.clearVariableData: time axis");
		vrtimeInc = new Vector();
		status = EMPTY;
		makeSingleNoTimeIncrement();

		// Clear variables
		Iterator iter = vrvariable.iterator();
		while (iter.hasNext()) {
			Variable variable = (Variable)iter.next();
			variable.clearVariableData();
		} // while

	} // clearVariableData

	/** Make a clone of the time axis = a copy with its own data. */
	public Object clone() {
		TimeAxis timeaxis = new TimeAxis();
		timeaxis.setStates(this);
		return (Object)timeaxis;
	} // clone

	/** Decode the time axis from XML code.
	* @param szXMLCode The XML code for the time axis. */
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		status = NEW;
		vrtimeInc.clear();
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("status"))  {
					status = Integer.parseInt(szSub);
				} // if("status")
				// Keep for compatibility with vrtime vector.
				if (szTag.equals("vrtime"))  {
					// vrtime.clear();
					StringTokenizer st = new StringTokenizer(szSub);
     				while (st.hasMoreTokens()) {
         				String token = st.nextToken();
         				double time = (new Double(token)).doubleValue();
         				TimeIncrement timeinc = new TimeIncrement();
         				timeinc.setStartTime(time);
         				vrtimeInc.add(timeinc);
     				} // while
				} // if("vrtime")
				if (szTag.equals("vrtimeInc"))  {
					decodeXMLTimeIncrements(szSub);
				} // if("vrtimeInc")
				szXMLCode = xml.remainder();
			} // while
		} catch(Exception e) {
			System.out.println(e);
		} // try/catch
	} // decodeXML

	/** Decode the time increments from XML code.
	* @param szXMLCode The XML code for the time axis. */
	public void decodeXMLTimeIncrements(String szXMLCode) {
		XMLParser xml = null;
		vrtimeInc.clear();
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				if (szTag.equals("timeinc"))  {
         			TimeIncrement timeinc = new TimeIncrement();
         			timeinc.decodeXML(szSub);
         			vrtimeInc.add(timeinc);
				} // if("timeinc")
				szXMLCode = xml.remainder();
			} // while
			cincOrig = vrtimeInc.size()-1;
		} catch(Exception e) {
			System.out.println(e);
		} // try/catch
	} // decodeXMLTimeIncrements

	/** Export time axis values to file *.dat
	* @return A string with rows of label and values on each line.
	*/
	public String export() {
		String szExport = "Time:;              ";
		Iterator iter = vrtimeInc.iterator();
		while (iter.hasNext()) {
			TimeIncrement timeinc = (TimeIncrement)iter.next();
			szExport += " " + timeinc.getStartTime() + ";";
		} // while
		szExport += "\n";
		return szExport;
	} // export

	/** Export time axis values to a Matlab file *.m
	* @return A string with rows of label and values on each line.
	*/
	public String exportMatlab() {
		String szExport = "Time=[";
		if (vrtimeInc.size()>=2) {
			TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(0);
			szExport += timeinc.getStartTime();
			for (int iinc=1; iinc<vrtimeInc.size(); iinc++) {
				timeinc = (TimeIncrement)vrtimeInc.elementAt(iinc);
				szExport += "," + timeinc.getStartTime();
			} // while
		} // if
		szExport += "];\n";
		return szExport;
	} // export

	/** Export time axis values to a Maple file *.mws, <br>
	* Time:=Vector(23,[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]);
	* @param szCommandStart Command start prompt.
	* @return A string with rows of label and values on each line.
	*/
	public String exportMaple(String szCommandStart) {
		int cinc = getSize();
		String szExport = szCommandStart + "Time:=Vector("+cinc+",[";
		if (cinc>=2) {
			TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(0);
			szExport += timeinc.getStartTime();
			for (int iinc=1; iinc<cinc; iinc++) {
				timeinc = (TimeIncrement)vrtimeInc.elementAt(iinc);
				szExport += "," + timeinc.getStartTime();
			} // while
		} // if
		szExport += "]);\n";
		return szExport;
	} // export

	/** Extend the time axis with increments before and/or after the original axis.
	* @param timeStart The new start time of the time axis.
	* @param timeEnd The new end time of the time axis.*/
	public void extend(double timeStart, double timeEnd) {
		try {
			// Set start time if single increment
			setStartTime(timeStart);
			// Extend witn new increments
			if (status==SINGLE_INCREMENT || status==EXTENDED || status==LOCKED) {
				vrtimeInc = new Vector(vrtimeInc.subList(iincStartOrig,iincStartOrig+cincOrig+1));
				int cincBefore = 0;
				// Extend to the left, before current axis
				double time = value(0)-1.0;
				while (time>=timeStart-eps) {
					vrtimeInc.insertElementAt(new TimeIncrement(time),0);
					time -= 1.0;
					cincBefore++;
				} // while

				// Remember indices of original axis
				iincStartOrig = cincBefore;
				// Move selected time increment by number of inserted increments
				iincSelected += cincBefore;

				// Extend to the right, after current axis
				time = valueLast()+1.0;
				while (time<=timeEnd+eps) {
					vrtimeInc.add(new TimeIncrement(time));
					time += 1.0;
				} // while
				status = EXTENDED;
			} // if
		} catch (Exception e) {System.out.println("TimeAxis.extend: \n" + e);}
	} // extend

	/** Get index of start increment of original time axis. */
	public int getiincStartOrig() { return iincStartOrig; }
	/** Get number of increments of original time axis. */
	private int getcincOrig() { return cincOrig; }

	/** Get index of selected increment of original time axis. */
	public int getSelectedIncrement() {
		return iincSelected;
	} // getSelectedIncrement

	/** Get description of selected time increment. */
	public String getSelectedIncrementDescription() {
		TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(iincSelected);
		return timeinc.getDescription();
	} // getSelectedTime

	/** Get start time of selected time increment. */
	public Double getSelectedTime() {
		TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(iincSelected);
		return new Double(timeinc.getStartTime());
	} // getSelectedTime

	/** Get the status of the time axis. */
	public int getStatus() { return status; }
	/** Get the status as text of the time axis. */
	public String getszStatus() { return szStatus[status]; }
	/** Get number of time increments + 1 (including end time). */
	public int getSize() { return vrtimeInc.size(); }

	/** Get time for the time increment iinc.
	* @param iinc Index of the time increment. */
	public double getTime(int iinc) {
		double time = 0.0;
		if (iinc>=0 && iinc<getSize())
			time = ((TimeIncrement)vrtimeInc.elementAt(iinc)).getStartTime();
		return time;
	} // getTimeIncrements

	/** Get a vector <code>Double</code>s of start times including end time
	* of the time axis. */
	public Vector getTimeIncrements() {
		return vrtimeInc;
	} // getTimeIncrements

	/** List XML code for this timeaxis.
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = "";
		// TimeAxis
		szXML += szBlanks +"<status>" + status +"</status>\n";
		szXML += szBlanks +"<vrtimeInc>\n";
		Iterator iter = vrtimeInc.iterator();
		while (iter.hasNext()) {
			szXML += szBlanks + "  <timeinc>";;
			TimeIncrement timeinc = (TimeIncrement)iter.next();
			szXML += timeinc.listXML("");
			szXML += "</timeinc>\n";
		} // while
		szXML += szBlanks +"</vrtimeInc>\n";
		return szXML;
	} // listXML

	/** Set the time increment in a locked state which means that
	* no times or durations can be changed. The time axis can however be
	* exteded to the left and to the right. */
	public void lock() {
		if (status==NEW || status==SINGLE_INCREMENT || status==EXTENDED) {
			iincStartOrig = 0;
			cincOrig = vrtimeInc.size()-1;
			status=LOCKED;
		} // if
	} // lock

	/** Make a new time axis with start time <code>timeStart</code> and
	* end time <code>timeEnd</code>.
	* @param timeStart Start time.
	* @param timeEnd End time. */
	public void makeNew(double timeStart, double timeEnd) {
		if (status==EMPTY || status==NEW) {
			double time = timeStart;
			vrtimeInc.clear();
			while (time<=timeEnd+eps) {
				vrtimeInc.add(new TimeIncrement(time));
				time += 1.0;
			} // while
			iincStartOrig = 0;
			cincOrig = vrtimeInc.size()-1;
			status = NEW;
		} // if
	} // makeTimeAxis

	/** Make a single time increment from time  0.0 to 1.0.*/
	public void makeSingleNoTimeIncrement() {
		if (status==EMPTY) {
			vrtimeInc.add(new TimeIncrement(0.0));
			vrtimeInc.add(new TimeIncrement(1.0));
			iincStartOrig = 0;
			cincOrig = 1;
			iincSelected = 0;
			status = SINGLE_NO_TIME_INCREMENT;
		} // if
	} // makeSingleNoTimeIncrement

	/**
	* Remove a Variable that was associated with the TimeAxis.
	* @param variable The Variable to be removed.
	*/
	public void remove(Variable variable) { vrvariable.remove(variable); }

	/**
	* Renumber cases of the TimeAxis.
	* @param nFirstCase First case number.
	*/
	public void renumber(double nFirstCase) {
		try {
			// Set start case
			setStartTime(nFirstCase);
			// Set case increments
			if (status==SINGLE_INCREMENT || status==EXTENDED || status==LOCKED) {
				double nCase = nFirstCase;
				Iterator iter = vrtimeInc.iterator();
				while (iter.hasNext()) {
					TimeIncrement timeInc = (TimeIncrement)iter.next();
					timeInc.setStartTime(nCase);
					timeInc.setDescription(" ");
					nCase++;
				} // while
				status = EXTENDED;
			} // if
		} catch (Exception e) {System.out.println("TimeAxis.renumber: \n" + e);}
	}

	/** Set description of selected time increment.
	* @param szDescription Description. */
	public void setSelectedIncrementDescription(String szDescription) {
		TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(iincSelected);
		timeinc.setDescription(szDescription);
	} // getSelectedTime

	/** Set the duration <code>dtime</code> of the time increment that
	* starts at time <code>timeStartOfIncrement</code>.
	* @param timeStartOfIncrement Start time of increment.
	* @param dtime Duration of time increment. */
	public void setIncrementTime(double timeStartOfIncrement, double dtime) {
		if (dtime<0.0) return;
		if (status==NEW || status==SINGLE_INCREMENT || status==EXTENDED) {
			int iincMod = 0;
			if (timeStartOfIncrement<value(0) ||timeStartOfIncrement>valueLast()-eps) return;
			while (value(iincMod+1)<timeStartOfIncrement+eps && iincMod+2<vrtimeInc.size()) iincMod++;
			setIncrementTime(iincMod,dtime);
		} // if
	} // setIncrementTime

	/**
	* Set duration of the increment with index <code>iincMod</code>.
	* @param iincMod The index of the increment to be modified.
	* @param dtime The duration of the time increment.
	* <A NAME="setSelectedIncrement">
	*/
	public void setIncrementTime(int iincMod, double dtime) {
		// Calculate time increments of whole axis
		int cinc = vrtimeInc.size()-1;
		if (iincMod>=cinc) return;
		double[] vrdtime = new double[cinc];
		for (int iinc=0; iinc<cinc; iinc++) vrdtime[iinc] = value(iinc+1)-value(iinc);

		// Modifiy times for increasing times
		if (status==NEW || status==SINGLE_INCREMENT || iincMod>=iincStartOrig+cincOrig) {
			// Modifiy times for increasing times
			// Set increment to be modified
			vrdtime[iincMod] = dtime;
			// Adjust all time after this increment
			int iinc = iincMod;
			while(iinc<cinc) {
				vrtimeInc.setElementAt(new TimeIncrement(value(iinc)+vrdtime[iinc]),iinc+1);
				iinc++;
			} // while
		} else if (iincMod<iincStartOrig) {
			// Modifiy times for decreasing times
			// Set increment to be modified
			vrdtime[iincMod] = dtime;
			// Adjust all time after this increment
			int iinc = iincMod;
			while(iinc>=0) {
				vrtimeInc.setElementAt(new TimeIncrement(value(iinc+1)-vrdtime[iinc]),iinc);
				iinc--;
			} // while
		} // else/if

	} // setIncrementTime

	/**
	* Select increment with index <code>iinc</code>.
	* @param iinc The index of the selected increment.
	*/
	public void setSelectedIncrement(int iinc) {iincSelected = iinc; }

	/**
	* Set start time of a single time increment. The duration is not changed.
	* @param timeStart Start time of the increment.
	*/
	public void setStartTime(double timeStart) {
		if (status==SINGLE_NO_TIME_INCREMENT || status==SINGLE_INCREMENT) {
			double timeIncrement = value(1)-value(0);
			vrtimeInc.setElementAt(new TimeIncrement(timeStart),0);
			vrtimeInc.setElementAt(new TimeIncrement(timeStart+timeIncrement),1);
			status = SINGLE_INCREMENT;
		} // if
	} // setStartTime

	/** Set the state variables of the time axis equal to the states of
	* an original time axis.
	* @param timeaxisOrig The time axis to copy states from. */
	public void setStates(TimeAxis timeaxisOrig) {
		vrtimeInc = new Vector();
		Iterator iter = timeaxisOrig.vrtimeInc.iterator();
		while (iter.hasNext()) {
			TimeIncrement timeincOrig = (TimeIncrement)iter.next();
			vrtimeInc.add(timeincOrig.clone());
		} // while
		status  = timeaxisOrig.getStatus();
		iincStartOrig  = timeaxisOrig.getiincStartOrig();
		cincOrig  = timeaxisOrig.getcincOrig();
		iincSelected = timeaxisOrig.iincSelected;
	} // setStates

	/**
	* Paint time axis
	* @param g Graphics context
	* @param xAxisStart x-coordiante (pixels) for start of axis
	* @param xAxisEnd x-coordiante (pixels) for end of axis
	* @param yAxis y-coordianate of axis
	*/
	public void paint(Graphics g, int xAxisStart, int xAxisEnd, int yAxis) {
		this.paint(g,xAxisStart,xAxisEnd,yAxis,false);
	} // paint

	/**
	* Paint time axis with an optional yellow bar indicating original time axis.
	* @param g Graphics context
	* @param xAxisStart x-coordiante (pixels) for start of axis
	* @param xAxisEnd x-coordiante (pixels) for end of axis
	* @param yAxis y-coordianate of axis
	* @param fEdit = true if yellow bar for original part of time axis shall
	*                be shown during editing
	*/
	public void paint(Graphics g, int xAxisStart, int xAxisEnd, int yAxis, boolean fEdit) {
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(Color.black);
		// Axis line
		g2.drawLine(xAxisStart, yAxis, xAxisEnd, yAxis);

		// Ticks and time labels
		int cinc = vrtimeInc.size();
		if (cinc==0) return;
		int lAxis = xAxisEnd - xAxisStart;
		double timeStart = value(0);
		double timeEnd = value(cinc-1);
		double timeTot = timeEnd - timeStart;

		// Draw tick marks and time labels
		int xtimeValue=xAxisStart;
		for (int iinc=0; iinc<cinc; iinc++) {
			// Tick mark
			double timeTick = value(iinc);
			int xTick = xAxisStart + (int)(lAxis*(timeTick-timeStart)/timeTot);
			g2.drawLine(xTick, yAxis, xTick, yAxis+10);
			// Time label
			TimeIncrement timeinc = (TimeIncrement)vrtimeInc.elementAt(iinc);
			String szTime = String.valueOf(timeinc.getStartTime());
			if (cinc*40>lAxis) {
				if (xTick>=xtimeValue) {
					g2.drawLine(xTick, yAxis, xTick, yAxis+15);
					g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
					AffineTransform transf = new AffineTransform();
					transf.rotate(Math.toRadians(45), xTick+10, yAxis+10);
					g2.setTransform(transf);
					g2.drawString(szTime, xTick+10, yAxis+20);
					transf = new AffineTransform();
					g2.setTransform(transf);
					xtimeValue+=40;
				} // if
			} else {
				g2.drawString(szTime, xTick+3, yAxis+20);
			} // if
		} // for
		// Last tick mark
		g2.drawLine(xAxisEnd, yAxis, xAxisEnd, yAxis+10);

		// Indicate original axis with a yellow bar if editing axis
		if (fEdit) {
			g2.setColor(Color.yellow);
			double timeStartOrig = value(iincStartOrig);
			double timeEndOrig = value(iincStartOrig+cincOrig);
			double dtime = timeEndOrig - timeStartOrig;
			timeStart = value(0);
			timeEnd = value(vrtimeInc.size()-1);
			timeTot = timeEnd - timeStart;
			int xStartOrig = xAxisStart + (int)(lAxis*(timeStartOrig-timeStart)/timeTot);
			int widthOrig = (int)(lAxis*dtime/timeTot);
			g2.fillRect(xStartOrig, yAxis-5, widthOrig,5);
		} // if

		// Show selected increment by two vertical lines
		g2.setColor(Color.red);
		double timeLine = value(iincSelected);
		int xLine = xAxisStart + (int)(lAxis*(timeLine-timeStart)/timeTot);
		g2.drawLine(xLine, yAxis, xLine, yAxis-30);
		timeLine = value(iincSelected+1);
		xLine = xAxisStart + (int)(lAxis*(timeLine-timeStart)/timeTot);
		g2.drawLine(xLine, yAxis, xLine, yAxis-30);

	} // paint

	/** Returns a string representation of the time axis.
	* @return A string representation of the time axis.*/
    public String toString() {
		return "TimeAxis: status=" + szStatus[status] + " original inc=[" +
				iincStartOrig + "," + (iincStartOrig+cincOrig) + "]" + " size=" + vrtimeInc.size();
	} // toString

	/** Get the double value for the time of the increment with index iinc.
	* @param iinc The index for the time increment. */
	public double value(int iinc) { return ((TimeIncrement)vrtimeInc.elementAt(iinc)).getStartTime(); }
	private double valueLast() { return ((TimeIncrement)vrtimeInc.lastElement()).getStartTime(); }

} // TimeAxis