import java.awt.*;

/** A connector that connects to one flowpath outside a frame and
* one flowpath inside the frame. */
public class TwinFlowConnector {

	/** The flow connector of the by which the flowpath is connected
	* to the parent sector.*/
	protected FlowConnector flowconnSelected;
	/** Index of the layer. */
	protected int ilayer = 0;
	/** Side of model frame = ECFNode.EAST, ECFNode.WEST, ECFNode.NORTH or ECFNode.SOUTH. */
	protected int iside = ECFNode.NONE;
	/** Height of model frame. */
	protected int height;
	/** Width of model frame. */
	protected int width;
	/** X-coordiante of upper left corner of model frame. */
	protected int x;
	/** Y-coordiante of upper left corner of model frame. */
	protected int y;
	/** Inside flow path. */
	protected FlowPathOfFrame flowpathInside;
	/** Outside flow path. */
	protected FlowPath flowpathOutside;

	/** Constructs a twin flow connector.
	* @param flowpath The flow path to be added. */
	public TwinFlowConnector(FlowPath flowpath, ModelFrame modelframe) {

		flowconnSelected = flowpath.getSelectedConnector();
		Flow flow = flowpath.getFlow();
		ilayer = flow.getLayerIndex();
		iside = flowconnSelected.getSide();

		flowpathOutside = new FlowPath(modelframe,ilayer,20,20,100,20,flow);
		flowpathInside = new FlowPathOfFrame(this,modelframe,ilayer,20,40,100,40,flow);
		modelframe.registerFlowPathOfFrame(flowpathInside);

		// Set head and tail ends.
		int iend = flowconnSelected.getEndOfFlow();
		if (iend==FlowPath.END_HEAD) {
			// Head end was connected to outside,
			// connect tail end of inside flow path to inside
			flowpathInside.setSelectedConnector(FlowPath.END_TAIL);
			flowpathInside.hideConnectorPermanently(FlowPath.END_TAIL);
			flowpathOutside.setSelectedConnector(FlowPath.END_HEAD);
		} else {
			// Tail end was connected to outside,
			// connect head end of inside flow path to inside
			flowpathInside.setSelectedConnector(FlowPath.END_HEAD);
			flowpathInside.hideConnectorPermanently(FlowPath.END_HEAD);
			flowpathOutside.setSelectedConnector(FlowPath.END_TAIL);
		} // if/else
		flowpathOutside.hideConnectorPermanently(FlowPath.END_HEAD);
		flowpathOutside.hideConnectorPermanently(FlowPath.END_TAIL);
		flowpathInside.hideLabelPermanently();

	} // TwinFlowConnector

	/**
	* Decode twin flow connector from XML code.
	* @param szXMLCode The xml description of the twin flow connector.
	*/
	public void decodeXML(String szXMLCode) {
//		System.out.println("TwinFlowConnector.decodeXML: " + szXMLCode);
		XMLParser xml = null;
		// Decode XML code
		try {
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
//				System.out.println("TwinFlowConnector.decodeXML: szTag= " + szTag);
				if (szTag.equals("flowpathOutside")) {
//					System.out.println("TwinFlowConnector.decodeXML: szSub= " + szSub);
					flowpathOutside.decodeXMLLabelPos(szSub);
				} // if
				szXMLCode = xml.remainder();
			} // while szXMLCode
		} catch(Exception e) {
			System.out.println("TwinFlowConnector.decodeXML: " + e +
			"\n  " + szXMLCode);
		} // try/catch
	} // decodeXML

	/** */
	public FlowPath findFlowPath(int ilayer,int x,int y) {
		if (flowpathInside!= null && flowpathInside.contains(ilayer,x,y))
			return flowpathInside;
		if (flowpathOutside!= null && flowpathOutside.contains(ilayer,x,y))
			return flowpathOutside;
		return null;
	} // findFlowPath

	/** Get flow path inside.
	* @param
	* @return The flow path on the inside.*/
	public FlowPathOfFrame getFlowPathInside() {
		return flowpathInside;
	} // getFlowPathInside

	/** Get side of frame to which connector is connected.
	* @return side of frame.*/
	public int getSideOfFrame() {
		return iside;
	} // iside;}

	/** Ceck if the connector has flow paths with flow.
	* @param flow The flow to be tested.
	* @return true if flow paths have flow, else false. */
	public boolean hasFlow(Flow flow) {
		if (flowpathInside.getFlow()==flow) return true;
		else return false;
	} // hasFlow


	/** List XML code for the twin flow connector.
	* @param szBlanks a string of blanks indenting the xml code
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = "";
		szXML += "<flowpathOutside>";
		szXML += flowpathOutside.listXMLLabelPos("");
		szXML += "</flowpathOutside>";
		return szXML;
	} // listXML

	/** Make outside flow paths on one side.
	* @param ilayer Index of the layer in question.
	* @param vrflowconn Vector of flow connectors.
	* @param lPathHoriz Length (pixels) of the horizontal flow paths.
	* @param lPathVert Length (pixels) of the vertical flow paths.
	* @param width for calculating horizontal step between connectors.
	* @param heigt for calculating vertical step between connectors.
	* @param offsetHorz Horizontal offset for connectors.
	* @param offsetVert Vertical offset for connectors.
	*/
	private void makeFlowPathsOnOneSide(int ilayer, FlowConnector flowconn, // Vector vrflowconn,
		int lPathHoriz, int lPathVert, int width, int height,
		int offsetHorz, int offsetVert) {
		int xf,yf; // Flow path coordinates (pixels) at frame.
		int cflow = 1;  // Provisional !!!!!!!!!!!!
		int stepHorz = width/cflow; // Step (pixels) between flow connecting points.
		int stepVert = height/cflow; // Step (pixels) between flow connecting points.

		xf = x + stepHorz/2 + offsetHorz;
		yf = y + stepVert/2 + offsetVert;
		Flow flow = flowconn.getFlowPath().getFlow();
		if (flowconn.isTailEnd()) {
			flowpathOutside = new FlowPath(null,ilayer,xf+lPathHoriz,yf+lPathVert,xf,yf,flow);
			flowpathInside = new FlowPathOfFrame(this,null,ilayer,xf,yf,xf-lPathHoriz,yf-lPathVert,flow);
		} else {
			flowpathOutside = new FlowPath(null,ilayer,xf,yf,xf+lPathHoriz,yf+lPathVert,flow);
			flowpathInside = new FlowPathOfFrame(this,null,ilayer,xf-lPathHoriz,yf-lPathVert,xf,yf,flow);
		} // if/else
	} // makeFlowPathsOnOneSide

	/**
	* Paint the flowpaths on layer with index ilayer.
	* @param g the graphics context.
	* @param ilayer index of the layer to which the flowpath belongs.
	*/
	public void paint(Graphics g,int ilayer) {
		if (this.ilayer==ilayer) {
			flowpathOutside.paint(g,ilayer);
			flowpathInside.paint(g,ilayer);
		} // if
	} // paint

	/** Remove outside and inside flow paths from the connector. */
	public void removeFlowPaths() {
		flowpathInside.disconnectThis();
		flowpathInside = null;
		flowpathOutside.disconnectThis();
		flowpathOutside = null;
	} // removeFlowPaths

	/** Set index of flow path, counting on each side of the frame.
	* @param ipath The index along on side of the frame. */
	public void setFlowPathIndex(int ipath) { flowpathInside.setFlowPathIndex(ipath); }


	/** Set the position of the connector and update flowlines.
	* @param x the x-coordinate of the reference point.
	* @param y the y-coordinate of the reference point.
	*/
	public void setPos(int x, int y, int height, int width) {
		FlowConnector connOutFree;  // Free end of outside flow path
		FlowConnector connOutFrame; // Frame end of outside flow path
		FlowConnector connInFrame;  // Frame end of inside flow path
		FlowConnector connInFree;   // Free end of inside flow path
		int lPathHoriz=100;
		int lPathVert = 50;

		if (flowconnSelected.isHeadEnd()) {
			connOutFree = flowpathOutside.getFlowConnector(FlowPath.END_TAIL);
			connOutFrame = flowpathOutside.getFlowConnector(FlowPath.END_HEAD);
			connInFrame = flowpathInside.getFlowConnector(FlowPath.END_TAIL);
			connInFree = flowpathInside.getFlowConnector(FlowPath.END_HEAD);
		} else {
			connOutFree = flowpathOutside.getFlowConnector(FlowPath.END_HEAD);
			connOutFrame = flowpathOutside.getFlowConnector(FlowPath.END_TAIL);
			connInFrame = flowpathInside.getFlowConnector(FlowPath.END_HEAD);
			connInFree = flowpathInside.getFlowConnector(FlowPath.END_TAIL);
		} // if /else

		// Set connector positions at frame
		connOutFrame.setPos(x,y);
		connInFrame.setPos(x,y);

		if (iside==ECFNode.WEST) {
			connOutFree.setPos(x-lPathHoriz,y);
			connInFrame.setSide(ECFNode.EAST);
			if (connInFree.getSide()==ECFNode.NONE) connInFree.setPos(x+lPathHoriz,y);
		} // if WEST

		if (iside==ECFNode.EAST) {
			connOutFree.setPos(x+lPathHoriz,y);
			connInFrame.setSide(ECFNode.WEST);
			if (connInFree.getSide()==ECFNode.NONE) connInFree.setPos(x-lPathHoriz,y);
		} // if EAST

		if (iside==ECFNode.NORTH) {
			connOutFree.setPos(x,y-lPathVert);
			connInFrame.setSide(ECFNode.SOUTH);
			if (connInFree.getSide()==ECFNode.NONE) connInFree.setPos(x,y+lPathVert);
		} // NORTH

		if (iside==ECFNode.SOUTH) {
			connOutFree.setPos(x,y+lPathVert);
			connInFrame.setSide(ECFNode.NORTH);
			if (connInFree.getSide()==ECFNode.NONE) connInFree.setPos(x,y-lPathVert);
		} // SOUTH

	} // setPos

	public String toString() {
		return "TwinFlowConnector: flowpathInside= "+ flowpathInside;
	} // toString

} // TwinFlowConnector
