import java.awt.*;
import java.util.*; // StringTokenizer

/**
* A <code>Variable</code> has an array of values and a <A href=TimeAxis.html>TimeAxis</A>.
* The variable has one element less than the time axis.<p>
* The base value is the value used in calculations.<br>
* The display value is the value used in text and diagrams
* and depends upon which unit from the set of units (inherited from Quantity)
* that was selected.<p>
* The paint method uses the paint
* methods of <code>TimeAxis</code>, <A href=ValueAxis.html>ValueAxis</A> and
* <A href=ValuesPlot.html>ValuesPlot</A>. A plot looks like this: <p>
* <img src="VariablePlot.gif">
*/

public class Variable extends Quantity implements SimpleVariable {

	/** Indicate if default equation shall be used.*/
	protected static final int DEFAULT=0;
	/** Indicate if flow is prescribed by user.*/
	protected static final int PRESCRIBED=1;
	/** Indicate if flow is calculated from node equation.*/
	protected static final int CALCULATED=2;
	/** The names of the states.*/
	public static String[] STATES = {"DEFAULT", "PRESCRIBED", "CALCULATED"};

	/** A time axis with intervals matching the values of Variable. */
	protected TimeAxis timeaxis;
	/** Array of prescribed base values. */
	protected double[] vrvaluePrescribed;
	/** Array of calculated base values. */
	protected double[] vrvalueCalculated;
	/** Variable state, DEFAULT, PRESCRIBED or CALCULATED.*/
	protected int state = DEFAULT;
	/** Flag indicating if values are observed or measured. */
	protected boolean fMeasured = false;
	/** The name of the units set for the Quantity. */
	protected String szNameUnits;
	/** Reference to the data source. */
	protected String szDataSource;

	/**
	* Creates a Variable that matches timeaxis, all values are zero.
	*/
	public Variable() {
		timeaxis = ECF.getTimeAxis();
		int cinc = timeaxis.getTimeIncrements().size()-1;
		vrvaluePrescribed = new double[cinc];
		vrvalueCalculated = new double[cinc];
		timeaxis.add(this); // register with timeaxis.
	} // Variable

	/**
	* Adjusts length of variable to the length of the time axis, when extended.
	* Previous values are kept at the same time positions as before.
	*/
	public void adjustToTimeAxis() {
		int cinc = timeaxis.getTimeIncrements().size()-1;
		if (vrvaluePrescribed.length<cinc) {
			// Make a new value array
			double[] vrvaluePrescribedNew = new double[cinc];
			double[] vrvalueCalculatedNew = new double[cinc];
			int cvalueAfter = vrvaluePrescribedNew.length;
			// Put zeroes into new arrays
			for (int ivalue=0; ivalue<cvalueAfter; ivalue++)  {
				vrvaluePrescribedNew[ivalue] = 0.0;
				vrvalueCalculatedNew[ivalue] = 0.0;
			} // for

			// Copy old values into new array
			// int ivalue;
			int cvalueBefore = vrvaluePrescribed.length;
			int iincStartOrig = timeaxis.getiincStartOrig();
			for (int ivalue=0; ivalue<cvalueBefore; ivalue++) {
				vrvaluePrescribedNew[ivalue+iincStartOrig] = vrvaluePrescribed[ivalue];
				vrvalueCalculatedNew[ivalue+iincStartOrig] = vrvalueCalculated[ivalue];
			} // for
			// Replace old arrays by new arrays
			vrvaluePrescribed = vrvaluePrescribedNew;
			vrvalueCalculated = vrvalueCalculatedNew;
		} // if
	} // setTimeAxis

	/**
	* Reset variables.
	* Variables will have unit length and all magnitudes erased.
	*/
	public void clearVariableData() {
		vrvaluePrescribed = new double[1];
		vrvalueCalculated = new double[1];
	} // clearVariableData

	/** Copy one value with index itime to all prescribed values.
	* @param itime Index of the time instance which value is copied. */
	public void copyToAll(int itime) {
		int ctime = timeaxis.getTimeIncrements().size()-1;
		for (int itimeCopyTo=0; itimeCopyTo<ctime; itimeCopyTo++) {
			vrvaluePrescribed[itimeCopyTo] = vrvaluePrescribed[itime];
		} // for
	} // copyToAll

	/**
	* Decodes variable from XML code.
	* @param szXMLCode The xml description of this model.
	*/
	public void decodeXML(String szXMLCode) {
		XMLParser xml = null;
		// Decode XML code
		try {
			// Make a new value array
			int cinc = timeaxis.getTimeIncrements().size()-1;
			double[] vrvalueNew = new double[cinc];
			while (szXMLCode.length()>0) {
				xml = new XMLParser(szXMLCode);
				String szTag = xml.tag();
				String szSub = xml.sub();
				// Own attributes
				if (szTag.equals("state")) state = Integer.parseInt(szSub);
				if (szTag.equals("fMeasured")) fMeasured = (Boolean.valueOf(szSub)).booleanValue();
				if (szTag.equals("szNameUnits")) szNameUnits = szSub;
				if (szTag.equals("szDataSource")) szDataSource = szSub;
				if (szTag.equals("vrvaluePrescribed"))  {
					StringTokenizer st = new StringTokenizer(szSub);
					int ctok = st.countTokens();
					int ivalue = 0;
     				while (st.hasMoreTokens() && ivalue<cinc) {
         				String token = st.nextToken();
         				try {
         					vrvalueNew[ivalue] = Double.parseDouble(token);
						} catch(Exception e) {
							System.out.println("Variable.decodeXML: token=" + token
							+ " time=\n" + timeaxis.getTime(ivalue)+ e);
						} // try/catch
         				ivalue++;
     				} // while
				} // if vrvaluePrescribed
				// Attribute inherited from Quantity
				if (szTag.equals("szName")) {
					szName = szSub;
				} // if szName
				szXMLCode = xml.remainder();
			} // while
			// Replace old array by new array
			vrvaluePrescribed = vrvalueNew;
		} catch(Exception e) {
			System.out.println("Variable.decodeXML:" + e);
		} // try/catch
	} // decodeXML

	/** Export Varialble values
	* @return A string with label and values separated by semicolon (;).
	*/
	public String export() {
		Units units = getUnits();
		String szNameUnit = " ";
		if (units!=null) szNameUnit = units.getNameOfDisplayUnit();
		String szExport = ECFUtil.removeLineFeeds(szName) + "(" + szNameUnit + ");";
		szExport = Quantity.padToLength(szExport,20);

		double[] vrvalueDisplay = getDisplayValuesCalculated();
		for (int ivalue=0; ivalue<vrvalueDisplay.length; ivalue++)
			szExport += " " + Quantity.formatDouble(vrvalueDisplay[ivalue]) + ";";
		szExport += "\n";
		return szExport;
	} // export

	/** Export Variable values in Matlab format.
	* @param szType = Type of variable.
	* @param ivar Variable index.
	* @return A string with label and values.
	*/
	public String exportMatlab(String szType, int ivar) {
		Units units = getUnits();
		String szNameUnit = " ";
		if (units!=null) szNameUnit = units.getNameOfDisplayUnit();
		String szExport = "vrName"+szType+"{" + ivar + "}='" +
			ECFUtil.removeLineFeeds(szName)	+ "(" + szNameUnit + ")';";
		szExport = Quantity.padToLength(szExport,30);

		double[] vrvalueDisplay = getDisplayValuesCalculated();
		if (vrvalueDisplay.length>=1) {
			szExport += "  vr"+szType+"("+ ivar + ",:)=["+ vrvalueDisplay[0];
			for (int ivalue=1; ivalue<vrvalueDisplay.length; ivalue++)
				szExport += "," + vrvalueDisplay[ivalue];

		} // if
		szExport += "];\n";
		return szExport;
	} // export

	/** Export Variable values in Maple format.
	* vrNameType:=Vector(cvar);
	* vrType:=Vector(cvar);
	* vrNameType[ivar]:= 'szName';
	* vrType[ivar]:=Vector(ctime,[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]);
	* @param szType = Type of variable (Flow, Ratio ...) shown in variable name.
	* @param ivar Variable index.
	* @param szCommandStart Command start prompt.
	* @return A string with label and values.
	*/
	public String exportMaple(String szType, int ivar, String szCommandStart) {
		Units units = getUnits();
		String szNameUnit = " ";
		if (units!=null) szNameUnit = units.getNameOfDisplayUnit();
		String szExport = szCommandStart +
			"vrName"+szType+"[" + ivar + "]:=\"" +
			ECFUtil.removeLineFeeds(szName) + "(" + szNameUnit + ")\";\n";

		double[] vrvalueDisplay = getDisplayValuesCalculated();
		if (vrvalueDisplay.length>=1) {
			szExport += szCommandStart +
				"vr"+szType+"["+ ivar + "]:=Vector("+size()+",["+ vrvalueDisplay[0];
			for (int ivalue=1; ivalue<vrvalueDisplay.length; ivalue++)
				szExport += "," + vrvalueDisplay[ivalue];
		} // if
		szExport += "]);\n";
		return szExport;
	} // export

	/** Gets the calculated base value for the current time increment.
	* @return The base value. */
	public double getBaseValueCalculated() {
		double value;
		int iinc = timeaxis.getSelectedIncrement();
		int cvalue = vrvalueCalculated.length;
		value = vrvalueCalculated[timeaxis.getSelectedIncrement()];
		return value;
	} // getBaseValueCalculated

	/** Gets the prescribed base value for the current time increment.
	* @return The base value. */
	public double getBaseValuePrescribed() {
		double value;
		int iinc = timeaxis.getSelectedIncrement();
		int cvalue = vrvaluePrescribed.length;
		if (iinc<cvalue)
			value =vrvaluePrescribed[timeaxis.getSelectedIncrement()];
		else
			value = 0.0;
		return value;
	} // getBaseValuePrescribed

	/** Get prescribed variable scaled value for equations.
	* @return Prescribed value scaled for calculations. */
	public double getCalcValuePrescribed() {
		return getBaseValuePrescribed()/getScaleForCalc();
	} // getCalcValuePrescribed

	/** Get reference to data source. */
	public String getDataSource() { return szDataSource; }

	/**
	* Get the arrays of display values for the variable.
	* @return Two arrays of display values, (using current unit).
	*/
	public Color[] getDisplayColors() {
		Color vrcolor[] = {ECFModel.COLOR_PRESCRIBED, ECFModel.COLOR_CALCULATED};
		return vrcolor;
	} // getDisplayColors

	/**
	* Gets the calculated display value of the Variable for the current time increment, selected
	* for the <A href="TimeAxis.html#setSelectedIncrement">TimeAxis</A>.
	* @return The display value at the selected time.
	*/
	public double getDisplayValueCalculated() {
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		return factor*vrvalueCalculated[timeaxis.getSelectedIncrement()];
	} // getDisplayValueCalculated

	/**
	* Gets the prescribed display value of the Variable for the current time increment, selected
	* for the <A href="TimeAxis.html#setSelectedIncrement">TimeAxis</A>.
	* @return The display value at the selected time.
	*/
	public double getDisplayValuePrescribed() {
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		return factor*vrvaluePrescribed[timeaxis.getSelectedIncrement()];
	} // getDisplayValuePrescribed

	/**
	* Get the arrays of display values for the variable.
	* @return Two arrays of display values for prescribed
	* and calculated values, (using current unit).
	*/
	public double[][] getDisplayValues() {
		double[][] mxvalueDisplay = new double[2][];
		mxvalueDisplay[0] = getDisplayValues(vrvaluePrescribed);
		mxvalueDisplay[1] = getDisplayValues(vrvalueCalculated);
		return mxvalueDisplay;
	} // getDisplayValues

	/**
	* Get the array of display values for the variable.
	* @param vrvalue An array of base values.
	* @return An array of display values, (using current unit).
	*/
	private double[] getDisplayValues(double[] vrvalue) {
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		double[] vrvalueDisplay = new double[vrvalue.length];
		for (int ivalue=0; ivalue<vrvalue.length; ivalue++)
			vrvalueDisplay[ivalue] = factor*vrvalue[ivalue];
		return vrvalueDisplay;
	} // getDisplayValues

	/**
	* Get an arrays of calculated display values for the variable.
	* @return One arrays of calculated display values, (using current unit).
	*/
	public double[] getDisplayValuesCalculated() {
		return getDisplayValues(vrvalueCalculated);
	} // getDisplayValues

	/** Get layer index of the variable. To be implemented by subtypes
	* of variable (flow, etc.).
	* @return Index of layer. */
	public int getLayerIndex() { return 0; }

	/** Get the calculated pixel value of the Variable.
	* @param cpixel The maximum width of an arrow symbol when the flow
	* has max display value. */
	public int getPixelValueCalculated(int cpixel) {
		if (units==null) return 0;
		return (int)(getBaseValueCalculated()*units.getScaleToPixels(cpixel));
	} // getPixelValue

	/** Get the pixel value of the difference between calculated
	* and measuresd values of Variable.
	* @param cpixel The maximum width of an arrow symbol when the flow
	* has max display value. */
	public int getPixelValueCalcMeasDiff(int cpixel) {
		if (units==null) return 0;
		double valueBaseDiff;
		if (fMeasured)
			// Prescribed values hold meassured values
			valueBaseDiff = getBaseValueCalculated()-getBaseValuePrescribed();
		else
			// No meassured values given, no difference can be calculated
			valueBaseDiff = 0;
		return (int)(valueBaseDiff*units.getScaleToPixels(cpixel));
	} // getPixelValueCalcMeasDiff

	/** Get the prescribed pixel value of the Variable.
	* @param cpixel The maximum width of an arrow symbol when the flow
	* has max display value. */
	public int getPixelValuePrescribed(int cpixel) {
		if (units==null) return 0;
		return (int)(getBaseValuePrescribed()*units.getScaleToPixels(cpixel));
	} // getPixelValue

	/** Get the scaling factor for setting variables in equation systems.
	* Variable in calculations = Base value / scale.
	* Base value = Value in calculations * scale.
	* @return Scaling factor for variables.*/
	public double getScaleForCalc() {
		Units units = getUnits();
		if (units!=null) return units.getScaleForCalc();
		else return 1.0;
	} // getScaleForCalc

	/** Get the state of the variable.
	* @return The state: DEFAULT,PRESCRIBED or CALCULATED. */
	public int getState() { return state; }

	/** Get a string with calculated display value and unit. */
	public String getStringQuantCalculated() {
		Units units = super.getUnits();
		if (units==null) return formatDouble(getDisplayValueCalculated()) +
			" no units set";
		return formatDouble(getDisplayValueCalculated()) + " " +
				units.getNameOfDisplayUnit();
	} // getStringQuantCalculated

	/** Get a string with calculated display value and unit. */
	public String getStringQuantCalcMeasDiff() {
		if (!fMeasured && !isPrescribed()) return "Not measured";
		Units units = super.getUnits();
		double valueDisplayDiff;
		if (fMeasured)
			// Prescribed values hold meassured values
			valueDisplayDiff = getDisplayValueCalculated() - getDisplayValuePrescribed();
		else
			// No meassured values given, no difference can be calculated
			valueDisplayDiff = 0;

		if (units==null) return "Diff " + formatDouble(valueDisplayDiff) +
			" no units set";
		else return "Diff " + formatDouble(valueDisplayDiff) + " " +
			units.getNameOfDisplayUnit();
//		else
//			return "0  " + units.getNameOfDisplayUnit();
	} // getStringQuantCalcMeasDiff

	/** Get a string with prescribed display value and unit. */
	public String getStringQuantPrescribed() {
		Units inits = getUnits();
		if (units==null) return formatDouble(getDisplayValuePrescribed()) +
			" no units set";
		return formatDouble(getDisplayValuePrescribed()) + " " +
				units.getNameOfDisplayUnit();
	} // getStringQuantPrescribed

	/**
	* Gets the time axis of the Variable.
	* @return The time axis.
	*/
	public TimeAxis getTimeAxis() { return timeaxis; }

	/**
	* Tell if a flow has DEFAULT state.
	* @return true if DEFAULT state, else false.
	*/
	public boolean isDefault() { return state==DEFAULT;}

	/** Tell if variable has measured values.
	* @return true if values are measured, false else. */
	public boolean isMeasured() {
		return fMeasured;
	} // setMeasured

	/**
	* Tell if a variable has a prescribed magnitude.
	* @return true if PRESCRIBED state, else false.
	*/
	public boolean isPrescribed() {
		return state==PRESCRIBED;
	} // isPrescribed

	/** Lists XML code for this variable.
	* @return XML code. */
	public String listXML(String szBlanks) {
		String szXML = szBlanks;
		// Own attributes
		szXML += "<state>"+state+"</state>";
		szXML += "<fMeasured>"+fMeasured+"</fMeasured>";
		Units units = getUnits();
		if (units==null)
			szXML += "<szNameUnits>"+"No units"+"</szNameUnits>";
		else
			szXML += "<szNameUnits>"+units.getName()+"</szNameUnits>";
		szXML += "<szDataSource>"+szDataSource+"</szDataSource>";
		// Values
		szXML += "<vrvaluePrescribed>";
		for (int ivalue=0; ivalue<vrvaluePrescribed.length; ivalue++) {
			szXML += " " + vrvaluePrescribed[ivalue];
		} // for
		szXML += "</vrvaluePrescribed>";
		// Attribute inherited from Quantity
		szXML += "<szName>"+szName+"</szName>";
		return szXML;
	} // listXML

	/**
	* Gets the largest display value of the Variable.
	* @return The largest display value.
	*/
	public double maxValue() {
		// Find max base value
		double max=0.0;
		for (int ivalue=0; ivalue<vrvaluePrescribed.length; ivalue++) {
			if (vrvaluePrescribed[ivalue]>max) max = vrvaluePrescribed[ivalue];
			if (vrvalueCalculated[ivalue]>max) max = vrvalueCalculated[ivalue];
		} // for
		// Convert to display value
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		return factor*max;
	} // maxValue

	/**
	* Paints the Variable in a diagram.
	* @param g The graphical context.
	* @param xOrigin The x-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param yOrigin The y-coordiante (pixels) in the graphical context
	* for the origin of the diagram.
	* @param width The height (pixels) of the time axis.
	* @param height The height (pixels) of the value axis.
	*/
	public void paint(Graphics g, int xOrigin, int yOrigin, int width, int height) {
		// Value axis
		Units units = getUnits(); String szNameUnit = " ";
		if (units!=null) szNameUnit = units.getNameOfDisplayUnit();
		ValueAxis valueaxis = new ValueAxis(maxValue(),szNameUnit);
		valueaxis.paint(g, xOrigin, height, yOrigin);
		// Time axis
		if (timeaxis.getTimeIncrements().size()==0) return;
		timeaxis.paint(g, xOrigin, xOrigin+width, yOrigin);
		// Values plot
		ValuesPlot valuesplot = new ValuesPlot(timeaxis.getTimeIncrements(),getDisplayValues(),getDisplayColors());
		valuesplot.paint(g,xOrigin,yOrigin,width,height,valueaxis.getScale(),
		           timeaxis.getSelectedIncrement());
	} // paint

	/** Set flow state to DEFAULT, only if CALCULATED.*/
	public void resetCalculated() { if(state==CALCULATED) state=DEFAULT; }

	/**
	* Set flow state to CALCULATED.
	*/
	public void setCalculated() { state=CALCULATED;	}

	/** Set flow state to DEFAULT.*/
	public void setDefault() {
		state=DEFAULT;
		Calculator.calculate();
	}

	/** Sets the calculated base value for the current time increment.
	* @param value The base value. */
	public void setBaseValueCalculated(double value) {
		if (units!=null) {
			vrvalueCalculated[timeaxis.getSelectedIncrement()] =
				value*units.getScaleForCalc();
		} else {
			vrvalueCalculated[timeaxis.getSelectedIncrement()] = value;
		} // if/else
	} // setBaseValueCalculated

	/** Sets the prescribed base value for the current time increment.
	* @param value The base value. */
	public void setBaseValuePrescribed(double value) {
		vrvaluePrescribed[timeaxis.getSelectedIncrement()] =value; }

	/** Set reference to data source.
	* @param szDataSource Reference to the data source. */
	public void setDataSource(String szDataSource) {
		this.szDataSource = szDataSource; }

	/**
	* Sets the base value of the Variable for the current time increment,
	* selected for the <A href="TimeAxis.html#setSelectedIncrement">
	* TimeAxis</A>.
	* @param valueDisplay The display value at the selected time.
	*        It is converted by display unit to base value.
	*/
	public void setDisplayValuePrescribed(double valueDisplay) {
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		vrvaluePrescribed[timeaxis.getSelectedIncrement()] =
			valueDisplay/factor;
	} // setDisplayValue

	/**
	* Sets the prescribed base values to <code>vrvalueDisplay</code>
	* converted from display unit.
	* @param vrvalueDisplay Value array in display unit.
	*/
	public void setDisplayValuesPrescribed(double[] vrvalueDisplay) {
		double factor = 1.0;
		if (units!=null) factor = units.getFactor();
		for (int ivalue=0; ivalue<vrvaluePrescribed.length; ivalue++)
			vrvaluePrescribed[ivalue] = vrvalueDisplay[ivalue]/factor;
	} // setDisplayValues

	/** Set state to PRESCRIBED. */
	public void setPrescribed() {
		state=PRESCRIBED;
		Calculator.calculate();
	} // setPrescribed

	/** Set flag telling that variable has measured values.
	* @param fMeasured = true if values are measured, = false else. */
	public void setMeasured(boolean fMeasured) {
		this.fMeasured=fMeasured;
	} // setMeasured

	/**
	* Gets the number of values of the Variable.
	* @return The number of values = length of value array.
	*/
	public int size() { return vrvaluePrescribed.length; }

	/** Shows properties of the Variable.
	* @return A string telling the properties.*/
	public String toString() {
		String sz = "\nVariable Prescribed: ";
		for (int ivalue=0; ivalue<vrvaluePrescribed.length; ivalue++) sz +=
			vrvaluePrescribed[ivalue] + " ";
		sz += "\nVariable Calculated: ";
		for (int ivalue=0; ivalue<vrvalueCalculated.length; ivalue++) sz +=
			vrvalueCalculated[ivalue] + " ";
		return sz;
	} // toString

	/** Print Maple format of base value assignment.
	* @param szCommandStart Command start prompt.
	* @return A string with a Maple base value assignment.
	*/
	public String toStringBaseValueMaple(String szCommandStart) {
		return "\n" + szCommandStart + getName() + ":="
			+ getBaseValuePrescribed() + ";";
	} // toStringBaseValueMaple

	/** Print matlab format of base value assignment.
	* @return A string with a Matlab base value assignment.
	*/
	public String toStringBaseValueMatlab() {
		return "\n" + getName() + "="
			+ getBaseValuePrescribed() + ";";
	} // toStringBaseValueMatlab

	/** Print Maple format of value calculation value and scale assignments.
	* @param szCommandStart Command start prompt.
	* @return A string with Maple value assignments.
	*/
	public String toStringCalcValueAndScaleMaple(String szCommandStart) {
		String sz = "";
		sz += "\n" + szCommandStart + getName() + ":="
				+ getCalcValuePrescribed() + ";";
		sz += "\n" + szCommandStart + "scale" + getName() + "Calc:="
				+ getScaleForCalc() + ";";
		return sz;
	} // toStringCalcValueAndScaleMaple

	/** Print matlab format of value calculation value and scale assignments.
	* @return A string with a Matlab value assignment.
	*/
	public String toStringCalcValueAndScaleMatlab() {
		String sz = "";
		sz += "\n" + getName() + "="
				+ getCalcValuePrescribed() + ";";
		sz += "\n" + "scale" + getName() + "Calc="
				+ getScaleForCalc() + ";";
		return sz;
	} // toStringCalcValueAndScaleMatlab

	/** Show properties of the Variable (Flow har its own).
  	* @param szType The typ of variable, e.g. FlowRatio
	* @return A string telling the properties.*/
	public String toStringState(String szType) {
		String szNameNoLineFeeds = ECFUtil.removeLineFeeds(szName);
		String sz = "\t" + szType + ": ";
		sz += Quantity.padToLength(szNameNoLineFeeds,25) + " ";
		if (state==DEFAULT)    sz += "DEFAULT=0  ";
		if (state==CALCULATED) sz += "CALCULATED= " + getStringQuantCalculated();
		if (state==PRESCRIBED) sz += "PRESCRIBED= " + getStringQuantPrescribed();
		return sz;
	} // toString

} // Variable